@@ -181,6 +181,13 @@ else
   fi
 fi
 
+#-------------------------------------------------------------------------------
+# Make sure the version string is not longer than 25 characters
+#-------------------------------------------------------------------------------
+if test ${#VERSION} -gt 25; then
+  VERSION="${VERSION:0:19}...${VERSION: -3}"
+fi
+
 #-------------------------------------------------------------------------------
 # Print the version info and exit if necassary
 #-------------------------------------------------------------------------------
