@@ -30,9 +30,10 @@
 #include <errno.h>
 #include <signal.h>
 #include <stdio.h>
+#include <sys/resource.h>
 #include <sys/types.h>
 #include <sys/wait.h>
-#ifdef __macos__
+#ifdef __APPLE__
 #include <AvailabilityMacros.h>
 #endif
 
@@ -94,6 +95,8 @@ XrdScheduler::XrdScheduler(XrdSysError *eP, XrdOucTrace *tP,
               : XrdJob("underused thread monitor"),
                 WorkAvail(0, "sched work")
 {
+    struct rlimit rlim;
+
     XrdLog      =  eP;
     XrdTrace    =  tP;
     min_Workers =  minw;
@@ -111,6 +114,29 @@ XrdScheduler::XrdScheduler(XrdSysError *eP, XrdOucTrace *tP,
     num_Limited =  0;
     firstPID    =  0;
     WorkFirst = WorkLast = TimerQueue = 0;
+
+// Make sure we are using the maximum number of threads allowed (Linux only)
+//
+#if defined(__linux__) && defined(RLIMIT_NPROC)
+
+// Get the resource thread limit and set to maximum. In Linux this may be -1
+// to indicate useless infnity, so we have to come up with a number, sigh.
+//
+   if (!getrlimit(RLIMIT_NPROC, &rlim) && rlim.rlim_max > 0)
+      {if (rlim.rlim_cur < rlim.rlim_max)
+          {rlim.rlim_cur = rlim.rlim_max;
+           setrlimit(RLIMIT_NPROC, &rlim);
+          }
+      }
+
+// Readjust our internal maximum to be the actual maximum
+//
+   if (!getrlimit(RLIMIT_NPROC, &rlim))
+      {if (rlim.rlim_cur < 1) max_Workers = 127000;
+          else max_Workers = static_cast<int>(rlim.rlim_cur);
+      }
+#endif
+
 }
  
 /******************************************************************************/
@@ -228,7 +254,7 @@ void *XrdScheduler::Reaper()
    int status;
    pid_t pid;
    XrdSchedulerPID *tp, *ptp, *xtp;
-#if defined(__macos__) && !defined(MAC_OS_X_VERSION_10_5)
+#if defined(__APPLE__) && !defined(MAC_OS_X_VERSION_10_5)
    struct timespec ts = { 1, 0 };
 #else
    sigset_t Sset;
@@ -256,7 +282,7 @@ void *XrdScheduler::Reaper()
                 } else {ptp = tp; tp = tp->next;}
              }
        ReaperMutex.UnLock();
-#if defined(__macos__) && !defined(MAC_OS_X_VERSION_10_5)
+#if defined(__APPLE__) && !defined(MAC_OS_X_VERSION_10_5)
        // Mac OS X sigwait() is broken on <= 10.4.
       } while (nanosleep(&ts, 0) <= 0);
 #else
@@ -425,8 +451,8 @@ void XrdScheduler::setParms(int minw, int maxw, int avlw, int maxi, int once)
 // get a consistent view of all the values
 //
    if (maxw <= 0) maxw = max_Workers;
-   if (minw < 0) minw = (maxw/10 ? maxw/10 : 1);
-      else if (minw > maxw) minw = maxw;
+   if (minw < 0)  minw = min_Workers;
+   if (minw > maxw) minw = maxw;
    if (avlw < 0) avlw = maxw/4*3;
       else if (avlw > maxw) avlw = maxw;
 
