@@ -21,7 +21,7 @@
 #include "XrdCl/XrdClLog.hh"
 #include "XrdCl/XrdClMessage.hh"
 #include "XrdCl/XrdClAsyncSocketHandler.hh"
-#include "XrdSys/XrdSysDNS.hh"
+#include <netinet/tcp.h>
 
 namespace XrdCl
 {
@@ -56,9 +56,9 @@ namespace XrdCl
     env->GetInt( "TimeoutResolution", timeoutResolution );
     pTimeoutResolution = timeoutResolution;
 
-    memset( &pSockAddr, 0, sizeof( pSockAddr ) );
     pSocket = new Socket();
     pIncHandler = std::make_pair( (IncomingMsgHandler*)0, false );
+    pLastActivity = time(0);
   }
 
   //----------------------------------------------------------------------------
@@ -76,13 +76,13 @@ namespace XrdCl
   Status AsyncSocketHandler::Connect( time_t timeout )
   {
     Log *log = DefaultEnv::GetLog();
-    pConnectionStarted = ::time(0);
+    pLastActivity = pConnectionStarted = ::time(0);
     pConnectionTimeout = timeout;
 
     //--------------------------------------------------------------------------
     // Initialize the socket
     //--------------------------------------------------------------------------
-    Status st = pSocket->Initialize();
+    Status st = pSocket->Initialize( pSockAddr.Family() );
     if( !st.IsOK() )
     {
       log->Error( AsyncSockMsg, "[%s] Unable to initialize socket: %s",
@@ -91,13 +91,55 @@ namespace XrdCl
       return st;
     }
 
+    //--------------------------------------------------------------------------
+    // Set the keep-alive up
+    //--------------------------------------------------------------------------
+    Env *env = DefaultEnv::GetEnv();
+
+    int keepAlive = DefaultTCPKeepAlive;
+    env->GetInt( "TCPKeepAlive", keepAlive );
+    if( keepAlive )
+    {
+      int    param = 1;
+      Status st    = pSocket->SetSockOpt( SOL_SOCKET, SO_KEEPALIVE, &param,
+                                          sizeof(param) );
+      if( !st.IsOK() )
+        log->Error( AsyncSockMsg, "[%s] Unable to turn on keepalive: %s",
+                    st.ToString().c_str() );
+
+#if defined(__linux__) && defined( TCP_KEEPIDLE ) && \
+    defined( TCP_KEEPINTVL ) && defined( TCP_KEEPCNT )
+
+      param = DefaultTCPKeepAliveTime;
+      env->GetInt( "TCPKeepAliveTime", param );
+      st = pSocket->SetSockOpt(SOL_TCP, TCP_KEEPIDLE, &param, sizeof(param));
+      if( !st.IsOK() )
+        log->Error( AsyncSockMsg, "[%s] Unable to set keepalive time: %s",
+                    st.ToString().c_str() );
+
+      param = DefaultTCPKeepAliveInterval;
+      env->GetInt( "TCPKeepAliveInterval", param );
+      st = pSocket->SetSockOpt(SOL_TCP, TCP_KEEPINTVL, &param, sizeof(param));
+      if( !st.IsOK() )
+        log->Error( AsyncSockMsg, "[%s] Unable to set keepalive interval: %s",
+                    st.ToString().c_str() );
+
+      param = DefaultTCPKeepAliveProbes;
+      env->GetInt( "TCPKeepAliveProbes", param );
+      st = pSocket->SetSockOpt(SOL_TCP, TCP_KEEPCNT, &param, sizeof(param));
+      if( !st.IsOK() )
+        log->Error( AsyncSockMsg, "[%s] Unable to set keepalive probes: %s",
+                    st.ToString().c_str() );
+#endif
+    }
+
     pHandShakeDone = false;
 
     //--------------------------------------------------------------------------
     // Initiate async connection to the address
     //--------------------------------------------------------------------------
     char nameBuff[256];
-    XrdSysDNS::IPFormat( (sockaddr*)&pSockAddr, nameBuff, sizeof(nameBuff) );
+    pSockAddr.Format( nameBuff, sizeof(nameBuff), XrdNetAddrInfo::fmtAdv6 );
     log->Debug( AsyncSockMsg, "[%s] Attempting connection to %s",
                 pStreamName.c_str(), nameBuff );
 
@@ -146,6 +188,11 @@ namespace XrdCl
 
     pPoller->RemoveSocket( pSocket );
     pSocket->Close();
+
+    if( !pIncHandler.second )
+      delete pIncoming;
+
+    pIncoming = 0;
     return Status();
   }
 
@@ -172,6 +219,7 @@ namespace XrdCl
     //--------------------------------------------------------------------------
     if( type & ReadyToRead )
     {
+      pLastActivity = time(0);
       if( likely( pHandShakeDone ) )
         OnRead();
       else
@@ -194,6 +242,7 @@ namespace XrdCl
     //--------------------------------------------------------------------------
     if( type & ReadyToWrite )
     {
+      pLastActivity = time(0);
       if( unlikely( pSocket->GetStatus() == Socket::Connecting ) )
         OnConnectionReturn();
       else if( likely( pHandShakeDone ) )
@@ -264,7 +313,7 @@ namespace XrdCl
     pHandShakeData = new HandShakeData( pStream->GetURL(),
                                         pStream->GetStreamNumber(),
                                         pSubStreamNum );
-    pHandShakeData->serverAddr = pSocket->GetServerAddress();
+    pHandShakeData->serverAddr = &pSocket->GetServerAddress();
     pHandShakeData->clientName = pSocket->GetSockName();
     pHandShakeData->streamName = pStreamName;
 
@@ -340,10 +389,9 @@ namespace XrdCl
 
       if( pOutHandler && pOutHandler->IsRaw() )
       {
-        log->Dump( AsyncSockMsg,
-                   "[%s] Will call raw handler to write payload for message: "
-                   "%s.", pStreamName.c_str(),
-                   pOutgoing->GetDescription().c_str() );
+        log->Dump( AsyncSockMsg, "[%s] Will call raw handler to write payload "
+                   "for message: %s (0x%x).", pStreamName.c_str(),
+                   pOutgoing->GetDescription().c_str(), pOutgoing );
       }
 
       pOutMsgDone = true;
@@ -368,9 +416,9 @@ namespace XrdCl
     }
 
     Log *log = DefaultEnv::GetLog();
-    log->Dump( AsyncSockMsg,
-               "[%s] successfully sent message: %s.", pStreamName.c_str(),
-               pOutgoing->GetDescription().c_str() );
+    log->Dump( AsyncSockMsg, "[%s] Successfully sent message: %s (0x%x).",
+               pStreamName.c_str(), pOutgoing->GetDescription().c_str(),
+               pOutgoing );
 
     pStream->OnMessageSent( pSubStreamNum, pOutgoing, pOutMsgSize );
     pOutgoing = 0;
@@ -443,9 +491,9 @@ namespace XrdCl
     //--------------------------------------------------------------------------
     // We have written the message successfully
     //--------------------------------------------------------------------------
-    log->Dump( AsyncSockMsg, "[%s] Wrote a message: %s, %d bytes",
+    log->Dump( AsyncSockMsg, "[%s] Wrote a message: %s (0x%x), %d bytes",
                pStreamName.c_str(), pOutgoing->GetDescription().c_str(),
-               pOutgoing->GetSize() );
+               pOutgoing, pOutgoing->GetSize() );
     return Status();
   }
 
@@ -484,9 +532,8 @@ namespace XrdCl
       if( st.code == suRetry )
         return;
 
-      log->Dump( AsyncSockMsg,
-                "[%s] Received message header, size: %d",
-                pStreamName.c_str(), pIncoming->GetCursor() );
+      log->Dump( AsyncSockMsg, "[%s] Received message header for 0x%x size: %d",
+                pStreamName.c_str(), pIncoming, pIncoming->GetCursor() );
       pIncMsgSize = pIncoming->GetCursor();
       pHeaderDone = true;
       std::pair<IncomingMsgHandler *, bool> raw;
@@ -494,9 +541,8 @@ namespace XrdCl
 
       if( pIncHandler.first )
       {
-        log->Dump( AsyncSockMsg,
-                   "[%s] Will use the raw handler to read message body.",
-                   pStreamName.c_str() );
+        log->Dump( AsyncSockMsg, "[%s] Will use the raw handler to read body "
+                   "of message 0x%x", pStreamName.c_str(), pIncoming );
       }
     }
 
@@ -507,7 +553,7 @@ namespace XrdCl
     {
       uint32_t bytesRead = 0;
       st = pIncHandler.first->ReadMessageBody( pIncoming, pSocket->GetFD(),
-                                                  bytesRead );
+                                               bytesRead );
       if( !st.IsOK() )
       {
         OnFault( st );
@@ -539,8 +585,8 @@ namespace XrdCl
     //--------------------------------------------------------------------------
     // Report the incoming message
     //--------------------------------------------------------------------------
-    log->Dump( AsyncSockMsg, "[%s] Received a message of %d bytes",
-               pStreamName.c_str(), pIncMsgSize );
+    log->Dump( AsyncSockMsg, "[%s] Received message 0x%x of %d bytes",
+               pStreamName.c_str(), pIncoming, pIncMsgSize );
 
     pStream->OnIncoming( pSubStreamNum, pIncoming, pIncMsgSize );
     pIncoming = 0;
