@@ -1,7 +1,9 @@
 //------------------------------------------------------------------------------
-// Copyright (c) 2011-2012 by European Organization for Nuclear Research (CERN)
+// Copyright (c) 2011-2014 by European Organization for Nuclear Research (CERN)
 // Author: Lukasz Janyst <ljanyst@cern.ch>
 //------------------------------------------------------------------------------
+// This file is part of the XRootD software suite.
+//
 // XRootD is free software: you can redistribute it and/or modify
 // it under the terms of the GNU Lesser General Public License as published by
 // the Free Software Foundation, either version 3 of the License, or
@@ -14,6 +16,10 @@
 //
 // You should have received a copy of the GNU Lesser General Public License
 // along with XRootD.  If not, see <http://www.gnu.org/licenses/>.
+//
+// In applying this licence, CERN does not waive the privileges and immunities
+// granted to it by virtue of its status as an Intergovernmental Organization
+// or submit itself to any jurisdiction.
 //------------------------------------------------------------------------------
 
 #include "XrdCl/XrdClFileStateHandler.hh"
@@ -181,7 +187,6 @@ namespace
                                             XrdCl::HostList     *hostList )
       {
         using namespace XrdCl;
-        XRDCL_SMART_PTR_T<XRootDStatus>    statusPtr( status );
         XRDCL_SMART_PTR_T<AnyObject>       responsePtr( response );
         pSendParams.hostList = hostList;
 
@@ -190,27 +195,13 @@ namespace
         //----------------------------------------------------------------------
         if( !status->IsOK() )
         {
-          statusPtr.release();
           pStateHandler->OnStateError( status, pMessage, this, pSendParams );
           return;
         }
 
-        //----------------------------------------------------------------------
-        // We have been sent out elsewhere
-        //----------------------------------------------------------------------
-        if( status->IsOK() && status->code == suXRDRedirect )
-        {
-          RedirectInfo *redirInfo = 0;
-          response->Get( redirInfo );
-          pStateHandler->OnStateRedirection( redirInfo, pMessage, this,
-                                             pSendParams );
-          return;
-        }
-
         //----------------------------------------------------------------------
         // We're clear
         //----------------------------------------------------------------------
-        statusPtr.release();
         responsePtr.release();
         pStateHandler->OnStateResponse( status, pMessage, response, hostList );
         pUserHandler->HandleResponseWithHosts( status, response, hostList );
@@ -250,7 +241,8 @@ namespace XrdCl
     pOpenFlags( 0 ),
     pSessionId( 0 ),
     pDoRecoverRead( true ),
-    pDoRecoverWrite( true )
+    pDoRecoverWrite( true ),
+    pFollowRedirects( true )
   {
     pFileHandle = new uint8_t[4];
     ResetMonitoringVars();
@@ -313,6 +305,13 @@ namespace XrdCl
     // Check if the parameters are valid
     //--------------------------------------------------------------------------
     Log *log = DefaultEnv::GetLog();
+
+    if (pFileUrl)
+    {
+      delete pFileUrl;
+      pFileUrl = 0;
+    }
+
     pFileUrl = new URL( url );
     if( !pFileUrl->IsValid() )
     {
@@ -369,6 +368,7 @@ namespace XrdCl
     XRootDTransport::SetDescription( msg );
     OpenHandler *openHandler = new OpenHandler( this, handler );
     MessageSendParams params; params.timeout = timeout;
+    params.followRedirects = pFollowRedirects;
     MessageUtils::ProcessSendParams( params );
 
     Status st = MessageUtils::SendMessage( *pFileUrl, msg, openHandler, params );
@@ -424,7 +424,10 @@ namespace XrdCl
     XRootDTransport::SetDescription( msg );
     msg->SetSessionId( pSessionId );
     CloseHandler *closeHandler = new CloseHandler( this, handler, msg );
-    MessageSendParams params; params.timeout = timeout;
+    MessageSendParams params;
+    params.timeout = timeout;
+    params.followRedirects = false;
+    params.stateful        = true;
     MessageUtils::ProcessSendParams( params );
 
     Status st = MessageUtils::SendMessage( *pDataServer, msg, closeHandler, params );
@@ -728,6 +731,80 @@ namespace XrdCl
     return SendOrQueue( *pDataServer, msg, stHandler, params );
   }
 
+  //----------------------------------------------------------------------------
+  // Performs a custom operation on an open file, server implementation
+  // dependent - async
+  //----------------------------------------------------------------------------
+  XRootDStatus FileStateHandler::Fcntl( const Buffer    &arg,
+                                        ResponseHandler *handler,
+                                        uint16_t         timeout )
+  {
+    XrdSysMutexHelper scopedLock( pMutex );
+
+    if( pFileState != Opened && pFileState != Recovering )
+      return XRootDStatus( stError, errInvalidOp );
+
+    Log *log = DefaultEnv::GetLog();
+    log->Debug( FileMsg, "[0x%x@%s] Sending a fcntl command for handle 0x%x to "
+                "%s", this, pFileUrl->GetURL().c_str(),
+                *((uint32_t*)pFileHandle), pDataServer->GetHostId().c_str() );
+
+    Message            *msg;
+    ClientQueryRequest *req;
+    MessageUtils::CreateRequest( msg, req, arg.GetSize() );
+
+    req->requestid = kXR_query;
+    req->infotype  = kXR_Qopaqug;
+    req->dlen      = arg.GetSize();
+    memcpy( req->fhandle, pFileHandle, 4 );
+    msg->Append( arg.GetBuffer(), arg.GetSize(), 24 );
+
+    MessageSendParams params;
+    params.timeout         = timeout;
+    params.followRedirects = false;
+    params.stateful        = true;
+    MessageUtils::ProcessSendParams( params );
+
+    XRootDTransport::SetDescription( msg );
+    StatefulHandler *stHandler = new StatefulHandler( this, handler, msg, params );
+    return SendOrQueue( *pDataServer, msg, stHandler, params );
+  }
+
+  //----------------------------------------------------------------------------
+  // Get access token to a file - async
+  //----------------------------------------------------------------------------
+  XRootDStatus FileStateHandler::Visa( ResponseHandler *handler,
+                                       uint16_t         timeout )
+  {
+    XrdSysMutexHelper scopedLock( pMutex );
+
+    if( pFileState != Opened && pFileState != Recovering )
+      return XRootDStatus( stError, errInvalidOp );
+
+    Log *log = DefaultEnv::GetLog();
+    log->Debug( FileMsg, "[0x%x@%s] Sending a visa command for handle 0x%x to "
+                "%s", this, pFileUrl->GetURL().c_str(),
+                *((uint32_t*)pFileHandle), pDataServer->GetHostId().c_str() );
+
+    Message            *msg;
+    ClientQueryRequest *req;
+    MessageUtils::CreateRequest( msg, req );
+
+    req->requestid = kXR_query;
+    req->infotype  = kXR_Qvisa;
+    memcpy( req->fhandle, pFileHandle, 4 );
+
+    MessageSendParams params;
+    params.timeout         = timeout;
+    params.followRedirects = false;
+    params.stateful        = true;
+    MessageUtils::ProcessSendParams( params );
+
+    XRootDTransport::SetDescription( msg );
+    StatefulHandler *stHandler = new StatefulHandler( this, handler, msg, params );
+    return SendOrQueue( *pDataServer, msg, stHandler, params );
+  }
+
   //----------------------------------------------------------------------------
   // Check if the file is open
   //----------------------------------------------------------------------------
@@ -741,45 +818,64 @@ namespace XrdCl
   }
 
   //----------------------------------------------------------------------------
-  // Enable/disable state recovery procedures while the file is open for
-  // reading
+  // Set file property
   //----------------------------------------------------------------------------
-  void FileStateHandler::EnableReadRecovery( bool enable )
+  bool FileStateHandler::SetProperty( const std::string &name,
+                                      const std::string &value )
   {
     XrdSysMutexHelper scopedLock( pMutex );
-    pDoRecoverRead = enable;
+    if( name == "ReadRecovery" )
+    {
+      if( value == "true" ) pDoRecoverRead = true;
+      else pDoRecoverRead = false;
+      return true;
+    }
+    else if( name == "WriteRecovery" )
+    {
+      if( value == "true" ) pDoRecoverWrite = true;
+      else pDoRecoverWrite = false;
+      return true;
+    }
+    else if( name == "FollowRedirects" )
+    {
+      if( value == "true" ) pFollowRedirects = true;
+      else pFollowRedirects = false;
+      return true;
+    }
+    return false;
   }
 
   //----------------------------------------------------------------------------
-  //! Enable/disable state recovery procedures while the file is open for
-  //! writing or read/write
+  // Get file property
   //----------------------------------------------------------------------------
-  void FileStateHandler::EnableWriteRecovery( bool enable )
+  bool FileStateHandler::GetProperty( const std::string &name,
+                                      std::string &value ) const
   {
     XrdSysMutexHelper scopedLock( pMutex );
-    pDoRecoverWrite = enable;
-  }
-
-  //----------------------------------------------------------------------------
-  // Get the data server the file is accessed at
-  //----------------------------------------------------------------------------
-  std::string FileStateHandler::GetDataServer() const
-  {
-    XrdSysMutexHelper scopedLock( pMutex );
-    if( pDataServer )
-      return pDataServer->GetHostId();
-    return "";
-  }
-
-  //----------------------------------------------------------------------------
-  // Get the final url with all the cgi information
-  //----------------------------------------------------------------------------
-  URL FileStateHandler::GetLastURL() const
-  {
-    XrdSysMutexHelper scopedLock( pMutex );
-    if( pDataServer )
-      return *pDataServer;
-    return URL();
+    if( name == "ReadRecovery" )
+    {
+      if( pDoRecoverRead ) value = "true";
+      else value = "false";
+      return true;
+    }
+    else if( name == "WriteRecovery" )
+    {
+      if( pDoRecoverWrite ) value = "true";
+      else value = "false";
+      return true;
+    }
+    else if( name == "FollowRedirects" )
+    {
+      if( pFollowRedirects ) value = "true";
+      else value = "false";
+      return true;
+    }
+    else if( name == "DataServer" && pDataServer )
+      { value = pDataServer->GetHostId(); return true; }
+    else if( name == "LastURL" && pDataServer )
+      { value =  pDataServer->GetURL(); return true; }
+    value = "";
+    return false;
   }
 
   //----------------------------------------------------------------------------
@@ -832,7 +928,7 @@ namespace XrdCl
     // We have failed
     //--------------------------------------------------------------------------
     pStatus = *status;
-    if( !pStatus.IsOK() )
+    if( !pStatus.IsOK() || !openInfo )
     {
       log->Debug( FileMsg, "[0x%x@%s] Error while opening at %s: %s",
                   this, pFileUrl->GetURL().c_str(), lastServer.c_str(),
@@ -928,6 +1024,24 @@ namespace XrdCl
                                        ResponseHandler   *userHandler,
                                        MessageSendParams &sendParams )
   {
+    //--------------------------------------------------------------------------
+    // It may be a redirection
+    //--------------------------------------------------------------------------
+    if( !status->IsOK() && status->code == errRedirect && pFollowRedirects )
+    {
+      std::string root = "root", xroot = "xroot";
+      std::string msg = status->GetErrorMessage();
+      if( !msg.compare( 0, root.size(), root ) ||
+          !msg.compare( 0, xroot.size(), xroot ) )
+      {
+        OnStateRedirection( msg, message, userHandler, sendParams );
+        return;
+      }
+    }
+
+    //--------------------------------------------------------------------------
+    // Handle error
+    //--------------------------------------------------------------------------
     Log *log = DefaultEnv::GetLog();
     XrdSysMutexHelper scopedLock( pMutex );
     pInTheFly.erase( message );
@@ -984,7 +1098,7 @@ namespace XrdCl
   //----------------------------------------------------------------------------
   // Handle stateful redirect
   //----------------------------------------------------------------------------
-  void FileStateHandler::OnStateRedirection( RedirectInfo      *redirectInfo,
+  void FileStateHandler::OnStateRedirection( const std::string &redirectUrl,
                                              Message           *message,
                                              ResponseHandler   *userHandler,
                                              MessageSendParams &sendParams )
@@ -999,9 +1113,7 @@ namespace XrdCl
     if( !pStateRedirect )
     {
       std::ostringstream o;
-      o << redirectInfo->host << ":" << redirectInfo->port << "//fakepath?";
-      o << redirectInfo->cgi;
-      pStateRedirect = new URL( o.str() );
+      pStateRedirect = new URL( redirectUrl );
       URL::ParamsMap params = pFileUrl->GetParams();
       MessageUtils::MergeCGI( params,
                               pStateRedirect->GetParams(),
@@ -1015,7 +1127,7 @@ namespace XrdCl
   //----------------------------------------------------------------------------
   // Handle stateful response
   //----------------------------------------------------------------------------
-  void FileStateHandler::OnStateResponse( XRootDStatus */*status*/,
+  void FileStateHandler::OnStateResponse( XRootDStatus *status,
                                           Message      *message,
                                           AnyObject    *response,
                                           HostList     */*urlList*/ )
@@ -1094,8 +1206,8 @@ namespace XrdCl
   //------------------------------------------------------------------------
   void FileStateHandler::Tick( time_t now )
   {
-    TimeOutRequests( now );
     XrdSysMutexHelper scopedLock( pMutex );
+    TimeOutRequests( now );
   }
 
   //----------------------------------------------------------------------------
@@ -1152,7 +1264,7 @@ namespace XrdCl
   //----------------------------------------------------------------------------
   // Send a message to a host or put it in the recovery queue
   //----------------------------------------------------------------------------
-  Status FileStateHandler::SendOrQueue( const URL         &/*url*/,
+  Status FileStateHandler::SendOrQueue( const URL         &url,
                                         Message           *msg,
                                         ResponseHandler   *handler,
                                         MessageSendParams &sendParams )
@@ -1262,13 +1374,14 @@ namespace XrdCl
     Status st;
     if( pStateRedirect )
     {
-      st = ReOpenFileAtServer( *pStateRedirect, 300 );
+      SendClose( 0 );
+      st = ReOpenFileAtServer( *pStateRedirect, 0 );
       delete pStateRedirect; pStateRedirect = 0;
     }
     else if( IsReadOnly() && pLoadBalancer )
-      st = ReOpenFileAtServer( *pLoadBalancer, 300 );
+      st = ReOpenFileAtServer( *pLoadBalancer, 0 );
     else
-      st = ReOpenFileAtServer( *pDataServer, 300 );
+      st = ReOpenFileAtServer( *pDataServer, 0 );
 
     if( !st.IsOK() )
     {
@@ -1279,6 +1392,31 @@ namespace XrdCl
     return st;
   }
 
+  //----------------------------------------------------------------------------
+  // Send a close and ignore the response
+  //----------------------------------------------------------------------------
+  Status FileStateHandler::SendClose( uint16_t timeout )
+  {
+    Message            *msg;
+    ClientCloseRequest *req;
+    MessageUtils::CreateRequest( msg, req );
+
+    req->requestid = kXR_close;
+    memcpy( req->fhandle, pFileHandle, 4 );
+
+    XRootDTransport::SetDescription( msg );
+    msg->SetSessionId( pSessionId );
+    NullResponseHandler *handler = new NullResponseHandler();
+    MessageSendParams params;
+    params.timeout         = timeout;
+    params.followRedirects = false;
+    params.stateful        = true;
+
+    MessageUtils::ProcessSendParams( params );
+
+    return MessageUtils::SendMessage( *pDataServer, msg, handler, params );
+  }
+
   //----------------------------------------------------------------------------
   // Re-open the current file at a given server
   //----------------------------------------------------------------------------
@@ -1286,7 +1424,7 @@ namespace XrdCl
   {
     Log *log = DefaultEnv::GetLog();
     log->Dump( FileMsg, "[0x%x@%s] Sending a recovery open command to %s",
-               this, pFileUrl->GetURL().c_str(), url.GetHostId().c_str() );
+               this, pFileUrl->GetURL().c_str(), url.GetURL().c_str() );
 
     //--------------------------------------------------------------------------
     // Remove the kXR_delete and kXR_new flags, we don't want the recovery
@@ -1298,7 +1436,11 @@ namespace XrdCl
 
     Message           *msg;
     ClientOpenRequest *req;
-    std::string        path = pFileUrl->GetPathWithParams();
+
+    URL u = *pFileUrl;
+    if( !url.GetPath().empty() )
+      u.SetPath( url.GetPath() );
+    std::string        path = u.GetPathWithParams();
     MessageUtils::CreateRequest( msg, req, path.length() );
 
     req->requestid = kXR_open;
