@@ -1,7 +1,9 @@
 //------------------------------------------------------------------------------
-// Copyright (c) 2011-2012 by European Organization for Nuclear Research (CERN)
+// Copyright (c) 2011-2014 by European Organization for Nuclear Research (CERN)
 // Author: Lukasz Janyst <ljanyst@cern.ch>
 //------------------------------------------------------------------------------
+// This file is part of the XRootD software suite.
+//
 // XRootD is free software: you can redistribute it and/or modify
 // it under the terms of the GNU Lesser General Public License as published by
 // the Free Software Foundation, either version 3 of the License, or
@@ -14,6 +16,10 @@
 //
 // You should have received a copy of the GNU Lesser General Public License
 // along with XRootD.  If not, see <http://www.gnu.org/licenses/>.
+//
+// In applying this licence, CERN does not waive the privileges and immunities
+// granted to it by virtue of its status as an Intergovernmental Organization
+// or submit itself to any jurisdiction.
 //------------------------------------------------------------------------------
 
 #include "XrdCl/XrdClPollerBuiltIn.hh"
@@ -57,8 +63,6 @@ namespace
                           void                      *cbArg,
                           int                        evFlags )
       {
-        (void)chP;
-        (void)cbArg;
         using namespace XrdCl;
         uint8_t ev      = 0;
 
@@ -107,7 +111,7 @@ namespace XrdCl
     for( it = pSocketMap.begin(); it != pSocketMap.end(); ++it )
     {
       PollerHelper *helper = (PollerHelper*)it->second;
-      delete helper->channel;
+             helper->channel->Delete();
       delete helper->callBack;
       delete helper;
     }
@@ -188,7 +192,21 @@ namespace XrdCl
 
     Log *log = DefaultEnv::GetLog();
     log->Debug( PollerMsg, "Stopping the poller..." );
+
     XrdSysMutexHelper scopedLock( pMutex );
+    if( !pPoller )
+    {
+      log->Debug( PollerMsg, "Stopping a poller that has not been started" );
+      return true;
+    }
+
+    XrdSys::IOEvents::Poller *poller = pPoller;
+    pPoller = 0;
+
+    scopedLock.UnLock();
+    poller->Stop();
+    delete poller;
+    scopedLock.Lock( &pMutex );
 
     SocketMap::iterator  it;
     const char          *errMsg = 0;
@@ -203,20 +221,10 @@ namespace XrdCl
         log->Error( PollerMsg, "%s Unable to disable write notifications: %s",
                     socket->GetName().c_str(), errMsg );
       }
-      delete helper->channel;
+      helper->channel->Delete();
       helper->channel = 0;
     }
 
-    if( !pPoller )
-    {
-      log->Debug( PollerMsg, "Stopping a poller that has not been started" );
-      return true;
-    }
-
-    pPoller->Stop();
-    delete pPoller;
-    pPoller = 0;
-
     return true;
   }
 
@@ -306,7 +314,7 @@ namespace XrdCl
                     socket->GetName().c_str(), errMsg );
         return false;
       }
-      delete helper->channel;
+      helper->channel->Delete();
     }
     delete helper->callBack;
     delete helper;
