@@ -1,7 +1,9 @@
 //------------------------------------------------------------------------------
-// Copyright (c) 2011-2012 by European Organization for Nuclear Research (CERN)
+// Copyright (c) 2011-2014 by European Organization for Nuclear Research (CERN)
 // Author: Lukasz Janyst <ljanyst@cern.ch>
 //------------------------------------------------------------------------------
+// This file is part of the XRootD software suite.
+//
 // XRootD is free software: you can redistribute it and/or modify
 // it under the terms of the GNU Lesser General Public License as published by
 // the Free Software Foundation, either version 3 of the License, or
@@ -14,6 +16,10 @@
 //
 // You should have received a copy of the GNU Lesser General Public License
 // along with XRootD.  If not, see <http://www.gnu.org/licenses/>.
+//
+// In applying this licence, CERN does not waive the privileges and immunities
+// granted to it by virtue of its status as an Intergovernmental Organization
+// or submit itself to any jurisdiction.
 //------------------------------------------------------------------------------
 
 #include "XrdCl/XrdClStream.hh"
@@ -27,9 +33,9 @@
 #include "XrdCl/XrdClOutQueue.hh"
 #include "XrdCl/XrdClMonitor.hh"
 #include "XrdCl/XrdClAsyncSocketHandler.hh"
-#include "XrdSys/XrdSysDNS.hh"
 
 #include <sys/types.h>
+#include <algorithm>
 #include <sys/socket.h>
 #include <sys/time.h>
 
@@ -111,6 +117,7 @@ namespace XrdCl
     pLastStreamError( 0 ),
     pConnectionCount( 0 ),
     pConnectionInitTime( 0 ),
+    pAddressType( Utils::IPAll ),
     pSessionId( 0 ),
     pQueueIncMsgJob(0),
     pBytesSent( 0 ),
@@ -130,11 +137,16 @@ namespace XrdCl
     pStreamErrorWindow = Utils::GetIntParameter( *url, "StreamErrorWindow",
                                                  DefaultStreamErrorWindow );
 
+    std::string netStack = Utils::GetStringParameter( *url, "NetworkStack",
+                                                      DefaultNetworkStack );
+
+    pAddressType = Utils::String2AddressType( netStack );
+
     Log *log = DefaultEnv::GetLog();
-    log->Debug( PostMasterMsg, "[%s] Stream parameters: Connection Window: %d, "
-                "ConnectionRetry: %d, Stream Error Widnow: %d",
-                pStreamName.c_str(), pConnectionWindow, pConnectionRetry,
-                pStreamErrorWindow );
+    log->Debug( PostMasterMsg, "[%s] Stream parameters: Network Stack: %s, "
+                "Connection Window: %d, ConnectionRetry: %d, Stream Error "
+                "Widnow: %d", pStreamName.c_str(), netStack.c_str(),
+                pConnectionWindow, pConnectionRetry, pStreamErrorWindow );
   }
 
   //----------------------------------------------------------------------------
@@ -170,6 +182,7 @@ namespace XrdCl
                                                     pChannelData,
                                                     0 );
     s->SetStream( this );
+
     pSubStreams.push_back( new SubStreamData() );
     pSubStreams[0]->socket = s;
     return Status();
@@ -231,7 +244,7 @@ namespace XrdCl
     //--------------------------------------------------------------------------
     // Resolve all the addresses of the host we're supposed to connect to
     //--------------------------------------------------------------------------
-    Status st = Utils::GetHostAddresses( pAddresses, *pUrl );
+    Status st = Utils::GetHostAddresses( pAddresses, *pUrl, pAddressType );
     if( !st.IsOK() )
     {
       log->Error( PostMasterMsg, "[%s] Unable to resolve IP address for "
@@ -246,12 +259,13 @@ namespace XrdCl
                              pAddresses );
 
     //--------------------------------------------------------------------------
-    // Initiate the connection process to the first one on the list
+    // Initiate the connection process to the first one on the list.
+    // It's more efficient to remove addresses from the back of a vector
+    // so we reverse the it.
     //--------------------------------------------------------------------------
-    sockaddr_in addr;
-    memcpy( &addr, &pAddresses.back(), sizeof( sockaddr_in ) );
+    std::reverse( pAddresses.begin(), pAddresses.end() );
+    pSubStreams[0]->socket->SetAddress( pAddresses.back() );
     pAddresses.pop_back();
-    pSubStreams[0]->socket->SetAddress( addr );
     st = pSubStreams[0]->socket->Connect( pConnectionWindow );
     if( st.IsOK() )
       pSubStreams[0]->status = Socket::Connecting;
@@ -280,7 +294,8 @@ namespace XrdCl
     //--------------------------------------------------------------------------
     // Decide on the path to send the message
     //--------------------------------------------------------------------------
-    PathID path = pTransport->MultiplexSubStream( msg, *pChannelData );
+    PathID path = pTransport->MultiplexSubStream( msg, pStreamNum,
+                                                  *pChannelData );
     if( pSubStreams.size() <= path.up )
     {
       log->Warning( PostMasterMsg, "[%s] Unable to send message %s through "
@@ -289,9 +304,9 @@ namespace XrdCl
       path.up = 0;
     }
 
-    log->Dump( PostMasterMsg, "[%s] Sending message %s through substream %d "
-               "expecting answer at %d", pStreamName.c_str(),
-               msg->GetDescription().c_str(), path.up, path.down );
+    log->Dump( PostMasterMsg, "[%s] Sending message %s (0x%x) through "
+               "substream %d expecting answer at %d", pStreamName.c_str(),
+               msg->GetDescription().c_str(), msg, path.up, path.down );
 
     //--------------------------------------------------------------------------
     // Enable *a* path and insert the message to the right queue
@@ -299,7 +314,7 @@ namespace XrdCl
     Status st = EnableLink( path );
     if( st.IsOK() )
     {
-      pTransport->MultiplexSubStream( msg, *pChannelData, &path );
+      pTransport->MultiplexSubStream( msg, pStreamNum, *pChannelData, &path );
       pSubStreams[path.up]->outQueue->PushBack( msg, handler,
                                                 expires, stateful );
     }
@@ -326,6 +341,13 @@ namespace XrdCl
   //----------------------------------------------------------------------------
   void Stream::Disconnect( bool /*force*/ )
   {
+    XrdSysMutexHelper scopedLock( pMutex );
+    SubStreamList::iterator it;
+    for( it = pSubStreams.begin(); it != pSubStreams.end(); ++it )
+    {
+      (*it)->socket->Close();
+      (*it)->status = Socket::Disconnected;
+    }
   }
 
   //----------------------------------------------------------------------------
@@ -333,9 +355,12 @@ namespace XrdCl
   //----------------------------------------------------------------------------
   void Stream::Tick( time_t now )
   {
+    //--------------------------------------------------------------------------
+    // Check for timed-out requests and incoming handlers
+    //--------------------------------------------------------------------------
     pMutex.Lock();
-    SubStreamList::iterator it;
     OutQueue q;
+    SubStreamList::iterator it;
     for( it = pSubStreams.begin(); it != pSubStreams.end(); ++it )
       q.GrabExpired( *(*it)->outQueue, now );
     pMutex.UnLock();
@@ -391,20 +416,23 @@ namespace XrdCl
     msg->SetSessionId( pSessionId );
     pBytesReceived += bytesReceived;
 
+    uint32_t streamAction = pTransport->MessageReceived( msg, pStreamNum,
+                                                         subStream,
+                                                         *pChannelData );
+    if( streamAction & TransportHandler::DigestMsg )
+      return;
+
     //--------------------------------------------------------------------------
     // No handler, we cache and see what comes later
     //--------------------------------------------------------------------------
     Log *log = DefaultEnv::GetLog();
     InMessageHelper &mh = pSubStreams[subStream]->inMsgHelper;
+
     if( !mh.handler )
     {
-      log->Dump( PostMasterMsg, "[%s] Queuing received message: %s.",
-                 pStreamName.c_str(), msg->GetDescription().c_str() );
+      log->Dump( PostMasterMsg, "[%s] Queuing received message: 0x%x.",
+                 pStreamName.c_str(), msg );
 
-      uint32_t streamAction = pTransport->StreamAction( msg, *pChannelData );
-
-      if( streamAction & TransportHandler::DigestMsg )
-        return;
       pJobManager->QueueJob( pQueueIncMsgJob, msg );
       return;
     }
@@ -412,17 +440,19 @@ namespace XrdCl
     //--------------------------------------------------------------------------
     // We have a handler, so we call the callback
     //--------------------------------------------------------------------------
-    log->Dump( PostMasterMsg, "[%s] Handling received message: %s.",
-               pStreamName.c_str(), msg->GetDescription().c_str() );
+    log->Dump( PostMasterMsg, "[%s] Handling received message: 0x%x.",
+               pStreamName.c_str(), msg );
 
     if( !(mh.action & IncomingMsgHandler::RemoveHandler) )
       pIncomingQueue->ReAddMessageHandler( mh.handler, mh.expires );
 
-    if( mh.action & IncomingMsgHandler::NoProcess )
+    if( mh.action & (IncomingMsgHandler::NoProcess|IncomingMsgHandler::Ignore) )
     {
-      log->Dump( PostMasterMsg, "[%s] Ignoring the processing handler for: %s.",
+      log->Dump( PostMasterMsg, "[%s] Ignoring the processing handler for: 0x%x.",
                  pStreamName.c_str(), msg->GetDescription().c_str() );
+      bool delit = ( mh.action & IncomingMsgHandler::Ignore );
       mh.Reset();
+      if (delit) delete msg;
       return;
     }
 
@@ -465,6 +495,8 @@ namespace XrdCl
                               Message  *msg,
                               uint32_t  bytesSent )
   {
+    pTransport->MessageSent( msg, pStreamNum, subStream, bytesSent,
+                             *pChannelData );
     OutMessageHelper &h = pSubStreams[subStream]->outMsgHelper;
     pBytesSent += bytesSent;
     if( h.handler )
@@ -601,10 +633,8 @@ namespace XrdCl
       //------------------------------------------------------------------------
       if( !pAddresses.empty() )
       {
-        sockaddr_in addr;
-        memcpy( &addr, &pAddresses.back(), sizeof( sockaddr_in ) );
+        pSubStreams[0]->socket->SetAddress( pAddresses.back() );
         pAddresses.pop_back();
-        pSubStreams[0]->socket->SetAddress( addr );
 
         Status st = pSubStreams[0]->socket->Connect( pConnectionWindow-elapsed );
         if( !st.IsOK() )
@@ -810,8 +840,59 @@ namespace XrdCl
   //----------------------------------------------------------------------------
   // Call back when a message has been reconstructed
   //----------------------------------------------------------------------------
-  void Stream::OnReadTimeout( uint16_t /*substream*/ )
+  void Stream::OnReadTimeout( uint16_t substream )
   {
+    //--------------------------------------------------------------------------
+    // We only take the main stream into account
+    //--------------------------------------------------------------------------
+    if( substream != 0 )
+      return;
+
+    //--------------------------------------------------------------------------
+    // Check if there is no outgoing messages and if the stream TTL is elapesed.
+    // It is assumed that the underlying transport makes sure that there is no
+    // pending requests that are not answered, ie. all possible virtual streams
+    // are de-allocated
+    //--------------------------------------------------------------------------
+    Log *log = DefaultEnv::GetLog();
+    SubStreamList::iterator it;
+    time_t                  now = time(0);
+
+    XrdSysMutexHelper scopedLock( pMutex );
+    uint32_t outgoingMessages = 0;
+    time_t   lastActivity     = 0;
+    for( it = pSubStreams.begin(); it != pSubStreams.end(); ++it )
+    {
+      outgoingMessages += (*it)->outQueue->GetSize();
+      time_t sockLastActivity = (*it)->socket->GetLastActivity();
+      if( lastActivity < sockLastActivity )
+        lastActivity = sockLastActivity;
+    }
+
+    if( !outgoingMessages )
+    {
+      bool disconnect = pTransport->IsStreamTTLElapsed( now-lastActivity,
+                                                        pStreamNum,
+                                                        *pChannelData );
+      if( disconnect )
+      {
+        log->Debug( PostMasterMsg, "[%s] Stream TTL elapsed, disconnecting...",
+                    pStreamName.c_str() );
+        Disconnect();
+      }
+    }
+
+    //--------------------------------------------------------------------------
+    // Check if the stream is broken
+    //--------------------------------------------------------------------------
+    Status st = pTransport->IsStreamBroken( now-lastActivity,
+                                            pStreamNum,
+                                            *pChannelData );
+    if( !st.IsOK() )
+    {
+      scopedLock.UnLock();
+      OnError( substream, st );
+    }
   }
 
   //----------------------------------------------------------------------------
