@@ -70,7 +70,7 @@ namespace XrdCl
     if( pos != std::string::npos )
     {
       pProtocol = url.substr( 0, pos );
-      current   = url.substr( pos+3, url.length()-pos-3 );
+      current   = url.substr( pos+3 );
     }
     else if( url[0] == '/' )
     {
@@ -105,7 +105,7 @@ namespace XrdCl
       else
       {
         hostInfo = current.substr( 0, pos );
-        path     = current.substr( pos+1, current.length()-pos );
+        path     = current.substr( pos+1 );
       }
     }
 
@@ -121,7 +121,7 @@ namespace XrdCl
       return false;
     }
 
-    pURL = url;
+    ComputeURL();
 
     //--------------------------------------------------------------------------
     // Dump the url
@@ -159,7 +159,7 @@ namespace XrdCl
     if( pos != std::string::npos )
     {
       std::string userPass = hostInfo.substr( 0, pos );
-      hostPort = hostInfo.substr( pos+1, hostInfo.length() );
+      hostPort = hostInfo.substr( pos+1 );
       pos = userPass.find( ":" );
 
       //------------------------------------------------------------------------
@@ -168,7 +168,7 @@ namespace XrdCl
       if( pos != std::string::npos )
       {
         pUserName = userPass.substr( 0, pos );
-        pPassword = userPass.substr( pos+1, userPass.length() );
+        pPassword = userPass.substr( pos+1 );
         if( pPassword.empty() )
           return false;
       }
@@ -202,7 +202,10 @@ namespace XrdCl
         // Check if we're IPv6 encoded IPv4
         //----------------------------------------------------------------------
         pos = pHostName.find( "." );
-        if( pos != std::string::npos )
+        size_t pos2 = pHostName.find( "[::ffff" );
+        size_t pos3 = pHostName.find( "[::" );
+        if( pos != std::string::npos && pos3 != std::string::npos &&
+            pos2 == std::string::npos )
         {
           pHostName.erase( 0, 3 );
           pHostName.erase( pHostName.length()-1, 1 );
