@@ -1,7 +1,9 @@
 //------------------------------------------------------------------------------
-// Copyright (c) 2011-2012 by European Organization for Nuclear Research (CERN)
+// Copyright (c) 2011-2014 by European Organization for Nuclear Research (CERN)
 // Author: Lukasz Janyst <ljanyst@cern.ch>
 //------------------------------------------------------------------------------
+// This file is part of the XRootD software suite.
+//
 // XRootD is free software: you can redistribute it and/or modify
 // it under the terms of the GNU Lesser General Public License as published by
 // the Free Software Foundation, either version 3 of the License, or
@@ -14,6 +16,10 @@
 //
 // You should have received a copy of the GNU Lesser General Public License
 // along with XRootD.  If not, see <http://www.gnu.org/licenses/>.
+//
+// In applying this licence, CERN does not waive the privileges and immunities
+// granted to it by virtue of its status as an Intergovernmental Organization
+// or submit itself to any jurisdiction.
 //------------------------------------------------------------------------------
 
 #include "XrdCl/XrdClXRootDMsgHandler.hh"
@@ -23,11 +29,11 @@
 #include "XrdCl/XrdClXRootDTransport.hh"
 #include "XrdCl/XrdClMessage.hh"
 #include "XrdCl/XrdClURL.hh"
+#include "XrdCl/XrdClUglyHacks.hh"
 #include "XrdCl/XrdClUtils.hh"
 #include "XrdCl/XrdClTaskManager.hh"
 #include "XrdCl/XrdClSIDManager.hh"
 #include "XrdCl/XrdClMessageUtils.hh"
-#include "XrdCl/XrdClUglyHacks.hh"
 
 #include <arpa/inet.h>              // for network unmarshalling stuff
 #include "XrdSys/XrdSysPlatform.hh" // same as above
@@ -140,7 +146,7 @@ namespace XrdCl
         return Take | RemoveHandler;
 
       case kXR_waitresp:
-        return Take;
+        return Take | Ignore; // This must be handled synchronously!
 
       //------------------------------------------------------------------------
       // Handle the potential raw cases
@@ -255,11 +261,29 @@ namespace XrdCl
                  pRequest->GetDescription().c_str() );
       Message *embededMsg = new Message( rsp->hdr.dlen-8 );
       embededMsg->Append( msg->GetBuffer( 16 ), rsp->hdr.dlen-8 );
-      delete msg;
+      XRDCL_SMART_PTR_T<Message> msgPtr( msg );
       pResponse = embededMsg; // this can never happen for oksofars
 
       // we need to unmarshall the header by hand
       XRootDTransport::UnMarshallHeader( embededMsg );
+
+      //------------------------------------------------------------------------
+      // Check if the dlen field of the embedded message is consistent with
+      // the dlen value of the original message
+      //------------------------------------------------------------------------
+      ServerResponse *embRsp = (ServerResponse *)embededMsg->GetBuffer();
+      if( embRsp->hdr.dlen != rsp->hdr.dlen-16 )
+      {
+        log->Error( XRootDMsg, "[%s] Sizes of the async response to %s and the "
+                    "embedded message are inconsistent. Expected %d, got %d.",
+                    pUrl.GetHostId().c_str(), pRequest->GetDescription().c_str(),
+                    rsp->hdr.dlen-16, embRsp->hdr.dlen);
+
+        pStatus = Status( stFatal, errInvalidMessage );
+        HandleResponse();
+        return;
+      }
+
       Process( embededMsg );
       return;
     }
@@ -279,7 +303,14 @@ namespace XrdCl
     //--------------------------------------------------------------------------
     // Process the message
     //--------------------------------------------------------------------------
-    XRootDTransport::UnMarshallBody( msg, req->header.requestid );
+    Status st = XRootDTransport::UnMarshallBody( msg, req->header.requestid );
+    if( !st.IsOK() )
+    {
+      pStatus = Status( stFatal, errInvalidMessage );
+      HandleResponse();
+      return;
+    }
+
     switch( rsp->hdr.status )
     {
       //------------------------------------------------------------------------
@@ -300,12 +331,16 @@ namespace XrdCl
       //------------------------------------------------------------------------
       case kXR_error:
       {
+        char *errmsg = new char[rsp->hdr.dlen-3]; errmsg[rsp->hdr.dlen-4] = 0;
+        memcpy( errmsg, rsp->body.error.errmsg, rsp->hdr.dlen-4 );
         log->Dump( XRootDMsg, "[%s] Got a kXR_error response to request %s "
                    "[%d] %s", pUrl.GetHostId().c_str(),
                    pRequest->GetDescription().c_str(), rsp->body.error.errnum,
-                   rsp->body.error.errmsg );
+                   errmsg );
+        delete [] errmsg;
 
-        HandleError( Status(stError, errErrorResponse), pResponse );
+        HandleError( Status(stError, errErrorResponse, rsp->body.error.errnum),
+                     pResponse );
         return;
       }
 
@@ -386,34 +421,52 @@ namespace XrdCl
         std::vector<std::string> urlComponents;
         std::string newCgi;
         Utils::splitString( urlComponents, urlInfo, "?" );
+
         std::ostringstream o;
-        o << urlComponents[0] << ":" << rsp->body.redirect.port << "/";
-        pUrl = URL( o.str() );
-        if( !pUrl.IsValid() )
+
+        o << urlComponents[0];
+        if( rsp->body.redirect.port != -1 )
+          o << ":" << rsp->body.redirect.port << "/";
+
+        URL newUrl = URL( o.str() );
+        if( !newUrl.IsValid() )
         {
           pStatus = Status( stError, errInvalidRedirectURL );
           log->Error( XRootDMsg, "[%s] Got invalid redirection URL: %s",
-                                pUrl.GetHostId().c_str(), urlInfo.c_str() );
+                      pUrl.GetHostId().c_str(), urlInfo.c_str() );
           HandleResponse();
           return;
         }
 
+        if( pUrl.GetUserName() != "" && newUrl.GetUserName() == "" )
+          newUrl.SetUserName( pUrl.GetUserName() );
+
+        if( pUrl.GetPassword() != "" && newUrl.GetPassword() == "" )
+          newUrl.SetPassword( pUrl.GetPassword() );
+
+        pUrl         = newUrl;
+        pRedirectUrl = newUrl.GetURL();
+
         URL cgiURL;
         if( urlComponents.size() > 1 )
         {
           std::ostringstream o;
           o << "fake://fake:111//fake?";
           o << urlComponents[1];
+          pRedirectUrl += "?";
+          pRedirectUrl += urlComponents[1];
           cgiURL = URL( o.str() );
-          pRedirectCgi = urlComponents[1];
         }
 
         //----------------------------------------------------------------------
         // Check if we need to return the URL as a response
         //----------------------------------------------------------------------
+        if( pUrl.GetProtocol() != "root" && pUrl.GetProtocol() != "xroot" )
+          pRedirectAsAnswer = true;
+
         if( pRedirectAsAnswer )
         {
-          pStatus   = Status( stOK, suXRDRedirect );
+          pStatus   = Status( stError, errRedirect );
           pResponse = msgPtr.release();
           HandleResponse();
           return;
@@ -422,7 +475,8 @@ namespace XrdCl
         //----------------------------------------------------------------------
         // Rewrite the message in a way required to send it to another server
         //----------------------------------------------------------------------
-        Status st = RewriteRequestRedirect( cgiURL.GetParams() );
+        Status st = RewriteRequestRedirect( cgiURL.GetParams(),
+                                            pUrl.GetPath() );
         if( !st.IsOK() )
         {
           pStatus = st;
@@ -468,7 +522,6 @@ namespace XrdCl
         }
 
         //----------------------------------------------------------------------
-
         // Some messages require rewriting before they can be sent again
         // after wait
         //----------------------------------------------------------------------
@@ -481,11 +534,24 @@ namespace XrdCl
         }
 
         //----------------------------------------------------------------------
-        // Register a task to resend the message in some seconds
+        // Register a task to resend the message in some seconds, if we still
+        // have time to do that, and report a timeout otherwise
         //----------------------------------------------------------------------
-        TaskManager *taskMgr = pPostMaster->GetTaskManager();
-        taskMgr->RegisterTask( new WaitTask( this ),
-                               ::time(0)+waitSeconds );
+        time_t resendTime = ::time(0)+waitSeconds;
+
+        if( resendTime < pExpiration )
+        {
+          TaskManager *taskMgr = pPostMaster->GetTaskManager();
+          taskMgr->RegisterTask( new WaitTask( this ), resendTime );
+        }
+        else
+        {
+          log->Debug( XRootDMsg, "[%s] Wait time is too long, timing out %s",
+                      pUrl.GetHostId().c_str(),
+                      pRequest->GetDescription().c_str() );
+          pStatus   = Status( stError, errOperationExpired );
+          HandleResponse();
+        }
         return;
       }
 
@@ -626,7 +692,6 @@ namespace XrdCl
                                          int       socket,
                                          uint32_t &bytesRead )
   {
-    (void)msg;
     if( pReadVRawMsgOffset == pAsyncMsgSize )
       return Status( stOK, suDone );
 
@@ -783,6 +848,12 @@ namespace XrdCl
       //------------------------------------------------------------------------
       if( !st.IsOK() )
         return st;
+
+      //------------------------------------------------------------------------
+      // If we are not done reading the header, return back to the event loop.
+      //------------------------------------------------------------------------
+      if( st.IsOK() && st.code != suDone )
+        return st;
     }
 
     //--------------------------------------------------------------------------
@@ -816,7 +887,6 @@ namespace XrdCl
                                          int       socket,
                                          uint32_t &bytesRead )
   {
-    (void)msg;
     if( !pOtherRawStarted )
     {
       pAsyncOffset     = 0;
@@ -1012,10 +1082,20 @@ namespace XrdCl
     ServerResponse *rsp = 0;
     if( pResponse )
       rsp = (ServerResponse *)pResponse->GetBuffer();
-    if( !pStatus.IsOK() && pStatus.code == errErrorResponse && rsp )
+
+    if( !pStatus.IsOK() && rsp )
     {
-      st->errNo = rsp->body.error.errnum;
-      st->SetErrorMessage( rsp->body.error.errmsg );
+      if( pStatus.code == errErrorResponse )
+      {
+        st->errNo = rsp->body.error.errnum;
+        char *errmsg = new char[rsp->hdr.dlen-3];
+        errmsg[rsp->hdr.dlen-4] = 0;
+        memcpy( errmsg, rsp->body.error.errmsg, rsp->hdr.dlen-4 );
+        st->SetErrorMessage( errmsg );
+        delete [] errmsg;
+      }
+      else if( pStatus.code == errRedirect )
+        st->SetErrorMessage( pRedirectUrl );
     }
     return st;
   }
@@ -1038,21 +1118,8 @@ namespace XrdCl
     //--------------------------------------------------------------------------
     if( rsp->hdr.status == kXR_redirect )
     {
-      if( !pRedirectAsAnswer )
-      {
-        log->Error( XRootDMsg, "Internal Error: trying to pass redirect as an "
-                    "answer even though this has never been requested" );
-        return 0;
-      }
-      log->Dump( XRootDMsg, "Parsing the response to %s as RedirectInfo",
-                 pRequest->GetDescription().c_str() );
-      AnyObject    *obj  = new AnyObject();
-      RedirectInfo *info = new RedirectInfo( pUrl.GetHostName(),
-                                             pUrl.GetPort(),
-                                             pRedirectCgi );
-      obj->Set( info );
-      response = obj;
-      return Status();
+      log->Error( XRootDMsg, "Internal Error: unable to process redirect" );
+      return 0;
     }
 
     //--------------------------------------------------------------------------
@@ -1125,10 +1192,25 @@ namespace XrdCl
       case kXR_locate:
       {
         AnyObject *obj = new AnyObject();
+
+        char *nullBuffer = new char[length+1];
+        nullBuffer[length] = 0;
+        memcpy( nullBuffer, buffer, length );
+
         log->Dump( XRootDMsg, "[%s] Parsing the response to %s as "
                    "LocateInfo: %s", pUrl.GetHostId().c_str(),
-                   pRequest->GetDescription().c_str(), buffer );
-        LocationInfo *data = new LocationInfo( buffer );
+                   pRequest->GetDescription().c_str(), nullBuffer );
+        LocationInfo *data = new LocationInfo();
+
+        if( data->ParseServerResponse( nullBuffer ) == false )
+        {
+          delete obj;
+          delete data;
+          delete [] nullBuffer;
+          return Status( stError, errInvalidResponse );
+        }
+        delete [] nullBuffer;
+
         obj->Set( data );
         response = obj;
         return Status();
@@ -1146,11 +1228,25 @@ namespace XrdCl
         //----------------------------------------------------------------------
         if( req->stat.options & kXR_vfs )
         {
+          StatInfoVFS *data = new StatInfoVFS();
+
+          char *nullBuffer = new char[length+1];
+          nullBuffer[length] = 0;
+          memcpy( nullBuffer, buffer, length );
+
           log->Dump( XRootDMsg, "[%s] Parsing the response to %s as "
-                     "StatInfoVFS", pUrl.GetHostId().c_str(),
-                     pRequest->GetDescription().c_str() );
+                     "StatInfoVFS: %s", pUrl.GetHostId().c_str(),
+                     pRequest->GetDescription().c_str(), nullBuffer );
+
+          if( data->ParseServerResponse( nullBuffer ) == false )
+          {
+              delete obj;
+              delete data;
+              delete [] nullBuffer;
+              return Status( stError, errInvalidResponse );
+          }
+          delete [] nullBuffer;
 
-          StatInfoVFS *data = new StatInfoVFS( buffer );
           obj->Set( data );
         }
         //----------------------------------------------------------------------
@@ -1158,11 +1254,24 @@ namespace XrdCl
         //----------------------------------------------------------------------
         else
         {
-          log->Dump( XRootDMsg, "[%s] Parsing the response to %s as StatInfo",
-                     pUrl.GetHostId().c_str(),
-                     pRequest->GetDescription().c_str() );
+          StatInfo *data = new StatInfo();
 
-          StatInfo *data = new StatInfo( buffer );
+          char *nullBuffer = new char[length+1];
+          nullBuffer[length] = 0;
+          memcpy( nullBuffer, buffer, length );
+
+          log->Dump( XRootDMsg, "[%s] Parsing the response to %s as StatInfo: "
+                     "%s", pUrl.GetHostId().c_str(),
+                     pRequest->GetDescription().c_str(), nullBuffer );
+
+          if( data->ParseServerResponse( nullBuffer ) == false )
+          {
+              delete obj;
+              delete data;
+              delete [] nullBuffer;
+              return Status( stError, errInvalidResponse );
+          }
+          delete [] nullBuffer;
           obj->Set( data );
         }
 
@@ -1207,9 +1316,24 @@ namespace XrdCl
         char *path = new char[req->dirlist.dlen+1];
         path[req->dirlist.dlen] = 0;
         memcpy( path, pRequest->GetBuffer(24), req->dirlist.dlen );
-        DirectoryList *data = new DirectoryList( pUrl.GetHostId(), path,
-                                                 length ? buffer : 0 );
+
+        DirectoryList *data = new DirectoryList();
+        data->SetParentName( path );
         delete [] path;
+
+        char *nullBuffer = new char[length+1];
+        nullBuffer[length] = 0;
+        memcpy( nullBuffer, buffer, length );
+
+        if( data->ParseServerResponse( pUrl.GetHostId(), nullBuffer ) == false )
+        {
+          delete data;
+          delete obj;
+          delete [] nullBuffer;
+          return Status( stError, errInvalidResponse );
+        }
+
+        delete [] nullBuffer;
         obj->Set( data );
         response = obj;
         return Status();
@@ -1234,13 +1358,38 @@ namespace XrdCl
         AnyObject *obj      = new AnyObject();
         StatInfo  *statInfo = 0;
 
+        //----------------------------------------------------------------------
+        // Handle StatInfo if requested
+        //----------------------------------------------------------------------
         if( req->open.options & kXR_retstat )
         {
-          log->Dump( XRootDMsg, "[%s] Found StatInfo in response to %s",
+          log->Dump( XRootDMsg, "[%s] Parsing StatInfo in response to %s",
                      pUrl.GetHostId().c_str(),
                      pRequest->GetDescription().c_str() );
+
           if( rsp->hdr.dlen >= 12 )
-            statInfo = new StatInfo( buffer+12 );
+          {
+            char *nullBuffer = new char[rsp->hdr.dlen-11];
+            nullBuffer[rsp->hdr.dlen-12] = 0;
+            memcpy( nullBuffer, buffer+12, rsp->hdr.dlen-12 );
+
+            statInfo = new StatInfo();
+            if( statInfo->ParseServerResponse( nullBuffer ) == false )
+            {
+              delete statInfo;
+              statInfo = 0;
+            }
+            delete [] nullBuffer;
+          }
+
+          if( rsp->hdr.dlen < 12 || !statInfo )
+          {
+            log->Error( XRootDMsg, "[%s] Unable to parse StatInfo in response "
+                        "to %s", pUrl.GetHostId().c_str(),
+                        pRequest->GetDescription().c_str() );
+            delete obj;
+            return Status( stError, errInvalidResponse );
+          }
         }
 
         OpenInfo *data = new OpenInfo( (uint8_t*)buffer,
@@ -1363,7 +1512,9 @@ namespace XrdCl
   // Perform the changes to the original request needed by the redirect
   // procedure - allocate new streamid, append redirection data and such
   //----------------------------------------------------------------------------
-  Status XRootDMsgHandler::RewriteRequestRedirect( const URL::ParamsMap &newCgi )
+  Status XRootDMsgHandler::RewriteRequestRedirect(
+    const URL::ParamsMap &newCgi,
+    const std::string    &newPath )
   {
     Log *log = DefaultEnv::GetLog();
     ClientRequest  *req = (ClientRequest *)pRequest->GetBuffer();
@@ -1399,11 +1550,8 @@ namespace XrdCl
     //--------------------------------------------------------------------------
     // Rewrite particular requests
     //--------------------------------------------------------------------------
-    if( newCgi.empty() )
-      return Status();
-
     XRootDTransport::UnMarshallRequest( pRequest );
-    MessageUtils::AppendCGI( pRequest, newCgi, true );
+    MessageUtils::RewriteCGIAndPath( pRequest, newCgi, true, newPath );
     XRootDTransport::MarshallRequest( pRequest );
     return Status();
   }
@@ -1566,8 +1714,6 @@ namespace XrdCl
   //----------------------------------------------------------------------------
   void XRootDMsgHandler::HandleError( Status status, Message *msg )
   {
-    (void)msg;
-
     //--------------------------------------------------------------------------
     // If there was no error then do nothing
     //--------------------------------------------------------------------------
@@ -1575,7 +1721,7 @@ namespace XrdCl
       return;
 
     Log *log = DefaultEnv::GetLog();
-    log->Error( XRootDMsg, "[%s] Handling error while processing %s: %s.",
+    log->Debug( XRootDMsg, "[%s] Handling error while processing %s: %s.",
                 pUrl.GetHostId().c_str(), pRequest->GetDescription().c_str(),
                 status.ToString().c_str() );
 
@@ -1599,6 +1745,7 @@ namespace XrdCl
           SwitchOnRefreshFlag();
         HandleError( RetryAtServer( pLoadBalancer.url ) );
         delete pResponse;
+        pResponse = 0;
         return;
       }
       else
@@ -1656,8 +1803,9 @@ namespace XrdCl
   //----------------------------------------------------------------------------
   Status XRootDMsgHandler::RetryAtServer( const URL &url )
   {
+    if( pUrl.GetHostId() != url.GetHostId() )
+      pHosts->push_back( url );
     pUrl = url;
-    pHosts->push_back( pUrl );
     return pPostMaster->Send( pUrl, pRequest, this, true, pExpiration );
   }
 
@@ -1667,9 +1815,31 @@ namespace XrdCl
   void XRootDMsgHandler::UpdateTriedCGI()
   {
     URL::ParamsMap cgi;
-    cgi["tried"] = pUrl.GetHostName();
+    std::string    tried = pUrl.GetHostName();
+
+    //--------------------------------------------------------------------------
+    // If our current load balancer is a metamanager and we failed either
+    // at a diskserver or at an unidentified node we also exclude the last
+    // known manager
+    //--------------------------------------------------------------------------
+    if( pLoadBalancer.url.IsValid() && (pLoadBalancer.flags & kXR_attrMeta) )
+    {
+      HostList::reverse_iterator it;
+      for( it = pHosts->rbegin()+1; it != pHosts->rend(); ++it )
+      {
+        if( it->loadBalancer )
+          break;
+
+        tried += "," + it->url.GetHostName();
+
+        if( it->flags & kXR_isManager )
+          break;
+      }
+    }
+
+    cgi["tried"] = tried;
     XRootDTransport::UnMarshallRequest( pRequest );
-    MessageUtils::AppendCGI( pRequest, cgi, false );
+    MessageUtils::RewriteCGIAndPath( pRequest, cgi, false, "" );
     XRootDTransport::MarshallRequest( pRequest );
   }
 
