@@ -41,7 +41,6 @@
 #include <sys/resource.h>
 #include <sys/stat.h>
 #include <sys/types.h>
-#include <sys/un.h>
 #include <dirent.h>
 
 #include "XrdVersion.hh"
@@ -49,11 +48,11 @@
 
 #include "XrdCms/XrdCmsAdmin.hh"
 #include "XrdCms/XrdCmsBaseFS.hh"
+#include "XrdCms/XrdCmsBlackList.hh"
 #include "XrdCms/XrdCmsCache.hh"
 #include "XrdCms/XrdCmsCluster.hh"
 #include "XrdCms/XrdCmsConfig.hh"
 #include "XrdCms/XrdCmsManager.hh"
-#include "XrdCms/XrdCmsManTree.hh"
 #include "XrdCms/XrdCmsMeter.hh"
 #include "XrdCms/XrdCmsNode.hh"
 #include "XrdCms/XrdCmsPrepare.hh"
@@ -65,10 +64,10 @@
 #include "XrdCms/XrdCmsState.hh"
 #include "XrdCms/XrdCmsSupervisor.hh"
 #include "XrdCms/XrdCmsTrace.hh"
-#include "XrdCms/XrdCmsXmi.hh"
-#include "XrdCms/XrdCmsXmiReq.hh"
+#include "XrdCms/XrdCmsUtils.hh"
 
 #include "XrdNet/XrdNetOpts.hh"
+#include "XrdNet/XrdNetUtils.hh"
 #include "XrdNet/XrdNetSecurity.hh"
 #include "XrdNet/XrdNetSocket.hh"
 
@@ -82,7 +81,6 @@
 #include "XrdOuc/XrdOucStream.hh"
 #include "XrdOuc/XrdOucUtils.hh"
 
-#include "XrdSys/XrdSysDNS.hh"
 #include "XrdSys/XrdSysError.hh"
 #include "XrdSys/XrdSysHeaders.hh"
 #include "XrdSys/XrdSysPlatform.hh"
@@ -96,29 +94,22 @@ using namespace XrdCms;
 /*                        G l o b a l   O b j e c t s                         */
 /******************************************************************************/
 
-       XrdCmsAdmin      XrdCms::Admin;
+namespace XrdCms
+{
+       XrdOucEnv        theEnv;
 
-       XrdCmsBaseFS     XrdCms::baseFS(&XrdCmsNode::do_StateDFS);
+       XrdCmsAdmin      Admin;
 
-       XrdCmsConfig     XrdCms::Config;
+       XrdCmsBaseFS     baseFS(&XrdCmsNode::do_StateDFS);
 
-       XrdSysError      XrdCms::Say(0, "");
+       XrdCmsConfig     Config;
 
-       XrdOucTrace      XrdCms::Trace(&Say);
+       XrdSysError      Say(0, "");
 
-       XrdScheduler    *XrdCms::Sched = 0;
+       XrdOucTrace      Trace(&Say);
 
-       XrdCmsXmi       *XrdCms::Xmi_Chmod  = 0;
-       XrdCmsXmi       *XrdCms::Xmi_Load   = 0;
-       XrdCmsXmi       *XrdCms::Xmi_Mkdir  = 0;
-       XrdCmsXmi       *XrdCms::Xmi_Mkpath = 0;
-       XrdCmsXmi       *XrdCms::Xmi_Prep   = 0;
-       XrdCmsXmi       *XrdCms::Xmi_Rename = 0;
-       XrdCmsXmi       *XrdCms::Xmi_Remdir = 0;
-       XrdCmsXmi       *XrdCms::Xmi_Remove = 0;
-       XrdCmsXmi       *XrdCms::Xmi_Select = 0;
-       XrdCmsXmi       *XrdCms::Xmi_Space  = 0;
-       XrdCmsXmi       *XrdCms::Xmi_Stat   = 0;
+       XrdScheduler    *Sched = 0;
+};
 
 /******************************************************************************/
 /*                S e c u r i t y   S y m b o l   T i e - I n                 */
@@ -164,6 +155,29 @@ void *XrdCmsStartSupervising(void *carg)
        return (void *)0;
       }
 
+/******************************************************************************/
+/*                    P i n g   C l o c k   H a n d l e r                     */
+/******************************************************************************/
+  
+namespace XrdCms
+{
+  
+class PingClock : XrdJob
+{
+public:
+
+       void DoIt() {Config.PingTick++;
+                    Sched->Schedule((XrdJob *)this,time(0)+Config.AskPing);
+                   }
+
+static void Start() {static PingClock selfie;}
+
+          PingClock() : XrdJob("ping clock") {DoIt();}
+         ~PingClock() {}
+private:
+};
+};
+
 /******************************************************************************/
 /*                               d e f i n e s                                */
 /******************************************************************************/
@@ -171,6 +185,7 @@ void *XrdCmsStartSupervising(void *carg)
 #define TS_String(x,m) if (!strcmp(x,var)) {free(m); m = strdup(val); return 0;}
 
 #define TS_Xeq(x,m)    if (!strcmp(x,var)) return m(eDest, CFile);
+#define TS_Xer(x,m,v)  if (!strcmp(x,var)) return m(eDest, CFile, v);
 
 #define TS_Set(x,v)    if (!strcmp(x,var)) {v=1; CFile.Echo(); return 0;}
 
@@ -207,22 +222,28 @@ int XrdCmsConfig::Configure1(int argc, char **argv, char *cfn)
 //
    opterr = 0; optind = 1;
    if (argc > 1 && '-' == *argv[1]) 
-      while ((c=getopt(argc,argv,"imsw")) && ((unsigned char)c != 0xff))
+      while ((c=getopt(argc,argv,"iw")) && ((unsigned char)c != 0xff))
      { switch(c)
        {
        case 'i': immed = 1;
                  break;
-       case 'm': isManager = 1;
-                 break;
-       case 's': isServer = 1;
-                 break;
        case 'w': immed = -1;   // Backward compatability only
                  break;
        default:  buff[0] = '-'; buff[1] = optopt; buff[2] = '\0';
-                 Say.Say("Config warning: unrecognized option,",buff,", ignored.");
+                 Say.Say("Config warning: unrecognized option, ",buff,", ignored.");
        }
      }
 
+// Accept a single parameter defining the overiding major role
+//
+   if (optind < argc)
+      {     if (!strcmp(argv[optind], "manager")) isManager = 1;
+       else if (!strcmp(argv[optind], "server" )) isServer  = 1;
+       else if (!strcmp(argv[optind], "super"  )) isServer  = isManager = 1;
+       else Say.Say("Config warning: unrecognized parameter, ",
+                    argv[optind],", ignored.");
+      }
+
 // Bail if no configuration file specified
 //
    inArgv = argv; inArgc = argc;
@@ -272,7 +293,7 @@ int XrdCmsConfig::Configure1(int argc, char **argv, char *cfn)
 // Create a text description of our role for use in messages
 //
    if (!myRole)
-      {XrdCmsRole::RoleID rid;
+      {XrdCmsRole::RoleID rid = XrdCmsRole::noRole;
              if (isMeta)        rid = XrdCmsRole::MetaManager;
         else if (isPeer)        rid = XrdCmsRole::Peer;
         else if (isProxy)
@@ -290,6 +311,11 @@ int XrdCmsConfig::Configure1(int argc, char **argv, char *cfn)
         myRoleID = static_cast<int>(rid);
       }
 
+// Export the role IN basic form and expanded form
+//
+   XrdOucEnv::Export("XRDROLE", myRole);
+   XrdOucEnv::Export("XRDROLETYPE", myRType);
+
 // For managers, make sure that we have a well designated port.
 // For servers or supervisors, force an ephemeral port to be used.
 //
@@ -382,10 +408,11 @@ int XrdCmsConfig::Configure2()
 //
    if ((LocalRoot || RemotRoot) && ConfigN2N()) NoGo = 1;
 
-// Configure the OSS and the base filesystem
+// Configure the OSS, the base filesystem, and initialize the prep queue
 //
    if (!NoGo) NoGo = ConfigOSS();
    if (!NoGo) baseFS.Start();
+   if (!NoGo) PrepQ.Init();
 
 // Setup manager or server, as needed
 //
@@ -413,10 +440,6 @@ int XrdCmsConfig::Configure2()
 //
    if (!NoGo) NoGo |= PidFile();
 
-// Load the XMI plugin
-//
-   if (!NoGo && XmiPath) NoGo = setupXmi();
-
 // All done, check for success or failure
 //
    sprintf(buff, " phase 2 %s initialization %s.", myRole,
@@ -459,6 +482,7 @@ int XrdCmsConfig::ConfigXeq(char *var, XrdOucStream &CFile, XrdSysError *eDest)
    TS_Xeq("adminpath",     xapath);  // Any,     non-dynamic
    TS_Xeq("allow",         xallow);  // Manager, non-dynamic
    TS_Xeq("altds",         xaltds);  // Server,  non-dynamic
+   TS_Xeq("blacklist",     xblk);    // Manager, non-dynamic
    TS_Xeq("defaults",      xdefs);   // Server,  non-dynamic
    TS_Xeq("dfs",           xdfs);    // Any,     non-dynamic
    TS_Xeq("export",        xexpo);   // Any,     non-dynamic
@@ -475,9 +499,10 @@ int XrdCmsConfig::ConfigXeq(char *var, XrdOucStream &CFile, XrdSysError *eDest)
    TS_Xeq("repstats",      xreps);   // Any,     non-dynamic
    TS_Xeq("role",          xrole);   // Server,  non-dynamic
    TS_Xeq("seclib",        xsecl);   // Server,  non-dynamic
+   TS_Xeq("subcluster",    xsubc);   // Manager, non-dynamic
    TS_Set("wait",          doWait);  // Server,  non-dynamic (backward compat)
    TS_unSet("nowait",      doWait);  // Server,  non-dynamic
-   TS_Xeq("xmilib",        xxmi);    // Any,     non-dynamic
+   TS_Xer("whitelist",     xblk,true);//Manager, non-dynamic
    }
 
    // The following are client directives that we will ignore
@@ -499,8 +524,6 @@ int XrdCmsConfig::ConfigXeq(char *var, XrdOucStream &CFile, XrdSysError *eDest)
 void XrdCmsConfig::DoIt()
 {
    XrdSysSemaphore SyncUp(0);
-   XrdCmsProtocol *pP;
-   XrdOucTList    *tp;
    pthread_t       tid;
    time_t          eTime = time(0);
    int             wTime;
@@ -535,6 +558,10 @@ void XrdCmsConfig::DoIt()
           }
       }
 
+// Start the ping clock if we are a manager of any kind
+//
+   if (isManager) PingClock::Start();
+
 // Start the admin thread if we need to, we will not continue until told
 // to do so by the admin interface.
 //
@@ -546,25 +573,9 @@ void XrdCmsConfig::DoIt()
        SyncUp.Wait();
       }
 
-// Start the server subsystem. We check here to make sure we will not be
-// tying to connect to ourselves. This is possible if the manager and meta-
-// manager were defined to be the same and we are a manager. We would have
-// liked to screen this out earlier but port discovery prevents it.
+// Start the manager subsystem.
 //
-   if (isManager || isServer || isPeer)
-      {tp = ManList;
-       while(tp)
-            {if (strcmp(tp->text, myName) || tp->val != PortTCP)
-                {pP = XrdCmsProtocol::Alloc(myRole, tp->text, tp->val);
-                 Sched->Schedule((XrdJob *)pP);
-                } else {
-                 char buff[512];
-                 sprintf(buff, "%s:%d", tp->text, tp->val);
-                 Say.Emsg("Config", "Circular connection to", buff, "ignored.");
-                }
-             tp = tp->next;
-            }
-      }
+   if (isManager || isServer || isPeer) XrdCmsManager::Start(ManList);
 
 // Start state monitoring thread
 //
@@ -614,10 +625,7 @@ int XrdCmsConfig::GenLocalPath(const char *oldp, char *newp)
 
 void XrdCmsConfig::ConfigDefaults(void)
 {
-   extern long timezone;
    static XrdVERSIONINFODEF(myVer, cmsd, XrdVNUMBER, XrdVERSION);
-   time_t myTime = time(0);
-   struct tm *tmP = localtime(&myTime);
    int myTZ, isEast = 0;
 
 // Preset all variables with common defaults
@@ -649,6 +657,9 @@ void XrdCmsConfig::ConfigDefaults(void)
    P_pag    = 0;
    AskPerf  = 10;         // Every 10 pings
    AskPing  = 60;         // Every  1 minute
+   PingTick = 0;
+   DoMWChk  = 1;
+   DoHnTry  = 1;
    MaxDelay = -1;
    LogPerf  = 10;         // Every 10 usage requests
    DiskMin  = 10240;      // 10GB*1024 (Min partition space) in MB
@@ -681,7 +692,10 @@ void XrdCmsConfig::ConfigDefaults(void)
    myRoleID  = XrdCmsRole::noRole;
    ManList   =0;
    NanList   =0;
+   SanList   =0;
    mySID    = 0;
+   mySite   = 0;
+   ifList    =0;
    perfint  = 3*60;
    perfpgm  = 0;
    AdminPath= strdup("/tmp/");
@@ -702,9 +716,9 @@ void XrdCmsConfig::ConfigDefaults(void)
    doWait   = 1;
    RefReset = 60*60;
    RefTurn  = 3*STMax*(DiskLinger+1);
-   XmiPath     = 0;
-   XmiParms    = 0;
    DirFlags    = 0;
+   blkList     = 0;
+   blkChk      = 0;
    SecLib      = 0;
    ossLib      = 0;
    ossParms    = 0;
@@ -716,7 +730,7 @@ void XrdCmsConfig::ConfigDefaults(void)
 
 // Compute the time zone we are in
 //
-   myTZ = timezone/(60*60);
+   myTZ = XrdSysTimer::TimeZone();
    if (myTZ <= 0) {isEast = 0x10; myTZ = -myTZ;}
    if (myTZ > 12) myTZ = 12;
    TimeZone = (myTZ | isEast);
@@ -736,7 +750,7 @@ int XrdCmsConfig::ConfigN2N()
 
 // Optimize the local case
 //
-   if (N2N_Lib || LocalRoot || (RemotRoot && XmiPath)) lcl_N2N = xeq_N2N;
+   if (N2N_Lib || LocalRoot) lcl_N2N = xeq_N2N;
 
 // All done
 //
@@ -751,7 +765,7 @@ int XrdCmsConfig::ConfigN2N()
 int XrdCmsConfig::ConfigOSS()
 {
    extern XrdOss *XrdOssGetSS(XrdSysLogger *, const char *, const char *,
-                              const char   *, XrdVersionInfo &);
+                              const char   *, XrdOucEnv  *, XrdVersionInfo &);
 
 // Set up environment for the OSS to keep it relevant for cmsd
 //
@@ -766,7 +780,8 @@ int XrdCmsConfig::ConfigOSS()
 
 // Load and return result
 //
-   return !(ossFS=XrdOssGetSS(Say.logger(),ConfigFN,ossLib,ossParms,*myVInfo));
+   return !(ossFS=XrdOssGetSS(Say.logger(),ConfigFN,ossLib,ossParms,
+                              &theEnv, *myVInfo));
 }
 
 /******************************************************************************/
@@ -814,7 +829,8 @@ int XrdCmsConfig::ConfigProc(int getrole)
                 ||  !strcmp(var, "all.manager")
                 ||  !strcmp(var, "all.pidpath")
                 ||  !strcmp(var, "all.role")
-                ||  !strcmp(var, "all.seclib"))
+                ||  !strcmp(var, "all.seclib")
+                ||  !strcmp(var, "all.subcluster"))
                    {if (ConfigXeq(var+4, CFile, 0)) {CFile.Echo(); NoGo = 1;}}
                    else if (!strcmp(var, "oss.stagecmd")) DiskSS = 1;
 
@@ -826,7 +842,7 @@ int XrdCmsConfig::ConfigProc(int getrole)
 
 // Merge Paths as needed
 //
-   if (!getrole && ManList) NoGo |= MergeP();
+   if (!getrole && (ManList || SanList)) NoGo |= MergeP();
 
 // Return final return code
 //
@@ -895,8 +911,11 @@ int XrdCmsConfig::MergeP()
 // Document what we will be declaring as available
 //
    if (!NoGo)
-      {const char *Who = (isManager ? (isServer ? "manager:" : "meta-manager:")
-                                    : "redirector:");
+      {const char *Who;
+       if (isManager)
+          {if (SanList) Who = "subcluster manager:";
+              else      Who = (isServer ? "manager:" : "meta-manager:");
+          } else        Who = "redirector:";
        Say.Say("The following paths are available to the ", Who);
        if (!(pp = PathList.First())) Say.Say("r  /");
         else while(pp)
@@ -986,6 +1005,28 @@ int XrdCmsConfig::setupManager()
    pthread_t tid;
    int rc;
 
+// If we are a subcluster then we need to replace the manager list with the
+// one specified on the subcluster directive.
+//
+   if (SanList)
+      {XrdOucTList *nP, *tP = ManList;
+       const char *urDom, *myDom = index(myName, '.');
+       bool isBad = false;
+       while(tP) {nP = tP; tP = tP->next; delete nP;}
+       ManList = tP = SanList;
+       if (myDom) while(tP)
+          {if ((urDom = index(tP->text, '.')) && strcmp(urDom, myDom))
+              {Say.Emsg("Config", "Subcluster's manager", tP->text,
+                                  "is in a different domain.");
+               isBad = true;
+              }
+           tP = tP->next;
+          }
+       if (isBad) {Say.Emsg("Config","Cross domain subclusters disallowed!");
+                   return 1;
+                  }
+      }
+
 // Setup supervisor mode if we are also a server
 //
    if (isServer && !XrdCmsSupervisor::Init(AdminPath, AdminMode)) return 1;
@@ -1024,6 +1065,11 @@ int XrdCmsConfig::setupManager()
 //
    if (SecLib && !XrdCmsSecurity::Configure(SecLib, ConfigFN)) return 1;
 
+// Initialize the black list
+//
+   if (!isServer && blkChk)
+      XrdCmsBlackList::Init(Sched, &Cluster, blkList, blkChk);
+
 // All done
 //
    return 0;
@@ -1045,13 +1091,12 @@ int XrdCmsConfig::setupServer()
        return 1;
       }
 
-// Count the number of managers we have and tell ManTree about it
+// Count the number of managers. Make sure there are not too many.
 //
    tp = ManList;
    while(tp) {n++; tp = tp->next;}
    if (n > XrdCmsManager::MTMax)
       {Say.Emsg("Config", "Too many managers have been specified"); return 1;}
-   ManTree.setMaxCon(n);
 
 // Calculate overload delay time
 //
@@ -1096,6 +1141,16 @@ char *XrdCmsConfig::setupSid()
    XrdOucTList *tp = (NanList ? NanList : ManList);
    char sfx;
 
+// Grab the interfaces. This is normally set as an envar. If present then
+// we will copy it because we must use it permanently.
+//
+   if (getenv("XRDIFADDRS")) ifList = strdup(getenv("XRDIFADDRS"));
+
+// Grab the site name
+//
+   if ((mySite = getenv("XRDSITE")) && *mySite) mySite = strdup(mySite);
+      else mySite = 0;
+
 // Determine what type of role we are playing
 //
    if (isManager && isServer) sfx = 'u';
@@ -1106,85 +1161,6 @@ char *XrdCmsConfig::setupSid()
    return XrdCmsSecurity::setSystemID(tp, myInsName, myName, sfx);
 }
 
-/******************************************************************************/
-/*                              s e t u p X m i                               */
-/******************************************************************************/
-  
-int XrdCmsConfig::setupXmi()
-{
-   EPNAME("setupXmi");
-   static XrdCmsXmiEnv XmiEnv;
-   XrdSysPlugin       *xmiLib;
-   XrdCmsXmi          *(*ep)(int, char **, XrdCmsXmiEnv *);
-   unsigned int isNormal, isDirect;
-   XrdCmsXmi          *XMI, *myXMI;
-   const char         *theMode;
-   int i;
-
-   struct {unsigned int   theMask;
-           XrdCmsXmi    **theAddr;
-           const char    *theName;} XmiTab[] =
-          {{XMI_CHMOD,  &Xmi_Chmod,  "chmod"},
-           {XMI_LOAD,   &Xmi_Load,   "load"},
-           {XMI_MKDIR,  &Xmi_Mkdir,  "mkdir"},
-           {XMI_MKPATH, &Xmi_Mkpath, "mkpath"},
-           {XMI_PREP,   &Xmi_Prep,   "prep"},
-           {XMI_RENAME, &Xmi_Rename, "rename"},
-           {XMI_REMDIR, &Xmi_Remdir, "remdir"},
-           {XMI_REMOVE, &Xmi_Remove, "remove"},
-           {XMI_SELECT, &Xmi_Select, "select"},
-           {XMI_SPACE,  &Xmi_Space,  "space"},
-           {XMI_STAT,   &Xmi_Stat,   "stat"}};
-   int numintab = sizeof(XmiTab)/sizeof(XmiTab[0]);
-
-// Fill out the rest of the XmiEnv structure
-//
-   XmiEnv.Role     = myRole;
-   XmiEnv.ConfigFN = ConfigFN;
-   XmiEnv.Parms    = XmiParms;
-   XmiEnv.eDest    = &Say;
-   XmiEnv.iNet     = NetTCP;
-   XmiEnv.Sched    = Sched;
-   XmiEnv.Trace    = &Trace;
-   XmiEnv.Name2Name= xeq_N2N;
-
-// Create a pluin object (we will throw this away without deletion because
-// the library must stay open but we never want to reference it again).
-//
-   if (!(xmiLib = new XrdSysPlugin(&Say, XmiPath, "xmilib", myVInfo))) return 1;
-
-// Now get the entry point of the object creator
-//
-   ep = (XrdCmsXmi *(*)(int, char **, XrdCmsXmiEnv *))(xmiLib->getPlugin("XrdCmsgetXmi"));
-   if (!ep) return 1;
-
-// Get the Object now
-//
-   if (!(XMI = ep(inArgc, inArgv, &XmiEnv))) return 1;
-   DEBUG("xmi library loaded; path=" <<XmiPath);
-
-// Obtain the execution mode
-//
-   XMI->XeqMode(isNormal, isDirect);
-
-// Check if we need to create an indirect XMI interface
-//
-   if ((isDirect & XMI_ALL) == XMI_ALL) myXMI = 0;
-      else myXMI = (XrdCmsXmi *)new XrdCmsXmiReq(XMI);
-
-// Now run throw all of the possibilities setting the execution mode as needed
-//
-   for (i = 0; i < numintab; i++)
-       {if (!(isNormal & XmiTab[i].theMask))
-           if (isDirect & XmiTab[i].theMask)
-                   {*XmiTab[i].theAddr =   XMI; theMode = "direct";}
-              else {*XmiTab[i].theAddr = myXMI; theMode = "queued";}
-           else theMode = "normal";
-        DEBUG(XmiTab[i].theName <<" is " <<theMode);
-       }
-   return 0;
-}
-
 /******************************************************************************/
 /*                                 U s a g e                                  */
 /******************************************************************************/
@@ -1280,7 +1256,7 @@ int XrdCmsConfig::xaltds(XrdSysError *eDest, XrdOucStream &CFile)
        {if (XrdOuca2x::a2i(*eDest,"data server port",val,&adsPort,1,65535))
            return 1;
        }
-       else if (!(adsPort = XrdSysDNS::getPort(val, "tcp")))
+       else if (!(adsPort = XrdNetUtils::ServPort(val, "tcp")))
                {eDest->Emsg("Config", "Unable to find tcp service '",val,"'.");
                 return 1;
                }
@@ -1313,7 +1289,6 @@ int XrdCmsConfig::xapath(XrdSysError *eDest, XrdOucStream &CFile)
 {
     char *pval, *val;
     mode_t mode = S_IRWXU;
-    struct sockaddr_un USock;
 
 // Get the path
 //
@@ -1325,13 +1300,6 @@ int XrdCmsConfig::xapath(XrdSysError *eDest, XrdOucStream &CFile)
 //
    if (*pval != '/')
       {eDest->Emsg("Config", "adminpath not absolute"); return 1;}
-
-// Make sure path is not too long (account for "/olbd.admin")
-//                                              12345678901
-   if (strlen(pval) > sizeof(USock.sun_path) - 11)
-      {eDest->Emsg("Config", "admin path", pval, "is too long");
-       return 1;
-      }
    pval = strdup(pval);
 
 // Get the optional access rights
@@ -1351,6 +1319,69 @@ int XrdCmsConfig::xapath(XrdSysError *eDest, XrdOucStream &CFile)
    return 0;
 }
 
+/******************************************************************************/
+/*                                  x b l k                                   */
+/******************************************************************************/
+
+/* Function: xblk
+
+   Purpose:  To parse the directive: blacklist [check <time>] [<path>]
+
+             <time>    how often to check for black list changes.
+             <path>    the path to the blacklist file
+
+  Output: 0 upon success or !0 upon failure.
+*/
+
+int XrdCmsConfig::xblk(XrdSysError *eDest, XrdOucStream &CFile, bool iswl)
+{
+   const char *fType = (iswl ? "whitelist" : "blacklist");
+   char *val = CFile.GetWord();
+
+// We only support this for managers
+//
+   if (!isManager || isServer) return CFile.noEcho();
+
+// Indicate blacklisting is active and free up any current blacklist path
+//
+   blkChk = 600;
+   if (blkList) {free(blkList); blkList = 0;}
+
+// Avoid echoing limitation in the stream object
+//
+   if (!val || !val[0])
+      {eDest->Say("=====> cms.", fType);
+       return 0;
+      }
+
+// Process any options
+//
+   do {     if (!strcmp(val, "check"))
+               {if (!(val = CFile.GetWord()) || !val[0])
+                   {eDest->Emsg("Config",fType,"check interval not specified");
+                    return 1;
+                   }
+                if (XrdOuca2x::a2tm(*eDest, "check value", val, &blkChk, 60)) return 1;
+               }
+       else break;
+      } while((val = CFile.GetWord()));
+
+// Handle the invert option
+//
+   if (iswl) blkChk = -blkChk;
+
+// Verify the path, if any. is absolute
+//
+   if (!val || !val[0]) return 0;
+   if (*val != '/')
+      {eDest->Emsg("Config", "blacklist path not absolute"); return 1;}
+
+// Record the path
+//
+   blkList = strdup(val);
+   return 0;
+}
+  
 /******************************************************************************/
 /*                                x d e l a y                                 */
 /******************************************************************************/
@@ -1535,7 +1566,7 @@ int XrdCmsConfig::xdfs(XrdSysError *eDest, XrdOucStream &CFile)
 do{     if (!strcmp("mdhold",  val))
            {if (!(val = CFile.GetWord()))
                {eDest->Emsg("Config","mdhold value not specified.");  return 1;}
-            if (XrdOuca2x::a2tm(*eDest, "hold value", val, &Hold, 1)) return 1;
+            if (XrdOuca2x::a2tm(*eDest, "hold value", val, &Hold, 0)) return 1;
            }
    else if (!strcmp("limit",   val))
            {if (!(val = CFile.GetWord()))
@@ -1817,14 +1848,12 @@ int XrdCmsConfig::xmang(XrdSysError *eDest, XrdOucStream &CFile)
     char **val1, **val2;
     };
 
-    static const int sockALen = sizeof(struct sockaddr);
-    struct sockaddr InetAddr[8];
-    XrdOucTList *tp = 0, *tpp = 0, *tpnew;
-    char *val, *bval = 0, *mval = 0;
-    StorageHelper SHelp(&bval, &mval);
-    int j, i, multi = 0, port = 0, xMeta = 0, xPeer = 0, xProxy = 0, Prt = 1;
+    XrdOucTList **theList = &ManList;
+    char *val, *hSpec = 0, *hPort = 0;
+    StorageHelper SHelp(&hSpec, &hPort);
+    int rc, xMeta = 0, xPeer = 0, xProxy = 0, *myPort = 0;
 
-//  Process the optional "peer" or "proxy"
+//  Process the optional "meta", "peer" or "proxy"
 //
     if ((val = CFile.GetWord()))
        {if ((xMeta  = !strcmp("meta", val))
@@ -1842,94 +1871,41 @@ int XrdCmsConfig::xmang(XrdSysError *eDest, XrdOucStream &CFile)
     if (val)
        if (!strcmp("any", val) || !strcmp("all", val)) val = CFile.GetWord();
 
-//  Get the actual host name
+//  Get the actual host name and copy it
 //
     if (!val)
        {eDest->Emsg("Config","manager host name not specified"); return 1;}
-    mval = strdup(val);
+    hSpec = strdup(val);
 
-//  Grab the port number
+//  Grab the port number (either in hostname or following token)
 //
-    if ((val = index(mval, ':'))) {*val = '\0'; val++;}
-       else val = CFile.GetWord();
+    if (!(hPort = XrdCmsUtils::ParseManPort(eDest, CFile, hSpec))) return 1;
 
-    if (val)
-       {if (isdigit(*val))
-           {if (XrdOuca2x::a2i(*eDest,"manager port",val,&port,1,65535))
-               port = 0;
-           }
-           else if (!(port = XrdSysDNS::getPort(val, "tcp")))
-                   {eDest->Emsg("Config", "Unable to find tcp service '",val,"'.");
-                    port = 0;
-                   }
-       }
-       else if (!(port = PortTCP))
-               eDest->Emsg("Config","manager port not specified for",mval);
-
-    if (!port) return 1;
-
-    if ((val = CFile.GetWord()))
-       {if (strcmp(val, "if"))
-           {eDest->Emsg("Config","expecting manager 'if' but",val,"found");
-            return 1;
-           }
-        if ((i=XrdOucUtils::doIf(eDest,CFile,"manager directive",
-                            myName,myInsName,myProg))<=0)
-           {if (!i) CFile.noEcho(); return i < 0;}
-       }
-
-    i = strlen(mval);
-    if (mval[i-1] != '+') 
-       {i = 1;
-        if (!XrdSysDNS::getHostAddr(mval, InetAddr))
-           {eDest->Emsg("CFile","Manager host", mval, "not found");
-            return 1;
-           }
-        free(mval); mval = XrdSysDNS::getHostName(InetAddr[0]);
-       }
-        else {bval = strdup(mval); mval[i-1] = '\0'; multi = 1;
-              if (!(i = XrdSysDNS::getHostAddr(mval, InetAddr, 8)))
-                 {eDest->Emsg("CFile","Manager host", mval, "not found");
-                  return 1;
-                 }
-             }
-
-// Now try to determine our port number if we are a [meta] manager
+// Check if this statement is gaurded by and "if" and process it
 //
-    if (isManager && !isServer)
-       {if ((xMeta && isMeta) || (!xMeta && !isMeta))
-           for (j = 0; j < i; j++)
-                if (!memcmp(&InetAddr[j], &myAddr, sockALen))
-                   {PortTCP = port; break;}
-        if (isMeta) return (xMeta ? 0: CFile.noEcho());
-        if (!xMeta) Prt = 0;
+   if ((val = CFile.GetWord()))
+      {if (strcmp(val, "if"))
+          {eDest->Emsg("Config","expecting manager 'if' but",val,"found");
+           return 1;
+          }
+       if ((rc = XrdOucUtils::doIf(eDest,CFile,"manager directive",
+                                   myName,myInsName,myProg))<=0)
+          {if (!rc) CFile.noEcho(); return rc < 0;}
+      }
+
+// Calculate the correct queue and port number to update
+//
+   if (isManager && !isServer)
+//    {if (((xMeta && isMeta) || (!xMeta && !isMeta)) && PortTCP < 1)
+      {if (((xMeta && isMeta) || (!xMeta && !isMeta)))
+          myPort = &PortTCP;
+        if (isMeta) theList = 0;
+           else theList = (xMeta ? &ManList : &NanList);
        }
 
-// Now construct the list of [meta] managers
+// Parse the specification and return
 //
-    do {if (multi)
-           {free(mval);
-            char mvBuff[1024];
-            if (Prt) sprintf(mvBuff, "%s -> all.manager ", bval);
-            mval = XrdSysDNS::getHostName(InetAddr[i-1]);
-            if (Prt) eDest->Say("Config ", mvBuff, mval);
-           }
-        tp = (Prt ? ManList : NanList); tpp = 0; j = 1;
-        while(tp) 
-             if ((j = strcmp(tp->text, mval)) < 0 || tp->val != port)
-                {tpp = tp; tp = tp->next;}
-                else {if (Prt && !j)
-                         eDest->Say("Config warning: duplicate manager ",mval);
-                      break;
-                     }
-        if (j) {tpnew = new XrdOucTList(mval, port, tp);
-                if (tpp) tpp->next = tpnew;
-                   else if (Prt) ManList = tpnew;
-                           else  NanList = tpnew;
-               }
-       } while(--i);
-
-    return 0;
+   return (XrdCmsUtils::ParseMan(eDest, theList, hSpec, hPort, myPort) ? 0 : 1);
 }
   
 /******************************************************************************/
@@ -2120,7 +2096,7 @@ int XrdCmsConfig::xping(XrdSysError *eDest, XrdOucStream &CFile)
     if (!(val = CFile.GetWord()))
        {eDest->Emsg("Config", "ping value not specified"); return 1;}
     if (XrdOuca2x::a2tm(*eDest, "ping interval",val,&ping,0)) return 1;
-
+    if (ping < 3) ping = 3;
 
     while((val = CFile.GetWord()))
         {     if (!strcmp("log", val))
@@ -2412,7 +2388,7 @@ int XrdCmsConfig::xrole(XrdSysError *eDest, XrdOucStream &CFile)
 {
     XrdCmsRole::RoleID roleID;
     char *val, *Tok1, *Tok2;
-    int rc, xMeta=0, xPeer=0, xProxy=0, xServ=0, xMan=0, xSolo=0, xSup=0;
+    int rc, xMeta=0, xPeer=0, xProxy=0, xServ=0, xMan=0, xSolo=0;
 
 // Get the first token
 //
@@ -2447,7 +2423,7 @@ int XrdCmsConfig::xrole(XrdSysError *eDest, XrdOucStream &CFile)
    switch(roleID)
          {case XrdCmsRole::MetaManager:  xMeta  = xMan  =         -1; break;
           case XrdCmsRole::Manager:               xMan  =         -1; break;
-          case XrdCmsRole::Supervisor:   xSup   = xMan  = xServ = -1; break;
+          case XrdCmsRole::Supervisor:            xMan  = xServ = -1; break;
           case XrdCmsRole::Server:                        xServ = -1; break;
           case XrdCmsRole::ProxyManager: xProxy = xMan  =         -1; break;
           case XrdCmsRole::ProxySuper:   xProxy = xMan  = xServ = -1; break;
@@ -2466,7 +2442,7 @@ int XrdCmsConfig::xrole(XrdSysError *eDest, XrdOucStream &CFile)
 // If the role was specified on the command line, issue warning and ignore this
 //
    if (isServer > 0 || isManager > 0 || isProxy > 0 || isPeer > 0)
-      {eDest->Say("Config warning: role directive over-ridden by command line options.");
+      {eDest->Say("Config warning: role directive over-ridden by command line.");
        return 0;
       }
 
@@ -2510,7 +2486,7 @@ int XrdCmsConfig::xrole(XrdSysError *eDest, XrdOucStream &CFile)
 int XrdCmsConfig::xsched(XrdSysError *eDest, XrdOucStream &CFile)
 {
     char *val;
-    int  i, ppp;
+    int  i, ppp, V_hntry = -1;
     static struct schedopts {const char *opname; int maxv; int *oploc;}
            scopts[] =
        {
@@ -2524,7 +2500,8 @@ int XrdCmsConfig::xsched(XrdSysError *eDest, XrdOucStream &CFile)
         {"pag",      100, &P_pag},
         {"space",    100, &P_dsk},
         {"maxload",  100, &MaxLoad},
-        {"refreset", -1,  &RefReset}
+        {"refreset", -1,  &RefReset},
+        {"tryhname",   1, &V_hntry}
        };
     int numopts = sizeof(scopts)/sizeof(struct schedopts);
 
@@ -2553,6 +2530,10 @@ int XrdCmsConfig::xsched(XrdSysError *eDest, XrdOucStream &CFile)
            val = CFile.GetWord();
           }
 
+// Handle non-int settings
+//
+   if (V_hntry >= 0) DoHnTry = static_cast<char>(V_hntry);
+
     return 0;
 }
 
@@ -2602,6 +2583,8 @@ int XrdCmsConfig::xsecl(XrdSysError *eDest, XrdOucStream &CFile)
 
                           [[min] {<mnp> [<min>] | <min>}  [[<hwp>] <hwm>]]
 
+                          [mwfiles]
+
              <num> Maximum number of times a server may be reselected without
                    a break. The default is 0.
 
@@ -2619,6 +2602,10 @@ int XrdCmsConfig::xsecl(XrdSysError *eDest, XrdOucStream &CFile)
              <sec> Number of seconds that must elapse before a disk free space
                    calculation will occur.
 
+             mwfiles
+                   space supports multiple writable file copies. This suppresses
+                   multiple file check when open a file in write mode.
+
    Notes:   This is used by the manager and the server.
 
    Type: All, dynamic.
@@ -2648,6 +2635,7 @@ int XrdCmsConfig::xspace(XrdSysError *eDest, XrdOucStream &CFile)
                   {eDest->Emsg("Config", "space min value not specified"); return 1;}
                break;
               }
+      else if (!strcmp("mwfiles", val)) DoMWChk = 0;
       else if (isdigit(*val)) break;
       else {eDest->Emsg("Config", "invalid space parameters"); return 1;}
       }
@@ -2715,6 +2703,59 @@ int XrdCmsConfig::xspace(XrdSysError *eDest, XrdOucStream &CFile)
        }
     return 0;
 }
+
+/******************************************************************************/
+/*                                 x s u b c                                  */
+/******************************************************************************/
+
+/* Function: subc
+
+   Purpose:  To parse the directive: subcluster of <host>[+][:<port>|<port>]
+
+   Type: Manager only, non-dynamic.
+
+   Output: 0 upon success or !0 upon failure.
+*/
+  
+int XrdCmsConfig::xsubc(XrdSysError *eDest, XrdOucStream &CFile)
+{
+    class StorageHelper
+    {public:
+          StorageHelper(char **v1, char **v2) : val1(v1), val2(v2) {}
+         ~StorageHelper() {if (*val1) free(*val1);
+                           if (*val2) free(*val2);
+                          }
+    char **val1, **val2;
+    };
+
+    char *val, *hSpec = 0, *hPort = 0;
+    StorageHelper SHelp(&hSpec, &hPort);
+
+// Ignore this call if we are not a simple manager
+//
+   if (isMeta || isServer || isPeer || isProxy) return CFile.noEcho();
+
+//  Skip the optional "of" keyword
+//
+    val = CFile.GetWord();
+    if (val && !strcmp("of", val)) val = CFile.GetWord();
+
+//  Get the actual host name and copy it
+//
+    if (!val)
+       {eDest->Emsg("Config","cluster manager host name not specified");
+        return 1;
+       }
+    hSpec = strdup(val);
+
+//  Grab the port number (either in hostname or following token)
+//
+    if (!(hPort = XrdCmsUtils::ParseManPort(eDest, CFile, hSpec))) return 1;
+
+// Parse the specification and return
+//
+   return (XrdCmsUtils::ParseMan(eDest, &SanList, hSpec, hPort) ? 0 : 1);
+}
   
 /******************************************************************************/
 /*                                x t r a c e                                 */
@@ -2765,41 +2806,3 @@ int XrdCmsConfig::xtrace(XrdSysError *eDest, XrdOucStream &CFile)
     Trace.What = trval;
     return 0;
 }
-  
-/******************************************************************************/
-/*                                  x x m i                                   */
-/******************************************************************************/
-
-/* Function: xxmi
-
-   Purpose:  To parse the directive: xmilib <path> [<parms>]
-
-             <path>    the SO path for the XrdCmsXmi plugin.
-             <parms>   optional parms to be passed to the Xmi object
-
-  Output: 0 upon success or !0 upon failure.
-*/
-
-int XrdCmsConfig::xxmi(XrdSysError *eDest, XrdOucStream &CFile)
-{
-    char *val, parms[1024];
-
-// Get the path
-//
-   if (!(val = CFile.GetWord()) || !val[0])
-      {eDest->Emsg("Config", "xmilib path not specified"); return 1;}
-
-// Record the path
-//
-   if (XmiPath) free(XmiPath);
-   XmiPath = strdup(val);
-
-// Record any parms
-//
-   if (!CFile.GetRest(parms, sizeof(parms)))
-      {eDest->Emsg("Config", "xmilib parameters too long"); return 1;}
-   if (XmiParms) free(XmiParms);
-   XmiParms = (*parms ? strdup(parms) : 0);
-
-   return 0;
-}
