@@ -44,10 +44,10 @@
 #include "XrdCms/XrdCmsBaseFS.hh"
 #include "XrdCms/XrdCmsCache.hh"
 #include "XrdCms/XrdCmsCluster.hh"
+#include "XrdCms/XrdCmsClustID.hh"
 #include "XrdCms/XrdCmsConfig.hh"
 #include "XrdCms/XrdCmsManager.hh"
 #include "XrdCms/XrdCmsManList.hh"
-#include "XrdCms/XrdCmsManTree.hh"
 #include "XrdCms/XrdCmsMeter.hh"
 #include "XrdCms/XrdCmsPList.hh"
 #include "XrdCms/XrdCmsPrepare.hh"
@@ -56,17 +56,16 @@
 #include "XrdCms/XrdCmsSelect.hh"
 #include "XrdCms/XrdCmsState.hh"
 #include "XrdCms/XrdCmsTrace.hh"
-#include "XrdCms/XrdCmsXmi.hh"
 
 #include "XrdOss/XrdOss.hh"
 
-#include "XrdSys/XrdSysDNS.hh"
 #include "XrdOuc/XrdOucName2Name.hh"
 #include "XrdOuc/XrdOucProg.hh"
 #include "XrdOuc/XrdOucPup.hh"
-#include "XrdOuc/XrdOucTokenizer.hh"
 #include "XrdOuc/XrdOucUtils.hh"
 
+#include "XrdNet/XrdNetUtils.hh"
+
 #include "XrdSys/XrdSysPlatform.hh"
 
 using namespace XrdCms;
@@ -78,26 +77,32 @@ using namespace XrdCms;
 XrdSysMutex XrdCmsNode::mlMutex;
 
 int         XrdCmsNode::LastFree = 0;
+
+namespace
+{
+XrdNetIF::ifType ifVec[4] = {XrdNetIF::PublicV4, XrdNetIF::Public46,
+                             XrdNetIF::PublicV6, XrdNetIF::Public64};
+};
   
 /******************************************************************************/
 /*                           C o n s t r u c t o r                            */
 /******************************************************************************/
   
-XrdCmsNode::XrdCmsNode(XrdLink *lnkp, int port,
-                       const char *nid,  int lvl, int id)
+XrdCmsNode::XrdCmsNode(XrdLink *lnkp, const char *theIF, const char *nid,
+                       int port, int lvl, int id)
 {
     static XrdSysMutex   iMutex;
     static const SMask_t smask_1(1);
     static int           iNum = 1;
 
     Link     =  lnkp;
-    IPAddr   =  0;
     NodeMask =  (id < 0 ? 0 : smask_1 << id);
     NodeID   = id;
-    isDisable=  0;
-    isNoStage=  0;
+    cidP     =  0;
+    hasNet   =  0;
+    isBad    =  0;
     isOffline=  (lnkp == 0);
-    isSuspend=  0;
+    isNoStage=  0;
     isBound  =  0;
     isConn   =  0;
     isGone   =  0;
@@ -105,11 +110,9 @@ XrdCmsNode::XrdCmsNode(XrdLink *lnkp, int port,
     isMan    =  0;
     isKnown  =  0;
     isPeer   =  0;
-    isProxy  =  0;
     myCost   =  0;
     myLoad   =  0;
     myMass   =  0;
-    myCNUM   = -3;
     DiskTotal=  0;
     DiskFree =  0;
     DiskMinF =  0;
@@ -129,7 +132,6 @@ XrdCmsNode::XrdCmsNode(XrdLink *lnkp, int port,
     myName   =  0;
     myNlen   =  0;
     Ident    =  0;
-    Port     =  0;
     myNID    = strdup(nid ? nid : "?");
     if ((myCID = index(myNID, ' '))) myCID++;
        else myCID = myNID;
@@ -137,10 +139,12 @@ XrdCmsNode::XrdCmsNode(XrdLink *lnkp, int port,
     ConfigID =  0;
     TZValid  = 0;
     TimeZone = 0;
+    subsPort = 0;
+    myVersion= kYR_Version;
 
-// setName() will set Ident, IPAddr, IPV6, myName, myNlen, & Port!
+// setName() will set the node identification information
 //
-   setName(lnkp, (nid ? port : 0));
+   setName(lnkp, theIF, (nid ? port : 0));
 
    iMutex.Lock();
    Instance =  iNum++;
@@ -158,6 +162,7 @@ XrdCmsNode::~XrdCmsNode()
 
 // Delete other appendages
 //
+   if (cidP) {cidP->RemNode(this); cidP = 0;}
    if (Ident) free(Ident);
    if (myNID) free(myNID);
    if (myName)free(myName);
@@ -167,43 +172,39 @@ XrdCmsNode::~XrdCmsNode()
 /*                               s e t N a m e                                */
 /******************************************************************************/
   
-void XrdCmsNode::setName(XrdLink *lnkp, int port)
+void XrdCmsNode::setName(XrdLink *lnkp, const char *theIF, int port)
 {
-   struct sockaddr netaddr;
-   char *bp, buff[512];
+   char buff[512];
    const char *hname = lnkp->Host();
-   unsigned int hAddr;
 
-// Get our address (the long way)
-//
-   lnkp->Name(&netaddr);
-   hAddr= XrdSysDNS::IPAddr(&netaddr);
-
-// Check if this is a duplicate
+// Check if this is a duplicate. Note that we check for strict equivalence.
 //
    if (myName)
-      {if (!strcmp(myName, hname) && port == Port && hAddr == IPAddr) return;
-          else free(myName);
+      {if (!strcmp(myName,hname) && port == netIF.Port()
+       &&  netID.Same(lnkp->NetAddr())) return;
+       free(myName);
       }
 
+// Get our address information but substitute data port for actual port
+//
+   netID = *(lnkp->NetAddr());
+
+// Set the network interface. Note that out of domain nodes are not allowed
+// to specify interface addresses as this does not make global sense.
+//
+   if (theIF && !netIF.InDomain(&netID)) theIF = 0;
+   netIF.SetIF(&netID, theIF, port);
+   hasNet = netIF.Mask();
+
 // Construct our identification
 //
-   IPAddr = hAddr;
    myName = strdup(hname);
-   myNlen = strlen(hname)+1;
-   Port = port;
+   myNlen = strlen(hname);
 
    if (!port) strcpy(buff, lnkp->ID);
       else    sprintf(buff, "%s:%d", lnkp->ID, port);
    if (Ident) free(Ident);
    Ident = strdup(buff);
-
-   strcpy(IPV6, "[::");
-   bp = IPV6+3;
-   bp += XrdSysDNS::IP2String(IPAddr, 0, bp, 24); // We're cheating
-   *bp++ = ']';
-   if (Port) {*bp++ = ':'; bp += sprintf(bp, "%d", Port);}
-   IPV6Len = bp - IPV6;
 }
 
 /******************************************************************************/
@@ -257,7 +258,7 @@ const char *XrdCmsNode::do_Avail(XrdCmsRRData &Arg)
 /*                              d o _ C h m o d                               */
 /******************************************************************************/
   
-// Chmod requests are forwarded to all subscribers subject to an Xmi callout.
+// Chmod requests are forwarded to all subscribers
 //
 const char *XrdCmsNode::do_Chmod(XrdCmsRRData &Arg)
 {
@@ -269,14 +270,6 @@ const char *XrdCmsNode::do_Chmod(XrdCmsRRData &Arg)
 //
    DEBUGR("mode " <<Arg.Mode <<' ' <<Arg.Path);
 
-// If we have an Xmi then call it
-//
-   if (Xmi_Chmod)
-      {XrdCmsReq Req(this, Arg.Request.streamid);
-       if (!getMode(Arg.Mode, mode)) return "invalid mode";
-          else if (Xmi_Chmod->Chmod(&Req, mode, Arg.Path, Arg.Opaque)) return 0;
-      }
-
 // We are don here if we have no data; otherwise convert the mode if we
 // haven't done so already.
 //
@@ -350,11 +343,11 @@ const char *XrdCmsNode::do_Gone(XrdCmsRRData &Arg)
 
 // If we have no managers and we still have the file or never had it, return
 //
-   if (!Manager.Present() || !newgone) return 0;
+   if (!XrdCmsManager::Present() || !newgone) return 0;
 
 // Back-propogate the gone to all of our managers
 //
-   Manager.Inform(Arg.Request, Arg.Buff, Arg.Dlen);
+   XrdCmsManager::Inform(Arg.Request, Arg.Buff, Arg.Dlen);
 
 // All done
 //
@@ -402,11 +395,11 @@ const char *XrdCmsNode::do_Have(XrdCmsRRData &Arg)
 
 // Return if we have no managers or we already informed the managers
 //
-   if (!Manager.Present() || !isnew) return 0;
+   if (!XrdCmsManager::Present() || !isnew) return 0;
 
 // Back-propogate the have to all of our managers
 //
-   Manager.Inform(Arg.Request, Arg.Buff, Arg.Dlen);
+   XrdCmsManager::Inform(Arg.Request, Arg.Buff, Arg.Dlen);
 
 // All done
 //
@@ -496,26 +489,51 @@ const char *XrdCmsNode::do_Locate(XrdCmsRRData &Arg)
    XrdCmsSelect    Sel(0, Arg.Path, Arg.PathLen-1);
    XrdCmsSelected *sP = 0;
    struct {kXR_unt32 Val; 
-           char outbuff[CmsLocateRequest::RILen*STMax];} Resp;
+           char outbuff[CmsLocateRequest::RHLen*STMax];} Resp;
    struct iovec ioV[2] = {{(char *)&Arg.Request, sizeof(Arg.Request)},
                           {(char *)&Resp,        0}};
    const char *Why;
-   char theopts[8], *toP = theopts;
+   char eBuff[128], theopts[8], *toP = theopts;
+   XrdCmsCluster::CmsLSOpts lsopts = XrdCmsCluster::LS_NULL;
+   XrdNetIF::ifType ifType;
    int rc, bytes;
+   bool oksel = false, lsall = (*Arg.Path == '*');
 
-// Do a callout to the external manager if we have one
+// Get the right interface selection options
 //
-   if (Xmi_Select)
-      {XrdCmsReq Req(this, Arg.Request.streamid);
-       if (Xmi_Select->Select(&Req, XMI_LOCATE, Arg.Path, Arg.Opaque)) return 0;
+   ifType = ifVec[(Arg.Opts & CmsLocateRequest::kYR_retipmsk)
+                  >> CmsLocateRequest::kYR_retipsft];
+
+// Indicate whether we want a name or an actual address
+//
+   lsopts = (Arg.Opts & CmsLocateRequest::kYR_retname
+          ?  XrdCmsCluster::LS_IDNT : XrdCmsCluster::LS_IPO);
+
+// Indicate whether we can ignore network restrictions
+//
+   if (Arg.Opts & CmsLocateRequest::kYR_listall)
+      lsopts |= XrdCmsCluster::LS_ANY;
+
+// Handle private networks here
+//
+   if (Arg.Opts & CmsLocateRequest::kYR_prvtnet)
+      {XrdNetIF::Privatize(ifType);
+       *toP++='P';
       }
 
+// Encode if type into the options
+//
+   Sel.Opts = static_cast<int>(ifType) & XrdCmsSelect::ifWant;
+   lsopts   = static_cast<XrdCmsCluster::CmsLSOpts>(lsopts | ifType);
+
 // Grab the refresh option (the only one we support)
 //
    if (Arg.Opts & CmsLocateRequest::kYR_refresh) 
       {Sel.Opts  = XrdCmsSelect::Refresh; *toP++='s';}
    if (Arg.Opts & CmsLocateRequest::kYR_asap)
-      {Sel.Opts |= XrdCmsSelect::Asap;    *toP++='i'; Sel.InfoP = &reqInfo;}
+      {Sel.Opts |= XrdCmsSelect::Asap;    *toP++='i'; Sel.InfoP = &reqInfo;
+       reqInfo.lsLU = static_cast<char>(lsopts);
+      }
       else                                            Sel.InfoP = 0;
 
 // Do some debugging
@@ -541,10 +559,19 @@ const char *XrdCmsNode::do_Locate(XrdCmsRRData &Arg)
 // List the servers
 //
    if (!rc)
-      {if (!Sel.Vec.hf || !(sP=Cluster.List(Sel.Vec.hf,XrdCmsCluster::LS_IPO)))
-          {Arg.Request.rrCode = kYR_error;
-           rc = kYR_ENOENT; Why = "none ";
-           bytes = strlcpy(Resp.outbuff, "No servers have the file",
+      {if (!Sel.Vec.hf || !(sP=Cluster.List(Sel.Vec.hf, lsopts, oksel)))
+          {const char *eTxt;
+           Arg.Request.rrCode = kYR_error;
+           if (oksel)
+              {rc = kYR_ENETUNREACH; Why = "unreachable ";
+               sprintf(eBuff, "No servers are reachable via %s network",
+                       XrdNetIF::Name(ifType));
+               eTxt = eBuff;
+              } else {
+               rc = kYR_ENOENT; Why = "none ";
+               eTxt = "No servers have the file";
+              }
+           bytes = strlcpy(Resp.outbuff, eTxt,
                           sizeof(Resp.outbuff)) + sizeof(Resp.Val) + 1;
           } else rc = 0;
       }
@@ -555,7 +582,8 @@ const char *XrdCmsNode::do_Locate(XrdCmsRRData &Arg)
       {Resp.Val           = htonl(rc);
        DEBUGR(Why <<Arg.Path);
       } else {
-       bytes=do_LocFmt(Resp.outbuff,sP,Sel.Vec.pf,Sel.Vec.wf)+sizeof(Resp.Val)+1;
+       bytes = do_LocFmt(Resp.outbuff, sP, Sel.Vec.pf, Sel.Vec.wf, lsall)
+             + sizeof(Resp.Val) + 1;
        Resp.Val            = 0;
        Arg.Request.rrCode  = kYR_data;
       }
@@ -573,9 +601,11 @@ const char *XrdCmsNode::do_Locate(XrdCmsRRData &Arg)
 /******************************************************************************/
   
 int XrdCmsNode::do_LocFmt(char *buff, XrdCmsSelected *sP,
-                          SMask_t pfVec, SMask_t wfVec)
+                          SMask_t pfVec, SMask_t wfVec, bool lsall)
 {
    static const int Skip = (XrdCmsSelected::Disable | XrdCmsSelected::Offline);
+   static const int Hung = (XrdCmsSelected::Disable | XrdCmsSelected::Offline
+                         |  XrdCmsSelected::Suspend);
    XrdCmsSelected *pP;
    char *oP = buff;
 
@@ -583,12 +613,22 @@ int XrdCmsNode::do_LocFmt(char *buff, XrdCmsSelected *sP,
 // 01234567810123456789212345678
 // xy[::123.123.123.123]:123456
 //
+if (lsall)
+   while(sP)
+        {*oP = (sP->Status & XrdCmsSelected::isMangr ? 'M' : 'S');
+         if (sP->Status & Hung) *oP = tolower(*oP);
+         *(oP+1) = (sP->Mask   & wfVec               ? 'w' : 'r');
+         strcpy(oP+2, sP->Ident); oP += sP->IdentLen + 2;
+         if (sP->next) *oP++ = ' ';
+         pP = sP; sP = sP->next; delete pP;
+        }
+   else
    while(sP)
         {if (!(sP->Status & Skip))
             {*oP     = (sP->Status & XrdCmsSelected::isMangr ? 'M' : 'S');
              if (sP->Mask & pfVec) *oP = tolower(*oP);
              *(oP+1) = (sP->Mask   & wfVec                   ? 'w' : 'r');
-             strcpy(oP+2, sP->IPV6); oP += sP->IPV6Len + 2;
+             strcpy(oP+2, sP->Ident); oP += sP->IdentLen + 2;
              if (sP->next) *oP++ = ' ';
             }
          pP = sP; sP = sP->next; delete pP;
@@ -604,7 +644,7 @@ int XrdCmsNode::do_LocFmt(char *buff, XrdCmsSelected *sP,
 /*                              d o _ M k d i r                               */
 /******************************************************************************/
   
-// Mkdir requests are forwarded to all subscribers subject to an Xmi callout.
+// Mkdir requests are forwarded to all subscribers
 //
 const char *XrdCmsNode::do_Mkdir(XrdCmsRRData &Arg)
 {
@@ -616,14 +656,6 @@ const char *XrdCmsNode::do_Mkdir(XrdCmsRRData &Arg)
 //
    DEBUGR("mode " <<Arg.Mode <<' ' <<Arg.Path);
 
-// If we have an Xmi then call it
-//
-   if (Xmi_Mkdir)
-      {XrdCmsReq Req(this, Arg.Request.streamid);
-       if (!getMode(Arg.Mode, mode)) return "invalid mode";
-          else if (Xmi_Mkdir->Mkdir(&Req, mode, Arg.Path, Arg.Opaque)) return 0;
-      }
-
 // We are don here if we have no data; otherwise convert the mode if we
 // haven't done so already.
 //
@@ -644,7 +676,7 @@ const char *XrdCmsNode::do_Mkdir(XrdCmsRRData &Arg)
 /*                             d o _ M k p a t h                              */
 /******************************************************************************/
   
-// Mkpath requests are forwarded to all subscribers subjectto an Xmi callout.
+// Mkpath requests are forwarded to all subscribers
 //
 const char *XrdCmsNode::do_Mkpath(XrdCmsRRData &Arg)
 {
@@ -656,14 +688,6 @@ const char *XrdCmsNode::do_Mkpath(XrdCmsRRData &Arg)
 //
    DEBUGR("mode " <<Arg.Mode <<' ' <<Arg.Path);
 
-// If we have an Xmi then call it
-//
-   if (Xmi_Mkpath)
-      {XrdCmsReq Req(this, Arg.Request.streamid);
-       if (!getMode(Arg.Mode, mode)) return "invalid mode";
-          else if (Xmi_Mkpath->Mkpath(&Req,mode,Arg.Path,Arg.Opaque)) return 0;
-      }
-
 // We are don here if we have no data; otherwise convert the mode if we
 // haven't done so already.
 //
@@ -684,7 +708,7 @@ const char *XrdCmsNode::do_Mkpath(XrdCmsRRData &Arg)
 /*                                 d o _ M v                                  */
 /******************************************************************************/
   
-// Mv requests are forwarded to all subscribers subject to an Xmi callout.
+// Mv requests are forwarded to all subscribers
 //
 const char *XrdCmsNode::do_Mv(XrdCmsRRData &Arg)
 {
@@ -696,14 +720,6 @@ const char *XrdCmsNode::do_Mv(XrdCmsRRData &Arg)
 //
    DEBUGR(Arg.Path <<" to " <<Arg.Path2);
 
-// If we have an Xmi then call it
-//
-   if (Xmi_Rename)
-      {XrdCmsReq Req(this, Arg.Request.streamid, Arg.Request.modifier & kYR_dnf);
-       if (Xmi_Rename->Rename(&Req,Arg.Path,Arg.Opaque,Arg.Path2,Arg.Opaque2))
-          return 0;
-      }
-
 // If we are not a server, if must remove references to the old and new names
 // from our cache. This is independent of how the raname is handled. We need
 // not back percolate the mv since it was hanled top down in the first place.
@@ -757,6 +773,7 @@ const char *XrdCmsNode::do_Ping(XrdCmsRRData &Arg)
 // Process: ping
 // Respond: pong
 //
+   if (isBad & isDoomed) return ".redirected";
    Link->Send((char *)&pongIt, sizeof(pongIt));
    return 0;
 }
@@ -788,14 +805,6 @@ const char *XrdCmsNode::do_PrepAdd(XrdCmsRRData &Arg)
    DEBUGR("parms: " <<Arg.Reqid <<' ' <<Arg.Notify <<' ' <<Arg.Prty <<' '
                     <<Arg.Mode  <<' ' <<Arg.Path);
 
-// Do an Xmi callout if need be
-//
-   if (Xmi_Prep
-   &&  Xmi_Prep->Prep(Arg.Reqid,
-                      Arg.Opts & CmsPrepAddRequest::kYR_write ? XMI_RW : 0,
-                      Arg.Path, Arg.Opaque))
-       return 0;
-
 // Queue this request for async processing
 //
    (new XrdCmsPrepArgs(Arg))->Queue();
@@ -814,10 +823,6 @@ const char *XrdCmsNode::do_PrepDel(XrdCmsRRData &Arg)
 //
    DEBUGR("reqid " <<Arg.Reqid);
 
-// Do a callout to the external manager if we have one
-//
-   if (Xmi_Prep && Xmi_Prep->Prep(Arg.Reqid, XMI_CANCEL, "", 0)) return 0;
-
 // Cancel the request if applicable.
 //
    if (Config.DiskOK)
@@ -833,7 +838,7 @@ const char *XrdCmsNode::do_PrepDel(XrdCmsRRData &Arg)
 /*                                 d o _ R m                                  */
 /******************************************************************************/
   
-// Rm requests are forwarded to all subscribers subject to an Xmi callout.
+// Rm requests are forwarded to all subscribers
 //
 const char *XrdCmsNode::do_Rm(XrdCmsRRData &Arg)
 {
@@ -845,13 +850,6 @@ const char *XrdCmsNode::do_Rm(XrdCmsRRData &Arg)
 //
    DEBUGR(Arg.Path);
 
-// If we have an Xmi then call it
-//
-   if (Xmi_Remove)
-      {XrdCmsReq Req(this, Arg.Request.streamid, Arg.Request.modifier & kYR_dnf);
-       if (Xmi_Remove->Remove(&Req, Arg.Path, Arg.Opaque)) return 0;
-      }
-
 // If we have no data then we should remove this file from our cache
 //
    if (!Config.DiskOK)
@@ -875,7 +873,7 @@ const char *XrdCmsNode::do_Rm(XrdCmsRRData &Arg)
 /*                              d o _ R m d i r                               */
 /******************************************************************************/
   
-// Rmdir requests are forwarded to all subscribers subject to an Xmi callout.
+// Rmdir requests are forwarded to all subscribers
 //
 const char *XrdCmsNode::do_Rmdir(XrdCmsRRData &Arg)
 {
@@ -887,13 +885,6 @@ const char *XrdCmsNode::do_Rmdir(XrdCmsRRData &Arg)
 //
    DEBUGR(Arg.Path);
 
-// If we have an Xmi then call it
-//
-   if (Xmi_Remdir)
-      {XrdCmsReq Req(this, Arg.Request.streamid, Arg.Request.modifier & kYR_dnf);
-       if (Xmi_Remdir->Remdir(&Req, Arg.Path, Arg.Opaque)) return 0;
-      }
-
 // If we have no data then we should remove this directory from our cache
 //
    if (!Config.DiskOK)
@@ -927,26 +918,21 @@ const char *XrdCmsNode::do_Select(XrdCmsRRData &Arg)
    XrdCmsSelect Sel(XrdCmsSelect::Peers, Arg.Path, Arg.PathLen-1);
    struct iovec ioV[2];
    char theopts[16], *Avoid, *toP = theopts;
+   XrdNetIF::ifType ifType;
    int rc, bytes;
 
-// Do a callout to the external manager if we have one
-//
-   if (Arg.Opts & CmsSelectRequest::kYR_stat && Xmi_Stat)
-      {XrdCmsReq Req(this, Arg.Request.streamid);
-       if (Xmi_Stat->Stat(&Req, Arg.Path, Arg.Opaque)) return 0;
-      } else 
-   if (Xmi_Select)
-      {XrdCmsReq Req(this, Arg.Request.streamid);
-       int opts = (Arg.Opts & CmsSelectRequest::kYR_write ? XMI_RW : 0);
-       if (Arg.Opts & CmsSelectRequest::kYR_create) opts |= XMI_NEW;
-       if (Arg.Opts & CmsSelectRequest::kYR_trunc)  opts |= XMI_TRUNC;
-       if (Xmi_Select->Select(&Req, opts, Arg.Path, Arg.Opaque)) return 0;
-      }
-
 // Init select data (note that refresh supresses fast redirects)
 //
    Sel.iovP  = 0; Sel.iovN  = 0; Sel.InfoP = &reqInfo;
 
+// Determine what interface to return to the client
+//
+   ifType = ifVec[(Arg.Opts & CmsSelectRequest::kYR_retipmsk)
+                  >> CmsSelectRequest::kYR_retipsft];
+   if (Arg.Opts & CmsSelectRequest::kYR_prvtnet)
+      {XrdNetIF::Privatize(ifType);                                *toP++='P';}
+   Sel.Opts |= static_cast<int>(ifType) & XrdCmsSelect::ifWant;
+
 // Complete the arguments to select
 //
          if (Arg.Opts & CmsSelectRequest::kYR_refresh) 
@@ -973,14 +959,14 @@ const char *XrdCmsNode::do_Select(XrdCmsRRData &Arg)
 //
    Sel.nmask = SMask_t(0);
    if ((Avoid = Arg.Avoid))
-      {unsigned int IPaddr;
+      {XrdNetAddr avoidAddr;
        char *Comma;
        DEBUGR(theopts <<' ' <<Arg.Path <<" avoiding " <<Avoid);
        Sel.InfoP = 0;
        do {if ((Comma = index(Avoid,','))) *Comma = '\0';
            if (*Avoid == '+') Sel.nmask |= Cluster.getMask(Avoid+1);
-              else if (XrdSysDNS::Host2IP(Avoid, &IPaddr))
-                              Sel.nmask |= Cluster.getMask(IPaddr);
+              else if (!avoidAddr.Set(Avoid,0))
+                              Sel.nmask |= Cluster.getMask(&avoidAddr);
            Avoid = Comma+1;
           } while(Comma && *Avoid);
       } else DEBUGR(theopts <<' ' <<Arg.Path);
@@ -1033,13 +1019,6 @@ int XrdCmsNode::do_SelPrep(XrdCmsPrepArgs &Arg) // Static!!!
    XrdCmsSelect Sel(XrdCmsSelect::Peers, Arg.path, Arg.pathlen-1);
    int rc;
 
-// Do a callout to the external manager if we have one
-//
-   if (Xmi_Prep)
-      {int opts = (Arg.options & CmsPrepAddRequest::kYR_write ? XMI_RW : 0);
-       if (Xmi_Prep->Prep(Arg.reqid, opts, Arg.path, Arg.opaque)) return 0;
-      }
-
 // Complete the arguments to select
 //
    if ( Arg.options & CmsPrepAddRequest::kYR_fresh)
@@ -1057,6 +1036,11 @@ int XrdCmsNode::do_SelPrep(XrdCmsPrepArgs &Arg) // Static!!!
    Sel.InfoP = 0;  // No fast redirects
    Sel.nmask = SMask_t(0);
 
+// We do not care what interface is being used. This may conflict with a
+// staging prepare but it's too complicated to handle at this point.
+//
+   Sel.Opts |= static_cast<char>(XrdNetIF::ifAny);
+
 // Check if co-location wanted relevant only when staging wanted
 //
    if (Arg.clPath && Sel.iovP)
@@ -1121,7 +1105,8 @@ const char *XrdCmsNode::do_Space(XrdCmsRRData &Arg)
 
 // Send the response
 //
-   if (Arg.Request.rrCode != kYR_space) Manager.Inform(mySpace.Hdr, buff, blen);
+   if (Arg.Request.rrCode != kYR_space)
+      XrdCmsManager::Inform(mySpace.Hdr, buff, blen);
       else {xmsg[0].iov_base = (char *)&mySpace;
             xmsg[0].iov_len  = sizeof(mySpace);
             xmsg[1].iov_base = buff;
@@ -1228,10 +1213,10 @@ void XrdCmsNode::do_StateDFS(XrdCmsBaseFR *rP, int rc)
 // point we will only inform the manager that actually wants to know. This
 // is encoded to the route passed to us.
 //
-   if (Manager.Present() && isNew
+   if (XrdCmsManager::Present() && isNew
    && !(rP->Mod & CmsStateRequest::kYR_noresp))
       {Request.rrCode   = kYR_have;
-       Manager.Inform(Request, rP->Path, rP->PathLen+1);
+       XrdCmsManager::Inform(Request, rP->Path, rP->PathLen+1);
       }
 }
   
@@ -1427,7 +1412,7 @@ const char *XrdCmsNode::do_Status(XrdCmsRRData &Arg)
    int Resume  = Arg.Request.modifier & CmsStatusRequest::kYR_Resume;
    int Suspend = Arg.Request.modifier & CmsStatusRequest::kYR_Suspend;
    int Reset   = Arg.Request.modifier & CmsStatusRequest::kYR_Reset;
-   int add2Activ, add2Stage;
+   int add2Activ, add2Stage, port;
 
 // Do some debugging
 //
@@ -1438,7 +1423,7 @@ const char *XrdCmsNode::do_Status(XrdCmsRRData &Arg)
 // Process reset requests. These are exclsuive to any other request
 //
    if (Reset)
-      {Manager.Reset();                // Propagate the reset to our managers
+      {XrdCmsManager::Reset();         // Propagate the reset to our managers
        Cache.Bounce(NodeMask, NodeID); // Now invalidate our cache lines
       }
 
@@ -1451,23 +1436,27 @@ const char *XrdCmsNode::do_Status(XrdCmsRRData &Arg)
 
 // Process suspend/resume
 //
-    if ((Resume && isSuspend) || (Suspend && !isSuspend))
-       if (Suspend) {add2Activ = -1; isSuspend = 1;
+    if ((Resume && (isBad & isSuspend)) || (Suspend && !(isBad & isSuspend)))
+       if (Suspend) {add2Activ = -1; isBad |=  isSuspend;
                      srvMsg="service suspended"; 
                      stgMsg = 0;
                     }
-          else      {add2Activ =  1; isSuspend = 0;
+          else      {add2Activ =  1; isBad &= ~isSuspend;
                      srvMsg="service resumed";
                      stgMsg = (isNoStage ? "(no staging)" : "(staging)");
-                     Port = ntohl(Arg.Request.streamid);
-                     DEBUGR("set data port to " <<Port);
+                     port = ntohl(Arg.Request.streamid);
+                     if (port && port != netIF.Port())
+                        {Lock(); netIF.Port(port); UnLock();
+                         DEBUGR("set data port to " <<port);
+                        }
                     }
        else         {add2Activ =  0; srvMsg = 0;}
 
 // Get the most important message out
 //
-   if (isOffline)         {srvMsg = "service offline";  stgMsg = 0;}
-      else if (isDisable) {srvMsg = "service disabled"; stgMsg = 0;}
+        if (isOffline)          {srvMsg = "service offline";     stgMsg = 0;}
+   else if (isBad & isDisabled) {srvMsg = "service disabled";    stgMsg = 0;}
+   else if (isBad & isBlisted ) {srvMsg = "service blacklisted"; stgMsg = 0;}
 
 // Now see if we need to change anything
 //
@@ -1483,8 +1472,7 @@ const char *XrdCmsNode::do_Status(XrdCmsRRData &Arg)
 /*                              d o _ T r u n c                               */
 /******************************************************************************/
   
-// Trunc requests are forwarded to all subscribers subject to an Xmi callout.
-// Currently, we have no definition in Xmi for trunc!
+// Trunc requests are forwarded to all subscribers
 //
 const char *XrdCmsNode::do_Trunc(XrdCmsRRData &Arg)
 {
@@ -1496,14 +1484,6 @@ const char *XrdCmsNode::do_Trunc(XrdCmsRRData &Arg)
 //
    DEBUGR("size " <<Arg.Mode <<' ' <<Arg.Path);
 
-// If we have an Xmi then call it
-//
-// if (Xmi_Trunc)
-//    {XrdCmsReq Req(this, Arg.Request.streamid);
-//     if (!getSize(Arg.Mode, Size)) return "invalid size";
-//        else if (Xmi_Trunc->Trunc(&Req, Size, Arg.Path, Arg.Opaque)) return 0;
-//    }
-
 // We are don here if we have no data; otherwise convert the mode if we
 // haven't done so already.
 //
@@ -1530,22 +1510,14 @@ const char *XrdCmsNode::do_Trunc(XrdCmsRRData &Arg)
 const char *XrdCmsNode::do_Try(XrdCmsRRData &Arg)
 {
    EPNAME("do_Try")
-   XrdOucTokenizer theList(Arg.Path);
-   char *tp;
 
 // Do somde debugging
 //
    DEBUGR(Arg.Path);
 
-// Delete any additions from this manager
-//
-   myMans.Del(IPAddr);
-
 // Add all the alternates to our alternate list
 //
-   tp = theList.GetLine();
-   while((tp = theList.GetToken()))
-         myMans.Add(IPAddr, tp, Config.PortTCP, myLevel);
+   if (Manager) Manager->myMans->Add(&netID, Arg.Path, Config.PortTCP, myLevel);
 
 // Close the link and return an error
 //
@@ -1616,7 +1588,7 @@ void XrdCmsNode::Report_Usage(XrdLink *lp)   // Static!
    xmsg[1].iov_base = respbuff;
    xmsg[1].iov_len  = blen;
    if (lp) lp->Send(xmsg, 2);
-      else Manager.Inform("usage", xmsg, 2);
+      else XrdCmsManager::Inform("usage", xmsg, 2);
 
 // Do some debugging
 //
