@@ -298,8 +298,8 @@ int XrdFrmAdmin::Find()
 /******************************************************************************/
 
 const char *XrdFrmAdmin::HelpHelp =
-"[help] {audit | chksum | exit | f[ind] | makelf | mark | mmap | pin | q[uery] "
-         "| quit | reloc | rm} ...";
+"[help] {audit | chksum | exit | f[ind] | makelf | mark | mmap | mv | pin "
+         "| q[uery] | quit | reloc | rm} ...";
   
 int XrdFrmAdmin::Help()
 {
@@ -314,6 +314,7 @@ int XrdFrmAdmin::Help()
                              {"makelf", 6, 6, MakeLFHelp},
                              {"mark",   4, 4, MarkHelp  },
                              {"mmap",   4, 4, MmapHelp  },
+                             {"mv",     2, 2, MvHelp    },
                              {"pin",    3, 3, PinHelp   },
                              {"query",  1, 5, QueryHelp },
                              {"reloc",  5, 5, RelocHelp },
@@ -481,6 +482,32 @@ int XrdFrmAdmin::Mmap()
    return 0;
 }
 
+/******************************************************************************/
+/*                                    M v                                     */
+/******************************************************************************/
+
+const char *XrdFrmAdmin::MvHelp = "mv oldlfn newlfn";
+
+int XrdFrmAdmin::Mv()
+{
+   static XrdOucArgs Spec(&Say, "frm_admin: ", "", (const char *)0);
+
+   static const char *Reqs[] = {"oldlfn", "newlfn", 0};
+
+   int rc;
+
+// Parse the request and do it
+//
+   if (!Parse("mv ", Spec, Reqs)) return 1;
+
+// Simply invoke the reloc function in the underlying FS
+//
+   if ((rc = Config.ossFS->Rename(Opt.Args[0], Opt.Args[1])))
+      Emsg(-rc, "rename ", Opt.Args[0]);
+      else Msg(Opt.Args[0], " renamed to ", Opt.Args[1]);
+   return rc != 0;
+}
+
 /******************************************************************************/
 /*                                   P i n                                    */
 /******************************************************************************/
@@ -691,6 +718,7 @@ int XrdFrmAdmin::xeqArgs(char *Cmd)
                              {"makelf", 6, 6, &XrdFrmAdmin::MakeLF},
                              {"mark",   1, 4, &XrdFrmAdmin::Mark},
                              {"mmap",   1, 4, &XrdFrmAdmin::Mmap},
+                             {"mv",     2, 2, &XrdFrmAdmin::Mv},
                              {"pin",    3, 3, &XrdFrmAdmin::Pin},
                              {"query",  1, 5, &XrdFrmAdmin::Query},
                              {"quit",   4, 4, &XrdFrmAdmin::Quit},
@@ -836,7 +864,7 @@ int XrdFrmAdmin::Parse(const char *What, XrdOucArgs &Spec, const char **Reqs)
 
 // Now process all the options
 //
-   while((theOpt = Spec.getopt()) != -1)
+   while((theOpt = Spec.getopt()) != (char)-1)
         {switch(theOpt)
                {case 'A': Opt.All     = 1; break;
                 case 'e': Opt.Erase   = 1; break;
@@ -936,7 +964,7 @@ int XrdFrmAdmin::ParseOwner(const char *What, char *Uname)
 // Process username
 //
    if (Uname)
-      {if (*Uname >= 0 && *Uname <= 9)
+      {if (*Uname >= '0' && *Uname <= '9')
           {if (XrdOuca2x::a2i(Say,"uid",Uname, &Unum)) return 0;
            Opt.Uid = Unum;
           }
@@ -949,7 +977,7 @@ int XrdFrmAdmin::ParseOwner(const char *What, char *Uname)
 // Process groupname
 //
    if (Gname)
-      {if (*Gname >= 0 && *Gname <= 9)
+      {if (*Gname >= '0' && *Gname <= '9')
           {if (XrdOuca2x::a2i(Say, "gid", Gname, &Gnum))  return 0;
            Opt.Gid = Gnum;
           }
