@@ -56,11 +56,23 @@
 #include "XrdNet/XrdNetConnect.hh"
 #include "XrdNet/XrdNetOpts.hh"
 #include "XrdNet/XrdNetSocket.hh"
+#include "XrdNet/XrdNetUtils.hh"
 #include "XrdOuc/XrdOucUtils.hh"
-#include "XrdSys/XrdSysDNS.hh"
 #include "XrdSys/XrdSysError.hh"
+#include "XrdSys/XrdSysFD.hh"
 #include "XrdSys/XrdSysPlatform.hh"
 
+/******************************************************************************/
+/*                               G l o b a l s                                */
+/******************************************************************************/
+  
+namespace XrdNetSocketCFG
+{
+       int ka_Idle = 0;
+       int ka_Itvl = 0;
+       int ka_Icnt = 0;
+};
+
 /******************************************************************************/
 /*                         l o c a l   d e f i n e s                          */
 /******************************************************************************/
@@ -69,6 +81,7 @@
 #define XRDNETSOCKET_LINGER    3
 
 #define Err(p,a,b,c) (ErrCode = (eroute ? eroute->Emsg(#p, a, b, c) : ErrCode),-1)
+#define ErrM(p,a,b,c) (ErrCode = (eroute ? eroute->Emsg(#p, a, b, c) : ErrCode),-1)
 
 /******************************************************************************/
 /*                           C o n s t r u c t o r                            */
@@ -77,9 +90,8 @@
 XrdNetSocket::XrdNetSocket(XrdSysError *erobj, int SockFileDesc)
 {
    ErrCode  = 0;
-   PeerName = 0;
-   SockFD   = SockFileDesc;
    eroute   = erobj;
+   SockFD   = SockFileDesc;
 }
 
 /******************************************************************************/
@@ -102,7 +114,7 @@ int XrdNetSocket::Accept(int timeout)
        if (!sfd.revents) return -1;
       }
 
-   do {ClientSock = accept(SockFD, (struct sockaddr *)0, 0);}
+   do {ClientSock = XrdSysFD_Accept(SockFD, (struct sockaddr *)0, 0);}
       while(ClientSock < 0 && errno == EINTR);
 
    if (ClientSock < 0 && eroute) eroute->Emsg("Accept",errno,"accept connection");
@@ -122,10 +134,6 @@ void XrdNetSocket::Close()
      //
      if (SockFD >= 0) {close(SockFD); SockFD=-1;} 
 
-     // Free any previous peername pointer
-     //
-     if (PeerName) {free(PeerName); PeerName = 0;}
-
      // Reset values and return.
      //
      ErrCode=0;
@@ -158,8 +166,9 @@ XrdNetSocket *XrdNetSocket::Create(XrdSysError *Say, const char *path,
    if (opts & XRDNET_FIFO)
       {if ((ASock->SockFD = mkfifo(fnbuff, mode)) < 0 && errno != EEXIST)
          {eMsg = "create fifo"; rc = errno;}
-         else if ((ASock->SockFD = open(fnbuff, O_RDWR, myMode)) < 0)
-                 {eMsg = "open fifo"; rc = ASock->LastError();}
+         else if ((ASock->SockFD = XrdSysFD_Open(fnbuff, O_RDWR, myMode)) < 0)
+                 {eMsg = "open fifo"; rc = errno;}
+                 else if (opts & XRDNET_NOCLOSEX) XrdSysFD_Yield(ASock->SockFD);
       } else if (ASock->Open(fnbuff, -1, sflags) < 0) 
                 {eMsg = "create socket"; rc = ASock->LastError();}
 #else
@@ -182,7 +191,6 @@ XrdNetSocket *XrdNetSocket::Create(XrdSysError *Say, const char *path,
 int XrdNetSocket::Detach()
 {  int oldFD = SockFD;
    SockFD = -1;
-   if (PeerName) {free(PeerName); PeerName = 0;}
    return oldFD;
 }
 
@@ -207,21 +215,19 @@ int XrdNetSocket::getWindow(int fd, int &Windowsz, XrdSysError *eDest)
   
 int XrdNetSocket::Open(const char *inpath, int port, int flags, int windowsz)
 {
-   struct sockaddr_in InetAddr;
-   struct sockaddr_un UnixAddr;
-   struct sockaddr *SockAddr;
-   char *errtxt = 0, pbuff[128];
-   const char *action = "configure socket";
-   const char *path = (inpath ? inpath : "");
-   const char *epath= (inpath ? inpath : pbuff);
-   int myEC, SockSize, backlog;
+   const char *epath, *eText, *action = "configure socket";
+   char pbuff[128];
+   int myEC, backlog, SockProt;
    int SockType = (flags & XRDNET_UDPSOCKET ? SOCK_DGRAM : SOCK_STREAM);
    const int one = 1;
    const SOCKLEN_t szone = (SOCKLEN_t)sizeof(one);
 
 // Supply actual port number in error messages
 //
-   if (!inpath) sprintf(pbuff, "port %d", port);
+   if (inpath) epath = inpath;
+      else {sprintf(pbuff, "port %d", port);
+            epath = pbuff;
+           }
 
 // Make sure this object is available for a new socket
 //
@@ -231,86 +237,74 @@ int XrdNetSocket::Open(const char *inpath, int port, int flags, int windowsz)
 //
    myEC = ErrCode = 0;
 
-// Allocate a socket descriptor and bind connection address, if requested.
-// In most OS's we must unlink the corresponding Unix path name or bind will
-// fail. In some OS's, this creates a problem (e.g., Solaris) since the
-// file inode is used to identify the socket and will likely change. This
-// means that connects occuring before the bind will hang up to 3 minutes.
-// So, the client better be prepared to timeout connects and try again.
+// Preset out address information
 //
-   if (port < 0 && *path == '/')
-      {if (strlen(path) >= sizeof(UnixAddr.sun_path))
-          return Err(Open, ENAMETOOLONG, "create unix socket ", epath);
-       if ((SockFD = socket(PF_UNIX, SockType, 0)) < 0)
-          return Err(Open, errno, "create unix socket ", epath);
-       UnixAddr.sun_family = AF_UNIX;
-       strcpy(UnixAddr.sun_path, path);
-       SockAddr = (struct sockaddr *)&UnixAddr;
-       SockSize = sizeof(UnixAddr);
-       if (flags & XRDNET_SERVER) unlink((const char *)path);
-      } else {
-       if ((SockFD = socket(PF_INET, SockType, 0)) < 0)
-          return Err(Open, errno, "create inet socket to", epath);
-       if (port < 0 && *path)
-                     XrdSysDNS::Host2Dest(inpath,(sockaddr &)InetAddr,&errtxt);
-             else   {XrdSysDNS::getHostAddr(path,(sockaddr &)InetAddr,&errtxt);
-                     XrdSysDNS::setPort((sockaddr &)InetAddr, port);
-                    }
-          if (errtxt)
-             {if(eroute) eroute->Emsg("Open", "Unable to obtain address for",
-                                     epath, errtxt);
-              Close();
-              ErrCode = EHOSTUNREACH;
-              return -1;
-             }
-       SockAddr = (struct sockaddr *)&InetAddr;
-       SockSize = sizeof(InetAddr);
+   if ((eText = SockInfo.Set(inpath,(port < 0 ? XrdNetAddr::PortInSpec:port))))
+      {ErrCode = EHOSTUNREACH;
+       if (eroute)
+          {char buff[256];
+           snprintf(buff, sizeof(buff), "'; %s", eText);
+           eroute->Emsg("Open", "Unable to create socket for '", epath, buff);
+          }
+       return -1;
       }
 
-// Set the options and window size, as needed (ignore errors)
+// Allocate a socket descriptor of the right type
+//
+   SockProt = SockInfo.Protocol();
+   if ((SockFD = XrdSysFD_Socket(SockProt, SockType, 0)) < 0)
+      return Err(Open, errno, "create socket for", epath);
+
+// Based on the type socket, set appropriate options. For server-side Unix
+// sockets we must unlink the corresponding Unix path name or bind will fail.
+// In some OS's, this creates a problem (e.g., Solaris) since the file inode is
+// used to identify the socket and will likely change. This means that connects
+// occuring before the bind will hang up to 3 minutes and client needs to retry.
+// For non-Unix socketsr be prepared to timeout connects and try again.
+//
+   if (SockProt == PF_UNIX)
+      {setOpts(SockFD, flags | XRDNET_UDPSOCKET, eroute);
+       if (flags & XRDNET_SERVER) unlink((const char *)inpath);
+      } else {
+       setOpts(SockFD, flags, eroute);
+       if (setsockopt(SockFD,SOL_SOCKET,SO_REUSEADDR, (Sokdata_t)&one, szone)
+       &&  eroute) eroute->Emsg("Open",errno,"set socket REUSEADDR for",epath);
+      }
+
+// Set the window size or udp buffer size, as needed (ignore errors)
 //
-   setOpts(SockFD, (flags | (*path == '/' ? XRDNET_UDPSOCKET : 0)), eroute);
    if (windowsz) setWindow(SockFD, windowsz, eroute);
 
-// Make sure the local endpoint can be reused
-//
-   if ((*path != '/') && setsockopt(SockFD,SOL_SOCKET,SO_REUSEADDR,
-                                   (Sokdata_t)&one, szone) && eroute)
-      eroute->Emsg("open", errno, "set socket REUSEADDR");
-
-// Either do a connect or a bind.
+// Either do a bind or a connect.
 //
    if (flags & XRDNET_SERVER)
       {action = "bind socket to";
-       if ( bind(SockFD, SockAddr, SockSize) ) myEC = errno;
+       if (bind(SockFD, SockInfo.SockAddr(), SockInfo.SockSize())) myEC = errno;
           else if (SockType == SOCK_STREAM)
                   {action = "listen on stream";
                    if (!(backlog = flags & XRDNET_BKLG))
                       backlog = XRDNETSOCKET_MAXBKLG;
                    if (listen(SockFD, backlog)) myEC = errno;
                   }
-       if (*path == '/') chmod(path, S_IRWXU);
+       if (SockProt == PF_UNIX) chmod(inpath, S_IRWXU);
       } else {
        if (SockType == SOCK_STREAM)
           {int tmo = flags & XRDNET_TOUT;
            action = "connect socket to";
-           if (tmo) myEC = XrdNetConnect::Connect(SockFD,SockAddr,SockSize,tmo);
-              else if (connect(SockFD, SockAddr, SockSize)) myEC = errno;
+           if (tmo) myEC = XrdNetConnect::Connect(SockFD, SockInfo.SockAddr(),
+                                                  SockInfo.SockSize(),tmo);
+              else if (connect(SockFD,SockInfo.SockAddr(),SockInfo.SockSize()))
+                      myEC = errno;
           }
-       if (!myEC) {PeerName = strdup((path ? path : "?"));
-                   if (*path == '/') XrdSysDNS::getHostAddr(0, PeerAddr);
-                      else memcpy((void *)&PeerAddr,SockAddr,sizeof(PeerAddr));
-                  }
       }
 
-// Check for any errors and return.
+// Check for any errors and return (Close() sets SockFD to -1).
 //
    if (myEC)
       {Close(); 
        ErrCode = myEC;
        if (!(flags & XRDNET_NOEMSG) && eroute)
           eroute->Emsg("Open", ErrCode, action, epath);
-       return -1;
       }
    return SockFD;
 }
@@ -319,9 +313,10 @@ int XrdNetSocket::Open(const char *inpath, int port, int flags, int windowsz)
 /*                              P e e r n a m e                               */
 /******************************************************************************/
   
-const char *XrdNetSocket::Peername(struct sockaddr **InetAddr)
+const char *XrdNetSocket::Peername(const struct sockaddr **InetAddr,
+                                                     int  *InetSize)
 {
-   char *errtxt;
+   const char *errtxt, *PeerName;
 
 // Make sure  we have something to look at
 //
@@ -333,16 +328,16 @@ const char *XrdNetSocket::Peername(struct sockaddr **InetAddr)
 
 // Get the host name on the other side of this socket
 //
-   if (!PeerName 
-   &&  !(PeerName = XrdSysDNS::Peername(SockFD, &PeerAddr, &errtxt)))
+   if (!(PeerName = SockInfo.Name(0, &errtxt)))
       {if (eroute) 
-          eroute->Emsg("Peername", "Unable to obtain peer name;",errtxt);
+          eroute->Emsg("Peername", "Unable to obtain peer name; ",errtxt);
        ErrCode = ESRCH;
       }
 
-// Return possible address and the name
+// Return possible address, length and the name
 //
-     if (InetAddr) *InetAddr = &PeerAddr;
+     if (InetAddr) *InetAddr = SockInfo.SockAddr();
+     if (InetSize) *InetSize = SockInfo.SockSize();
      return PeerName;
 }
 
@@ -354,12 +349,13 @@ int XrdNetSocket::setOpts(int xfd, int opts, XrdSysError *eDest)
 {
    int rc = 0;
    const int one = 1;
+   const int szint = sizeof(int);
    const SOCKLEN_t szone = (SOCKLEN_t)sizeof(one);
-   static int tcpprotid = XrdSysDNS::getProtoID("tcp");
+   static int tcpprotid = XrdNetUtils::ProtoID("tcp");
    static struct linger liopts = {1, XRDNETSOCKET_LINGER};
    const SOCKLEN_t szlio = (SOCKLEN_t)sizeof(liopts);
 
-   if (!(opts & XRDNET_NOCLOSEX) && fcntl(xfd, F_SETFD, FD_CLOEXEC))
+   if (opts & XRDNET_NOCLOSEX && !XrdSysFD_Yield(xfd))
       {rc = 1;
        if (eDest) eDest->Emsg("setOpts", errno, "set fd close on exec");
       }
@@ -372,10 +368,30 @@ int XrdNetSocket::setOpts(int xfd, int opts, XrdSysError *eDest)
        if (eDest) eDest->Emsg("setOpts", errno, "set socket LINGER");
       }
 
-   if ((opts & XRDNET_KEEPALIVE)
-   &&  setsockopt(xfd,SOL_SOCKET,SO_KEEPALIVE,(Sokdata_t)&one,szone))
-      {rc = 1;
-       if (eDest) eDest->Emsg("setOpts", errno, "set socket KEEPALIVE");
+   if (opts & XRDNET_KEEPALIVE)
+      {if (setsockopt(xfd,SOL_SOCKET,SO_KEEPALIVE,(Sokdata_t)&one,szone))
+          {rc = 1;
+           if (eDest) eDest->Emsg("setOpts", errno, "set socket KEEPALIVE");
+          }
+#ifdef __linux__
+           else if (opts & XRDNET_SERVER) // Following are inherited in Linux
+      {if (XrdNetSocketCFG::ka_Idle
+       &&  setsockopt(xfd,SOL_TCP,TCP_KEEPIDLE,&XrdNetSocketCFG::ka_Idle,szint))
+          {rc = 1;
+           if (eDest) eDest->Emsg("setOpts", errno, "set socket KEEPIDLE");
+          }
+       if (XrdNetSocketCFG::ka_Itvl
+       &&  setsockopt(xfd,SOL_TCP,TCP_KEEPINTVL,&XrdNetSocketCFG::ka_Itvl,szint))
+          {rc = 1;
+           if (eDest) eDest->Emsg("setOpts", errno, "set socket KEEPINTVL");
+          }
+       if (XrdNetSocketCFG::ka_Icnt
+       &&  setsockopt(xfd,SOL_TCP,TCP_KEEPCNT,  &XrdNetSocketCFG::ka_Icnt,szint))
+          {rc = 1;
+           if (eDest) eDest->Emsg("setOpts", errno, "set socket KEEPCNT");
+          }
+      }
+#endif
       }
 
    if (!(opts & XRDNET_DELAY)
@@ -410,49 +426,6 @@ int XrdNetSocket::setWindow(int xfd, int Windowsz, XrdSysError *eDest)
    return rc;
 }
 
-/******************************************************************************/
-/*                            s o c k e t A d d r                             */
-/******************************************************************************/
-  
-const char *XrdNetSocket::socketAddr(XrdSysError *Say, const char *dest,
-                                     struct sockaddr **sockAP, int &sockAL)
-{
-   struct sockaddr_in InetAddr;
-   struct sockaddr_un UnixAddr;
-   struct sockaddr *SockAddr;
-   char *errtxt = 0;
-   int port = 0, SockSize;
-
-   if (*dest == '/')
-      {if (strlen(dest) >= sizeof(UnixAddr.sun_path))
-          {if (Say) Say->Emsg("Net", ENAMETOOLONG, "generate addr from", dest);
-           return "socket address path too long";
-          }
-       UnixAddr.sun_family = AF_UNIX;
-       strcpy(UnixAddr.sun_path, dest);
-       SockAddr = (struct sockaddr *)&UnixAddr;
-       SockSize = sizeof(UnixAddr);
-      } else {
-       if (*dest && *dest != ':')
-                     XrdSysDNS::Host2Dest(dest,(sockaddr &)InetAddr,&errtxt);
-          else      {XrdSysDNS::getHostAddr(dest,(sockaddr &)InetAddr,&errtxt);
-                     if (*dest && *dest == ':') port = atoi(dest+1);
-                     XrdSysDNS::setPort((sockaddr &)InetAddr, port);
-                    }
-       if (errtxt)
-          {if (Say) Say->Emsg("Net","Unable to obtain address for",dest,errtxt);
-           return errtxt;
-          }
-       SockAddr = (struct sockaddr *)&InetAddr;
-       SockSize = sizeof(InetAddr);
-      }
-
-    sockAL = SockSize;
-   *sockAP = (struct sockaddr *)malloc(SockSize);
-   memcpy(*sockAP, SockAddr, SockSize);
-   return 0;
-}
-
 /******************************************************************************/
 /*                            s o c k e t P a t h                             */
 /******************************************************************************/
