@@ -46,6 +46,10 @@
 #include "XrdCks/XrdCksConfig.hh"
 #include "XrdCks/XrdCksData.hh"
 
+#include "XrdNet/XrdNetAddr.hh"
+#include "XrdNet/XrdNetIF.hh"
+#include "XrdNet/XrdNetUtils.hh"
+
 #include "XrdOfs/XrdOfs.hh"
 #include "XrdOfs/XrdOfsEvs.hh"
 #include "XrdOfs/XrdOfsHandle.hh"
@@ -59,7 +63,6 @@
 
 #include "XrdOss/XrdOss.hh"
 
-#include "XrdSys/XrdSysDNS.hh"
 #include "XrdSys/XrdSysError.hh"
 #include "XrdSys/XrdSysHeaders.hh"
 #include "XrdSys/XrdSysLogger.hh"
@@ -76,6 +79,7 @@
 #include "XrdOuc/XrdOucTrace.hh"
 #include "XrdSec/XrdSecEntity.hh"
 #include "XrdSfs/XrdSfsAio.hh"
+#include "XrdSfs/XrdSfsFlags.hh"
 #include "XrdSfs/XrdSfsInterface.hh"
 
 #ifdef AIX
@@ -123,23 +127,17 @@ XrdOss *XrdOfsOss;
 
 XrdOfs::XrdOfs()
 {
-   unsigned int myIPaddr = 0;
-   char buff[256], *bp;
-   int i;
+   const char *bp;
 
 // Establish defaults
 //
-   AuthLib       = 0;
-   AuthParm      = 0;
+   ofsConfig     = 0;
    Authorization = 0;
-   CmsLib        = 0;
-   CmsParms      = 0;
-   OssLib        = 0;
-   OssParms      = 0;
    Finder        = 0;
    Balancer      = 0;
    evsObject     = 0;
    myRole        = strdup("server");
+   ossRW         =' ';
 
 // Obtain port number we will be using. Note that the constructor must occur
 // after the port number is known (i.e., this cannot be a global static).
@@ -153,20 +151,6 @@ XrdOfs::XrdOfs()
    poscHold= 10*60;
    poscAuto= 0;
 
-// Establish our hostname and IPV4 address
-//
-   HostName      = XrdSysDNS::getHostName();
-   if (!XrdSysDNS::Host2IP(HostName, &myIPaddr)) myIPaddr = 0x7f000001;
-   strcpy(buff, "[::"); bp = buff+3;
-   bp += XrdSysDNS::IP2String(myIPaddr, 0, bp, 128);
-   *bp++ = ']'; *bp++ = ':';
-   sprintf(bp, "%d", myPort);
-   locResp = strdup(buff); locRlen = strlen(buff);
-   for (i = 0; HostName[i] && HostName[i] != '.'; i++);
-   HostName[i] = '\0';
-   HostPref = strdup(HostName);
-   HostName[i] = '.';
-
 // Set the configuration file name and dummy handle
 //
    ConfigFN = 0;
@@ -175,8 +159,7 @@ XrdOfs::XrdOfs()
 // Set checksum pointers
 //
    Cks       = 0;
-   CksConfig = 0;
-   CksRdsz   = 0;
+   CksPfn    = true;
 }
   
 /******************************************************************************/
@@ -353,6 +336,46 @@ int XrdOfsDirectory::close()
    return retc;
 }
 
+/******************************************************************************/
+/*                              a u t o S t a t                               */
+/******************************************************************************/
+
+int XrdOfsDirectory::autoStat(struct stat *buf)
+/*
+  Function: Set stat buffer to automaticaly return stat information
+
+  Input:    Pointer to stat buffer which will be filled in on each
+            nextEntry() and represent stat information for that entry.
+
+  Output:   Upon success, returns zero. Upon error returns SFS_ERROR and sets
+            the error object to contain the reason.
+
+  Notes: 1. If autoStat() is not supported he caller will need to follow up
+            with a manual stat() call for the full path, a slow and tedious
+            process. The autoStat function significantly reduces overhead by
+            automatically providing stat information for the entry read.
+*/
+{
+   EPNAME("autoStat");
+   int retc;
+
+// Check if this directory is actually open
+//
+   if (!dp) {XrdOfsFS->Emsg(epname, error, EBADF, "autostat directory");
+             return SFS_ERROR;
+            }
+
+// Set the stat buffer in the storage system directory.
+//
+    if ((retc = dp->StatRet(buf)))
+       retc = XrdOfsFS->Emsg(epname, error, retc, "autostat", fname);
+       else retc = SFS_OK;
+
+// All done
+//
+   return retc;
+}
+  
 /******************************************************************************/
 /*                                                                            */
 /*                F i l e   O b j e c t   I n t e r f a c e s                 */
@@ -407,7 +430,8 @@ int XrdOfsFile::open(const char          *path,      // In
                        : Path(path), hP(0), fP(0), poscNum(0) {}
 
                        ~OpenHelper()
-                       {if (hP) hP->Retire();
+                       {int retc;
+                        if (hP) hP->Retire(retc);
                         if (fP) delete fP;
                         if (poscNum > 0) XrdOfsFS->poscQ->Del(Path, poscNum, 1);
                        }
@@ -599,14 +623,14 @@ int XrdOfsFile::open(const char          *path,      // In
 // Verify that we can actually use this file
 //
    if (oP.poscNum > 0)
-      {if ((retc = oP.fP->Fchmod(static_cast<mode_t>(theMode | S_ISUID))))
+      {if ((retc = oP.fP->Fchmod(static_cast<mode_t>(theMode|XRDSFS_POSCPEND))))
           return XrdOfsFS->Emsg(epname, error, retc, "fchmod", path);
        XrdOfsFS->poscQ->Commit(path, oP.poscNum);
       }
 
 // Set compression values and activate the handle
 //
-   if (oP.fP->isCompressed())
+   if (oP.fP->isCompressed() > 0)
       {oP.hP->isCompressed = 1;
        dorawio = (open_mode & SFS_O_RAWIO ? 1 : 0);
       }
@@ -656,7 +680,7 @@ int XrdOfsFile::close()  // In
    static XrdOfsHanCB *hCB = static_cast<XrdOfsHanCB *>(new CloseFH);
 
    XrdOfsHandle *hP;
-   int   poscNum, retc;
+   int   poscNum, retc, cRetc = 0;
    short theMode;
 
 // Trace the call
@@ -691,12 +715,14 @@ int XrdOfsFile::close()  // In
 
 // If this file was tagged as a POSC then we need to make sure it will persist
 // Note that we unpersist the file immediately when it's inactive or if no hold
-// time is allowed.  `Also, close events occur only for active handles.
+// time is allowed. Also, close events occur only for active handles. If the
+// entry was via delete then we ignore the close return code as there is no
+// one to handle it on the other side.
 //
    if ((poscNum = hP->PoscGet(theMode, !viaDel)))
       {if (viaDel)
           {if (hP->Inactive() || !XrdOfsFS->poscHold)
-              {XrdOfsFS->Unpersist(hP, !hP->Inactive()); hP->Retire();}
+              {XrdOfsFS->Unpersist(hP, !hP->Inactive()); hP->Retire(cRetc);}
               else hP->Retire(hCB, XrdOfsFS->poscHold);
            return SFS_OK;
           }
@@ -715,21 +741,21 @@ int XrdOfsFile::close()  // In
 //
    if (XrdOfsFS->evsObject && tident
    &&  XrdOfsFS->evsObject->Enabled(hP->isRW ? XrdOfsEvs::Closew
-                                            : XrdOfsEvs::Closer))
+                                             : XrdOfsEvs::Closer))
       {long long FSize, *retsz;
        char pathbuff[MAXPATHLEN+8];
        XrdOfsEvs::Event theEvent;
        if (hP->isRW) {theEvent = XrdOfsEvs::Closew; retsz = &FSize;}
           else {      theEvent = XrdOfsEvs::Closer; retsz = 0; FSize=0;}
-       if (!(hP->Retire(retsz, pathbuff, sizeof(pathbuff))))
+       if (!(hP->Retire(cRetc, retsz, pathbuff, sizeof(pathbuff))))
           {XrdOfsEvsInfo evInfo(tident, pathbuff, "" , 0, 0, FSize);
            XrdOfsFS->evsObject->Notify(theEvent, evInfo);
-          } else hP->Retire();
-      } else     hP->Retire();
+          }
+      } else hP->Retire(cRetc);
 
 // All done
 //
-    return SFS_OK;
+  return (cRetc ? XrdOfsFS->Emsg(epname, error, cRetc, "close file") : SFS_OK);
 }
 
 /******************************************************************************/
@@ -749,10 +775,41 @@ int            XrdOfsFile::fctl(const int               cmd,
 
 // We don't support this
 //
-   out_error.setErrInfo(EEXIST, "fctl operation not supported");
+   out_error.setErrInfo(ENOTSUP, "fctl operation not supported");
    return SFS_ERROR;
 }
 
+/******************************************************************************/
+
+int            XrdOfsFile::fctl(const int               cmd,
+                                      int               alen,
+                                const char             *args,
+                                const XrdSecEntity     *client)
+{                             // 12345678901234
+   static const char *fctlArg = "ofs.tpc cancel";
+   static const int   fctlAsz = 15;
+
+// See if the is a tpc cancellation (the only thing we support here)
+//
+   if (cmd != SFS_FCTL_SPEC1 || !args || alen < fctlAsz || strcmp(fctlArg,args))
+      {error.setErrInfo(ENOTSUP, "fctl operation not supported");
+       return SFS_ERROR;
+      }
+
+// Check if we have a tpc operation in progress
+//
+   if (!myTPC)
+      {error.setErrInfo(ESRCH, "tpc operation not found");
+       return SFS_ERROR;
+      }
+
+// Cancel the tpc
+//
+   myTPC->Del();
+   myTPC = 0;
+   return SFS_OK;
+}
+
 /******************************************************************************/
 /*                                  r e a d                                   */
 /******************************************************************************/
@@ -839,6 +896,35 @@ XrdSfsXferSize XrdOfsFile::read(XrdSfsFileOffset  offset,    // In
 //
    return nbytes;
 }
+
+/******************************************************************************/
+/*                                  r e a d v                                 */
+/******************************************************************************/
+
+XrdSfsXferSize XrdOfsFile::readv(XrdOucIOVec     *readV,     // In
+                                 int              readCount) // In
+/*
+  Function: Perform all the reads specified in the readV vector.
+
+  Input:    readV     - A description of the reads to perform; includes the
+                        absolute offset, the size of the read, and the buffer
+                        to place the data into.
+            readCount - The size of the readV vector.
+
+  Output:   Returns the number of bytes read upon success and SFS_ERROR o/w.
+            If the number of bytes read is less than requested, it is considered
+            an error.
+*/
+{
+   EPNAME("readv");
+
+   XrdSfsXferSize nbytes = oh->Select().ReadV(readV, readCount);
+   if (nbytes < 0)
+       return XrdOfsFS->Emsg(epname, error, (int)nbytes, "readv", oh->Name());
+
+   return nbytes;
+
+}
   
 /******************************************************************************/
 /*                              r e a d   A I O                               */
@@ -1273,7 +1359,7 @@ int XrdOfs::chksum(      csFunc            Func,   // In
 
 // At this point we need to convert the lfn to a pfn
 //
-   if (!(Path = XrdOfsOss->Lfn2Pfn(Path, buff, MAXPATHLEN, rc)))
+   if (CksPfn && !(Path = XrdOfsOss->Lfn2Pfn(Path, buff, MAXPATHLEN, rc)))
       return Emsg(epname, einfo, rc, "checksum", Path);
 
 // Now determine what to do
@@ -1434,6 +1520,7 @@ int XrdOfs::fsctl(const int               cmd,
 
   Input:    cmd       - Operation command (currently supported):
                         SFS_FSCTL_LOCATE - locate file
+                        SFS_FSCTL_STATCC - return cluster config status
                         SFS_FSCTL_STATFS - return file system info (physical)
                         SFS_FSCTL_STATLS - return file system info (logical)
                         SFS_FSCTL_STATXA - return file extended attributes
@@ -1457,8 +1544,7 @@ int XrdOfs::fsctl(const int               cmd,
                                'w'};
    static const int PrivNum = sizeof(PrivLet);
 
-   int retc, find_flag = SFS_O_LOCATE | (cmd & (SFS_O_NOWAIT | SFS_O_RESET));
-   int i, blen, privs, opcode = cmd & SFS_FSCTL_CMD;
+   int retc, i, blen, privs, opcode = cmd & SFS_FSCTL_CMD;
    const char *tident = einfo.getErrUser();
    char *bP, *cP;
    XTRACE(fsctl, args, "");
@@ -1467,23 +1553,41 @@ int XrdOfs::fsctl(const int               cmd,
 //
    if (opcode == SFS_FSCTL_LOCATE)
       {struct stat fstat;
-       char pbuff[1024], rType[3], *Resp[] = {rType, locResp};
+       char pbuff[1024], rType[3];
+       const char *Resp[2] = {rType, pbuff};
        const char *locArg, *opq, *Path = Split(args,&opq,pbuff,sizeof(pbuff));
-            if (*Path == '*')      {locArg = Path; Path++;}
-       else if (cmd & SFS_O_TRUNC) {locArg = (char *)"*";}
-       else                         locArg = Path;
+       XrdNetIF::ifType ifType;
+       int Resp1Len;
+       int find_flag = SFS_O_LOCATE
+                     | (cmd&(SFS_O_FORCE|SFS_O_NOWAIT|SFS_O_RESET|SFS_O_HNAME));
        XrdOucEnv loc_Env(opq ? opq+1 : 0,0,client);
-       AUTHORIZE(client,0,AOP_Stat,"locate",Path,einfo);
+
+       if (cmd & SFS_O_TRUNC)           locArg = (char *)"*";
+          else {     if (*Path == '*') {locArg = Path; Path++;}
+                        else            locArg = Path;
+                AUTHORIZE(client,0,AOP_Stat,"locate",Path,einfo);
+               }
        if (Finder && Finder->isRemote()
        &&  (retc = Finder->Locate(einfo, locArg, find_flag, &loc_Env)))
           return fsError(einfo, retc);
-       if ((retc = XrdOfsOss->Stat(Path, &fstat, 0, &loc_Env)))
-          return XrdOfsFS->Emsg(epname, einfo, retc, "locate", Path);
-       rType[0] = ((fstat.st_mode & S_IFBLK) == S_IFBLK ? 's' : 'S');
-       rType[1] =  (fstat.st_mode & S_IWUSR             ? 'w' : 'r');
+
+       if (cmd & SFS_O_TRUNC) {rType[0] = 'S'; rType[1] = ossRW;}
+          else {if ((retc = XrdOfsOss->Stat(Path, &fstat, 0, &loc_Env)))
+                   return XrdOfsFS->Emsg(epname, einfo, retc, "locate", Path);
+                rType[0] = ((fstat.st_mode & S_IFBLK) == S_IFBLK ? 's' : 'S');
+                rType[1] =  (fstat.st_mode & S_IWUSR             ? 'w' : 'r');
+               }
        rType[2] = '\0';
-       einfo.setErrInfo(locRlen+3, (const char **)Resp, 2);
-       return SFS_DATA;
+
+       ifType = XrdNetIF::GetIFType((einfo.getUCap() & XrdOucEI::uIPv4)  != 0,
+                                    (einfo.getUCap() & XrdOucEI::uIPv64) != 0,
+                                    (einfo.getUCap() & XrdOucEI::uPrip)  != 0);
+       bool retHN = (cmd & SFS_O_HNAME) != 0;
+       if ((Resp1Len = myIF->GetDest(pbuff, sizeof(pbuff), ifType, retHN)))
+           {einfo.setErrInfo(Resp1Len+3, (const char **)Resp, 2);
+            return SFS_DATA;
+           }
+       return Emsg(epname, einfo, ENETUNREACH, "locate", Path);
       }
 
 // Process the STATFS request
@@ -1546,6 +1650,18 @@ int XrdOfs::fsctl(const int               cmd,
        return SFS_DATA;
       }
 
+// Process the STATCC request (this should always succeed)
+//
+   if (opcode == SFS_FSCTL_STATCC)
+      {static const int lcc_flag = SFS_O_LOCATE | SFS_O_LOCAL;
+       XrdOucEnv lcc_Env(0,0,client);
+            if (Finder)   retc = Finder  ->Locate(einfo,".",lcc_flag,&lcc_Env);
+       else if (Balancer) retc = Balancer->Locate(einfo,".",lcc_flag,&lcc_Env);
+       else retc = SFS_ERROR;
+       if (retc != SFS_DATA) einfo.setErrInfo(5, "none|");
+       return fsError(einfo, SFS_DATA);
+      }
+
 // Operation is not supported
 //
    return XrdOfsFS->Emsg(epname, einfo, ENOTSUP, "fsctl", args);
@@ -1637,7 +1753,15 @@ int XrdOfs::mkdir(const char             *path,    // In
 
 // If we have a redirector, tell it that we now have this path
 //
-   if (Balancer) Balancer->Added(path);
+   if (Balancer)
+      {if (!mkpath) Balancer->Added(path);
+          else     {char *slash, *myPath = strdup(path);
+                    do {Balancer->Added(myPath);
+                        if ((slash = rindex(myPath, '/'))) *slash = 0;
+                       } while(slash && slash != myPath);
+                    free(myPath);
+                   }
+      }
 
     return SFS_OK;
 }
@@ -2181,7 +2305,7 @@ void XrdOfs::Unpersist(XrdOfsHandle *oh, int xcev)
 
 char *XrdOfs::WaitTime(int stime, char *buff, int blen)
 {
-   int mlen, hr, min, sec;
+   int hr, min, sec;
 
 // Compute hours, minutes, and seconds
 //
@@ -2193,17 +2317,17 @@ char *XrdOfs::WaitTime(int stime, char *buff, int blen)
 // Now format the message based on time duration
 //
         if (!hr && !min)
-           mlen = snprintf(buff,blen,"%d second%s",sec,(sec > 1 ? "s" : ""));
+           snprintf(buff,blen,"%d second%s",sec,(sec > 1 ? "s" : ""));
    else if (!hr)
           {if (sec > 10) min++;
-           mlen = snprintf(buff,blen,"%d minute%s",min,(min > 1 ? "s" : ""));
+           snprintf(buff,blen,"%d minute%s",min,(min > 1 ? "s" : ""));
           }
    else   {if (hr == 1)
               if (min <= 30)
-                      mlen = snprintf(buff,blen,"%d minutes",min+60);
-                 else mlen = snprintf(buff,blen,"%d hour and %d minutes",hr,min);
+                      snprintf(buff,blen,"%d minutes",min+60);
+                 else snprintf(buff,blen,"%d hour and %d minutes",hr,min);
               else {if (min > 30) hr++;
-                      mlen = snprintf(buff,blen,"%d hours",hr);
+                      snprintf(buff,blen,"%d hours",hr);
                    }
           }
 
