@@ -43,6 +43,7 @@
 #include "XrdVersion.hh"
 
 #include "XrdFrc/XrdFrcProxy.hh"
+#include "XrdOss/XrdOssPath.hh"
 #include "XrdOss/XrdOssApi.hh"
 #include "XrdOss/XrdOssCache.hh"
 #include "XrdOss/XrdOssConfig.hh"
@@ -57,10 +58,10 @@
 #include "XrdOuc/XrdOucExport.hh"
 #include "XrdOuc/XrdOucMsubs.hh"
 #include "XrdOuc/XrdOucN2NLoader.hh"
+#include "XrdOuc/XrdOucPinLoader.hh"
 #include "XrdOuc/XrdOucProg.hh"
 #include "XrdOuc/XrdOucStream.hh"
 #include "XrdOuc/XrdOucUtils.hh"
-#include "XrdSys/XrdSysFAttr.hh"
 #include "XrdSys/XrdSysHeaders.hh"
 #include "XrdSys/XrdSysPthread.hh"
 #include "XrdSys/XrdSysPlatform.hh"
@@ -133,12 +134,6 @@ const char *XrdOssErrorText[] =
 
 #define xrdmax(a,b)       (a < b ? b : a)
 
-// Set the following value to establish the ulimit for FD numbers. Zero
-// sets it to whatever the current hard limit is. Negative leaves it alone.
-//
-#define XrdOssFDLIMIT     -1
-#define XrdOssFDMINLIM    64
-
 /******************************************************************************/
 /*            E x t e r n a l   T h r e a d   I n t e r f a c e s             */
 /******************************************************************************/
@@ -186,7 +181,7 @@ XrdOssSys::XrdOssSys()
    RemoteRoot    = 0;
    cscanint      = 600;
    FDFence       = -1;
-   FDLimit       = XrdOssFDLIMIT;
+   FDLimit       = -1;
    MaxSize       = 0;
    minalloc      = 0;
    ovhalloc      = 0;
@@ -209,13 +204,27 @@ XrdOssSys::XrdOssSys()
    xfrFdln       = 0;
    RSSProg       = 0;
    StageProg     = 0;
+   prPBits       = (long long)sysconf(_SC_PAGESIZE);
+   prPSize       = static_cast<int>(prPBits);
+   prPBits--;
+   prPMask       = ~prPBits;
+   prBytes       = 0;
+   prActive      = 0;
+   prDepth       = 0;
+   prQSize       = 0;
+   STT_Lib       = 0;
+   STT_Parms     = 0;
+   STT_Func      = 0;
+   STT_PreOp     = 0;
+   STT_V2        = 0;
 }
   
 /******************************************************************************/
 /*                             C o n f i g u r e                              */
 /******************************************************************************/
   
-int XrdOssSys::Configure(const char *configfn, XrdSysError &Eroute)
+int XrdOssSys::Configure(const char *configfn, XrdSysError &Eroute,
+                                               XrdOucEnv   *envP)
 {
 /*
   Function: Establish default values using a configuration file.
@@ -226,6 +235,8 @@ int XrdOssSys::Configure(const char *configfn, XrdSysError &Eroute)
 */
    XrdSysError_Table *ETab = new XrdSysError_Table(XRDOSS_EBASE, XRDOSS_ELAST,
                                                    XrdOssErrorText);
+   static const int maxFD = 1048576;
+   struct rlimit rlim;
    char *val;
    int  retc, NoGo = XrdOssOK;
    pthread_t tid;
@@ -240,17 +251,33 @@ int XrdOssSys::Configure(const char *configfn, XrdSysError &Eroute)
 //
    ConfigFN = (configfn && *configfn ? strdup(configfn) : 0);
 
+// Establish the FD limit and the fence (half way)
+//
+  if (getrlimit(RLIMIT_NOFILE, &rlim))
+     Eroute.Emsg("Config", errno, "get fd limit");
+     else {if (rlim.rlim_max == RLIM_INFINITY) rlim.rlim_max = maxFD;
+           if (rlim.rlim_cur != rlim.rlim_max)
+              {rlim.rlim_cur  = rlim.rlim_max;
+               if (setrlimit(RLIMIT_NOFILE, &rlim))
+                  Eroute.Emsg("Config", errno, "set fd limit");
+                  else FDLimit = rlim.rlim_cur;
+              } else {FDFence = static_cast<int>(rlim.rlim_cur)>>1;
+                      FDLimit = rlim.rlim_cur;
+                     }
+
 // Process the configuration file
 //
    NoGo = ConfigProc(Eroute);
 
+// Configure dependent plugins
+//
+   if (!NoGo)
+      {if (N2N_Lib || LocalRoot || RemoteRoot) NoGo |= ConfigN2N(Eroute, envP);
+       if (STT_Lib && !NoGo) NoGo |= ConfigStatLib(Eroute, envP);
+      }
+
 // Establish the FD limit
 //
-   {struct rlimit rlim;
-    if (getrlimit(RLIMIT_NOFILE, &rlim) < 0)
-       Eroute.Emsg("Config", errno, "get resource limits");
-       else Hard_FD_Limit = rlim.rlim_max;
-
     if (FDLimit <= 0) FDLimit = rlim.rlim_cur;
        else {rlim.rlim_cur = FDLimit;
             if (setrlimit(RLIMIT_NOFILE, &rlim) < 0)
@@ -287,9 +314,12 @@ int XrdOssSys::Configure(const char *configfn, XrdSysError &Eroute)
 //
    ConfigSpace();
 
-// Configure extended attributes (really not much to do here)
-//
-   XrdSysFAttr::Msg(&Eroute);
+// Set the prefix for files in cache file systems  
+   if ( OptFlags & XrdOss_CacheFS ) 
+       if (!NoGo) {
+           NoGo = XrdOssPath::InitPrefix();
+           if (NoGo) Eroute.Emsg("Config", "cache file prefix initialization failed");
+       }
 
 // Configure statiscal reporting
 //
@@ -466,7 +496,7 @@ void XrdOssSys::ConfigMio(XrdSysError &Eroute)
 /*                             C o n f i g N 2 N                              */
 /******************************************************************************/
 
-int XrdOssSys::ConfigN2N(XrdSysError &Eroute)
+int XrdOssSys::ConfigN2N(XrdSysError &Eroute, XrdOucEnv *envP)
 {
    XrdOucN2NLoader n2nLoader(&Eroute,ConfigFN,N2N_Parms,LocalRoot,RemoteRoot);
 
@@ -521,10 +551,6 @@ int XrdOssSys::ConfigProc(XrdSysError &Eroute)
                  &&  xpath(Config, Eroute)) {Config.Echo(); NoGo = 1;}
         }
 
-// All done scanning the file, set dependent parameters.
-//
-   if (N2N_Lib || LocalRoot || RemoteRoot) NoGo |= ConfigN2N(Eroute);
-
 // Now check if any errors occured during file i/o
 //
    if ((retc = Config.LastError()))
@@ -803,6 +829,32 @@ int XrdOssSys::ConfigStageC(XrdSysError &Eroute)
    return NoGo;
 }
 
+/******************************************************************************/
+/*                         C o n f i g S t a t L i b                          */
+/******************************************************************************/
+
+int XrdOssSys::ConfigStatLib(XrdSysError &Eroute, XrdOucEnv *envP)
+{
+   XrdOucPinLoader myLib(&Eroute, myVersion, "statlib", STT_Lib);
+
+// Get the plugin and stat function
+//
+   if (STT_V2)
+      {XrdOssStatInfoInit2_t siGet2;
+       if (!(siGet2=(XrdOssStatInfoInit2_t)myLib.Resolve("XrdOssStatInfoInit2"))
+       ||  !(STT_Fund = siGet2(this,Eroute.logger(),ConfigFN,STT_Parms,envP)))
+          return 1;
+      } else {
+       XrdOssStatInfoInit_t siGet;
+       if (!(siGet = (XrdOssStatInfoInit_t)myLib.Resolve("XrdOssStatInfoInit"))
+       ||  !(STT_Func = siGet (this,Eroute.logger(),ConfigFN,STT_Parms)))
+          return 1;
+      }
+
+// Return success
+//
+   return 0;
+}
   
 /******************************************************************************/
 /*                           C o n f i g S t a t s                            */
@@ -897,8 +949,10 @@ int XrdOssSys::ConfigXeq(char *var, XrdOucStream &Config, XrdSysError &Eroute)
    TS_Xeq("memfile",       xmemf);
    TS_Xeq("namelib",       xnml);
    TS_Xeq("path",          xpath);
+   TS_Xeq("preread",       xprerd);
    TS_Xeq("space",         xspace);
    TS_Xeq("stagecmd",      xstg);
+   TS_Xeq("statlib",       xstl);
    TS_Xeq("trace",         xtrace);
    TS_Xeq("usage",         xusage);
    TS_Xeq("xfr",           xxfr);
@@ -1093,27 +1147,12 @@ int XrdOssSys::xdefault(XrdOucStream &Config, XrdSysError &Eroute)
 
 int XrdOssSys::xfdlimit(XrdOucStream &Config, XrdSysError &Eroute)
 {
-    char *val;
-    int fence = 0, fdmax = XrdOssFDLIMIT;
 
-      if (!(val = Config.GetWord()))
-         {Eroute.Emsg("Config", "fdlimit fence not specified"); return 1;}
+    while(Config.GetWord()) {}
 
-      if (!strcmp(val, "*")) fence = -1;
-         else if (XrdOuca2x::a2i(Eroute,"fdlimit fence",val,&fence,0)) return 1;
+    Eroute.Say("Config warning: ", "fdlimit directive no longer supported.");
 
-      if (!(val = Config.GetWord())) fdmax = -1;
-         else if (!strcmp(val, "max")) fdmax = Hard_FD_Limit;
-                 else if (XrdOuca2x::a2i(Eroute, "fdlimit value", val, &fdmax,
-                              xrdmax(fence,XrdOssFDMINLIM))) return -EINVAL;
-                         else if (fdmax > Hard_FD_Limit)
-                                 {fdmax = Hard_FD_Limit;
-                                  Eroute.Say("Config warning: ",
-                                              "'fdlimit' forced to hard max");
-                                 }
-      FDFence = fence;
-      FDLimit = fdmax;
-      return 0;
+    return 0;
 }
   
 /******************************************************************************/
@@ -1277,12 +1316,94 @@ int XrdOssSys::xnml(XrdOucStream &Config, XrdSysError &Eroute)
 
 int XrdOssSys::xpath(XrdOucStream &Config, XrdSysError &Eroute)
 {
+   XrdOucPList *pP;
 
 // Parse the arguments
 //
-   return (XrdOucExport::ParsePath(Config, Eroute, RPList, DirFlags) ? 0 : 1);
+   pP = XrdOucExport::ParsePath(Config, Eroute, RPList, DirFlags);
+   if (!pP) return 1;
+
+// This plugin does not support relative paths so check for this
+//
+   if (*(pP->Path()) == '/') return 0;
+   Eroute.Emsg("Config", "Unsupported export -", pP->Path());
+   return 1;
 }
 
+/******************************************************************************/
+/*                                x p r e r d                                 */
+/******************************************************************************/
+
+/* Function: xprerd
+
+   Purpose:  To parse the directive: preread {<depth> | on} [limit <bytes>]
+                                             [ qsize [=]<qsz> ]
+
+             <depth>  the number of request to preread ahead of the read.
+                      A value of 0, the inital default, turns off prereads.
+                      Specifying "on" sets the value (currently) to 3.
+             <bytes>  Maximum number of bytes to preread. Prereading stops,
+                      regardless of depth, once <bytes> have been preread.
+                      The default is 1M (i.e.1 megabyte). The max is 16M.
+             <qsz>    the queue size after which preread blocking would occur.
+                      The value must be greater than or equal to <depth>.
+                      The value is adjusted to max(<qsz>/(<depth>/2+1),<depth>)
+                      unless the number is preceeded by an equal sign. The
+                      default <qsz> is 128.
+
+   Output: 0 upon success or !0 upon failure.
+*/
+
+int XrdOssSys::xprerd(XrdOucStream &Config, XrdSysError &Eroute)
+{
+    static const long long m16 = 16777216LL;
+    char *val;
+    long long lim = 1048576;
+    int depth, qeq = 0, qsz = 128;
+
+      if (!(val = Config.GetWord()))
+         {Eroute.Emsg("Config", "preread depth not specified"); return 1;}
+
+      if (!strcmp(val, "on")) depth = 3;
+         else if (XrdOuca2x::a2i(Eroute,"preread depth",val,&depth,0, 1024))
+                 return 1;
+
+      while((val = Config.GetWord()))
+           {     if (!strcmp(val, "limit"))
+                    {if (!(val = Config.GetWord()))
+                        {Eroute.Emsg("Config","preread limit not specified");
+                         return 1;
+                        }
+                     if (XrdOuca2x::a2sz(Eroute,"preread limit",val,&lim,0,m16))
+                        return 1;
+                    }
+            else if (!strcmp(val, "qsize"))
+                    {if (!(val = Config.GetWord()))
+                        {Eroute.Emsg("Config","preread qsize not specified");
+                         return 1;
+                        }
+                     if (XrdOuca2x::a2i(Eroute,"preread qsize",val,&qsz,0,1024))
+                        return 1;
+                     if (qsz < depth)
+                        {Eroute.Emsg("Config","preread qsize must be >= depth");
+                         return 1;
+                        }
+                    }
+            else {Eroute.Emsg("Config","invalid preread option -",val); return 1;}
+         }
+
+      if (lim < prPSize || !qsz) depth = 0;
+      if (!qeq && depth)
+         {qsz = qsz/(depth/2+1);
+          if (qsz < depth) qsz = depth;
+         }
+
+      prDepth = depth;
+      prQSize = qsz;
+      prBytes = lim;
+      return 0;
+}
+  
 /******************************************************************************/
 /*                                x s p a c e                                 */
 /******************************************************************************/
@@ -1453,7 +1574,57 @@ int XrdOssSys::xstg(XrdOucStream &Config, XrdSysError &Eroute)
    return 0;
 }
 
-  
+/******************************************************************************/
+/*                                  x s t l                                   */
+/******************************************************************************/
+
+/* Function: xstl
+
+   Purpose:  To parse the directive: statlib [-2] [preopen] <path> [<parms>]
+
+             preopen   issue the stat() prior to opening the file.
+             <path>    the path of the stat library to be used.
+             <parms>   optional parms to be passed
+
+  Output: 0 upon success or !0 upon failure.
+*/
+
+int XrdOssSys::xstl(XrdOucStream &Config, XrdSysError &Eroute)
+{
+    char *val, parms[1040];
+
+// Get the path or preopen option
+//
+   if (!(val = Config.GetWord()) || !val[0])
+      {Eroute.Emsg("Config", "statlib not specified"); return 1;}
+
+// Check for options
+//
+   STT_V2 = 0; STT_PreOp = 0;
+do{     if (!strcmp(val, "-2"))      STT_V2    = 1;
+   else if (!strcmp(val, "preopen")) STT_PreOp = 1;
+   else break;
+  } while((val = Config.GetWord()) && val[0]);
+
+// Make sure we have a statlib
+//
+   if (!val || !(*val))
+      {Eroute.Emsg("Config", "statlib not specified"); return 1;}
+
+// Record the path
+//
+   if (STT_Lib) free(STT_Lib);
+   STT_Lib = strdup(val);
+
+// Record any parms
+//
+   if (!Config.GetRest(parms, sizeof(parms)))
+      {Eroute.Emsg("Config", "statlib parameters too long"); return 1;}
+   if (STT_Parms) free(STT_Parms);
+   STT_Parms = (*parms ? strdup(parms) : 0);
+   return 0;
+}
+
 /******************************************************************************/
 /*                                x t r a c e                                 */
 /******************************************************************************/
