@@ -42,12 +42,24 @@
 #include <sys/stat.h>
 #include <sys/types.h>
 #endif
-#include "XrdSys/XrdSysDNS.hh"
+#include "XrdNet/XrdNetUtils.hh"
+#include "XrdOuc/XrdOucEnv.hh"
+#include "XrdOuc/XrdOucUtils.hh"
 #include "XrdSys/XrdSysError.hh"
 #include "XrdSys/XrdSysPlatform.hh"
 #include "XrdOuc/XrdOucStream.hh"
-#include "XrdOuc/XrdOucUtils.hh"
   
+/******************************************************************************/
+/*                              e n d s W i t h                               */
+/******************************************************************************/
+  
+bool XrdOucUtils::endsWith(const char *text, const char *ending, int endlen)
+{
+   int tlen = strlen(text);
+
+   return (tlen >= endlen && !strcmp(text+(tlen-endlen), ending));
+}
+
 /******************************************************************************/
 /*                                 e T e x t                                  */
 /******************************************************************************/
@@ -77,8 +89,11 @@ char *XrdOucUtils::eText(int rc, char *eBuff, int eBlen, int AsIs)
 /*                                  d o I f                                   */
 /******************************************************************************/
   
-// doIf() parses "if [<hostlist>] [<altopt> [&& <altop> [ ... ]]]"
-// altop: [exec <pgmlist> [&& named <namelist>]] | [named <namelist>]
+// doIf() parses "if [<hostlist>] [<conds>]"
+// conds: <cond1> | <cond2> | <cond3>
+// cond1: defined <varlist> [&& <conds>]
+// cond2: exec <pgmlist> [&& <cond3>]
+// cond3: named <namelist>
 
 // Returning 1 if true (i.e., this machine is one of the named hosts in hostlist 
 // and is running one of the programs pgmlist and named by one of the names in 
@@ -91,9 +106,10 @@ int XrdOucUtils::doIf(XrdSysError *eDest, XrdOucStream &Config,
                       const char *what,  const char *hname,
                       const char *nname, const char *pname)
 {
-   static const char *brk[] = {"exec", "named", 0};
+   static const char *brk[] = {"defined", "exec", "named", 0};
+   XrdOucEnv *theEnv = 0;
    char *val;
-   int hostok;
+   int hostok, isDef;
 
 // Make sure that at least one thing appears after the if
 //
@@ -105,7 +121,7 @@ int XrdOucUtils::doIf(XrdSysError *eDest, XrdOucStream &Config,
 // Check if we are one of the listed hosts
 //
    if (!is1of(val, brk))
-      {do {hostok = XrdSysDNS::isMatch(hname, val);
+      {do {hostok = XrdNetUtils::Match(hname, val);
            val = Config.GetWord();
           } while(!hostok && val && !is1of(val, brk));
       if (hostok)
@@ -115,6 +131,45 @@ int XrdOucUtils::doIf(XrdSysError *eDest, XrdOucStream &Config,
          } else return 0;
       }
 
+// Check if this is a defined test
+//
+   while(!strcmp(val, "defined"))
+      {if (!(val = Config.GetWord()) || *val != '?')
+          {if (eDest)
+              eDest->Emsg("Config","'?var' missing after 'defined' in",what);
+              return -1;
+          }
+       // Get environment if we have none
+       //
+          if (!theEnv && (theEnv = Config.SetEnv(0))) Config.SetEnv(theEnv);
+          if (!theEnv && *(val+1) != '~') return 0;
+
+       // Check if any listed variable is defined.
+       //
+       isDef = 0;
+       while(val && *val == '?')
+            {if (*(val+1) == '~' ? getenv(val+2) : theEnv->Get(val+1)) isDef=1;
+             val = Config.GetWord();
+            }
+       if (!val || !isDef) return isDef;
+       if (strcmp(val, "&&"))
+          {if (eDest)
+              eDest->Emsg("Config",val,"is invalid for defined test in",what);
+              return -1;
+          } else {
+           if (!(val = Config.GetWord()))
+              {if (eDest)
+                   eDest->Emsg("Config","missing keyword after '&&' in",what);
+                   return -1;
+              }
+          }
+       if (!is1of(val, brk))
+          {if (eDest)
+              eDest->Emsg("Config",val,"is invalid after '&&' in",what);
+              return -1;
+          }
+      }
+
 // Check if we need to compare program names (we are here only if we either
 // passed the hostlist test or there was no hostlist present)
 //
@@ -280,7 +335,7 @@ char *XrdOucUtils::Ident(long long  &mySID, char *iBuff, int iBlen,
    const char *sP;
    char sName[64], uName[256];
    long long urSID;
-   int  n, myPid;
+   int  myPid;
 
 // Generate our server ID
 //
@@ -391,6 +446,30 @@ int XrdOucUtils::makePath(char *path, mode_t mode)
    return 0;
 }
  
+/******************************************************************************/
+/*                                R e L i n k                                 */
+/******************************************************************************/
+
+int XrdOucUtils::ReLink(const char *path, const char *target, mode_t mode)
+{
+   const mode_t AMode = S_IRWXU|S_IRWXG; // 770
+   char pbuff[MAXPATHLEN+64];
+   int n;
+
+// Copy the path
+//
+   n = strlen(path);
+   if (n >= (int)sizeof(pbuff)) return ENAMETOOLONG;
+   strcpy(pbuff, path);
+
+// Unlink the target, make the path, and create the symlink
+//
+   unlink(path);
+   makePath(pbuff, (mode ? mode : AMode));
+   if (symlink(target, path)) return errno;
+   return 0;
+}
+
 /******************************************************************************/
 /*                              s u b L o g f n                               */
 /******************************************************************************/
