@@ -29,11 +29,10 @@
 /******************************************************************************/
 
 #include <errno.h>
-#include <fcntl.h>
-#include <unistd.h>
 #include <stdio.h>
 #include <stdlib.h>
   
+#include "XrdSys/XrdSysFD.hh"
 #include "XrdSys/XrdSysIOEvents.hh"
 #include "XrdSys/XrdSysHeaders.hh"
 #include "XrdSys/XrdSysPlatform.hh"
@@ -79,9 +78,9 @@
 
 #define TRACE_LOK " channel now " <<(isLocked ? "locked" : "unlocked")
 
-#define TRACE_MOD(x,fd,y,z) \
+#define TRACE_MOD(x,fd,y) \
                  IF_TRACE(x,fd,"Modify(" <<y <<") == " \
-                 <<BOOLNAME(retval) <<" [prior events=" <<z <<']' <<TRACE_LOK)
+                 <<BOOLNAME(retval) <<TRACE_LOK)
 
 #define TRACE_NOD(x,fd,y) \
                  IF_TRACE(x,fd,"Modify(" <<y <<") skipped; no events changed")
@@ -105,12 +104,13 @@ namespace XrdSys
 namespace IOEvents
 {
 struct pollArg
-       {Poller         *pollP;
-        const char     *retMsg;
-        int             retCode;
-        XrdSysSemaphore pollSync;
+       {Poller          *pollP;
+        const char      *retMsg;
+        int              retCode;
+        XrdSysSemaphore *pollSync;
 
-        pollArg() : retMsg(0), retCode(0), pollSync(0, "poll sync") {}
+        pollArg() : retMsg(0), retCode(0)
+                  {pollSync = new XrdSysSemaphore(0, "poll sync");}
        ~pollArg() {}
        };
 
@@ -125,9 +125,11 @@ void *BootStrap::Start(void *parg)
 {
    struct pollArg *pollArg = (struct pollArg *)parg;
    Poller *thePoller = pollArg->pollP;
+   XrdSysSemaphore *theSem = pollArg->pollSync;
    thePoller->pollTid = XrdSysThread::ID();
 
-   thePoller->Begin(&(pollArg->pollSync), pollArg->retCode, &(pollArg->retMsg));
+   thePoller->Begin(theSem, pollArg->retCode, &(pollArg->retMsg));
+   delete theSem;
 
    return (void *)0;
 }
@@ -273,20 +275,22 @@ XrdSys::IOEvents::Channel::Channel(Poller   *pollP, int   fd,
 }
 
 /******************************************************************************/
-/*                            D e s t r u c t o r                             */
+/*                                D e l e t e                                 */
 /******************************************************************************/
 
-XrdSys::IOEvents::Channel::~Channel()
+void XrdSys::IOEvents::Channel::Delete()
 {
    Poller *myPoller;
    bool isLocked = true;
 
 // Lock ourselves during the delete process. If the channel is disassociated
-// then we need do nothing more.
+// or the real poller is set to the error poller then this channel is clean
+// and can be deleted (i.e. the channel ran through Detach()).
 //
    chMutex.Lock();
    if (!chPollXQ || chPollXQ == &pollErr1)
       {chMutex.UnLock();
+       delete this;
        return;
       }
 
@@ -312,6 +316,9 @@ XrdSys::IOEvents::Channel::~Channel()
            cbDone.Wait();
           }
       }
+// It is now safe to release the storage
+//
+   delete this;
 }
   
 /******************************************************************************/
@@ -320,7 +327,7 @@ XrdSys::IOEvents::Channel::~Channel()
   
 bool XrdSys::IOEvents::Channel::Disable(int events, const char **eText)
 {
-   int eNum, newev, curev;
+   int eNum = 0, newev, curev;
    bool retval = true, isLocked = true;
 
 // Lock this channel
@@ -332,6 +339,10 @@ bool XrdSys::IOEvents::Channel::Disable(int events, const char **eText)
    if (chPoller == &pollWait) curev = static_cast<int>(reMod);
       else                    curev = static_cast<int>(chEvents);
 
+// Trace this entry
+//
+   IF_TRACE(Disable,chFD,"->Disable(" <<events <<") chev=" <<curev);
+
 // Calculate new event mask
 //
    events &= allEvents;
@@ -343,7 +354,7 @@ bool XrdSys::IOEvents::Channel::Disable(int events, const char **eText)
    if (newev != curev)
       {chEvents = newev;
        retval = chPoller->Modify(this, eNum, eText, isLocked);
-       TRACE_MOD(Disable,chFD,newev,curev);
+       TRACE_MOD(Disable,chFD,newev);
       } else {
        TRACE_NOD(Disable,chFD,newev);
       }
@@ -362,13 +373,8 @@ bool XrdSys::IOEvents::Channel::Disable(int events, const char **eText)
 bool XrdSys::IOEvents::Channel::Enable(int events, int timeout,
                                        const char **eText)
 {
-   time_t newDL;
-   int eNum, newev, curev;
-   bool retval, isLocked = true, setTO = false;
-
-// Trace this entry
-//
-   IF_TRACE(Enable,chFD,"->Enable(" <<events <<',' <<timeout <<')');
+   int eNum = 0, newev, curev, tmoSet = 0;
+   bool retval, setTO, isLocked = true;
 
 // Lock ourselves against any changes (this is a recursive mutex)
 //
@@ -379,6 +385,10 @@ bool XrdSys::IOEvents::Channel::Enable(int events, int timeout,
    if (chPoller == &pollWait) curev = static_cast<int>(reMod);
       else                    curev = static_cast<int>(chEvents);
 
+// Trace this entry
+//
+   IF_TRACE(Enable,chFD,"->Enable("<<events<<','<<timeout<<") chev="<<curev);
+
 // Establish events that should be enabled
 //
    events &= allEvents;
@@ -387,32 +397,22 @@ bool XrdSys::IOEvents::Channel::Enable(int events, int timeout,
 
 // Handle timeout changes now
 //
-   if (timeout)
-      {newDL = (timeout > 0 ? timeout + time(0) : Poller::maxTime);
-       if (events & readEvents)
-          {chRTO = (timeout < 0 ? 0 : timeout);
-           if (newDL != rdDL) {rdDL  = newDL; setTO = true;}
-          }
-       if (events & writeEvents)
-          {chWTO = (timeout < 0 ? 0 : timeout);
-           if (newDL != wrDL) {wrDL  = newDL; setTO = true;}
-          }
-      } else {
-       time_t nowTime = (chRTO || chWTO ? time(0) : 0);
-       if (events & readEvents)
-          {newDL = (chRTO ? nowTime + chRTO : Poller::maxTime);
-           if (newDL != rdDL) {rdDL  = newDL; setTO = true;}
-          }
-       if (events & writeEvents)
-          {newDL = (chWTO ? nowTime + chWTO : Poller::maxTime);
-           if (newDL != wrDL) {wrDL  = newDL; setTO = true;}
-          }
+   if (REVENTS(events))
+      {     if (timeout > 0) chRTO = timeout;
+       else if (timeout < 0) chRTO = 0;
+       if (rdDL != Poller::maxTime || chRTO) tmoSet |= CallBack::ReadyToRead;
+      }
+
+   if (WEVENTS(events))
+      {     if (timeout > 0) chWTO = timeout;
+       else if (timeout < 0) chWTO = 0;
+       if (wrDL != Poller::maxTime || chWTO) tmoSet |= CallBack::ReadyToWrite;
       }
 
 // Check if we have to reset the timeout. We need to hold the channel lock here.
 //
-   if (setTO && chPoller != &pollErr1)
-           setTO = chPollXQ->TmoAdd(this);
+   if (tmoSet && chPoller != &pollErr1)
+           setTO = chPollXQ->TmoAdd(this, tmoSet);
       else setTO = false;
 
 // Check if any modifcations needed here. If so, invoke the modifier. Note that
@@ -423,7 +423,7 @@ bool XrdSys::IOEvents::Channel::Enable(int events, int timeout,
 //
    if (newev)
       {retval = chPoller->Modify(this, eNum, eText, isLocked);
-       TRACE_MOD(Enable,chFD,(curev | events),curev);
+       TRACE_MOD(Enable,chFD,(curev | events));
       } else {
        retval = true;
        TRACE_NOD(Enable,chFD,(curev | events));
@@ -436,7 +436,8 @@ bool XrdSys::IOEvents::Channel::Enable(int events, int timeout,
 // that we cannot hold the channel mutex for this call because it may wait.
 //
    if (isLocked) chMutex.UnLock();
-   if (retval && !(chPollXQ->wakePend) && setTO && isLocked) chPollXQ->WakeUp();
+   bool isWakePend = CPP_ATOMIC_LOAD(chPollXQ->wakePend, std::memory_order_consume);
+   if (retval && !isWakePend && setTO && isLocked) chPollXQ->WakeUp();
 
 // All done
 //
@@ -553,6 +554,7 @@ XrdSys::IOEvents::Poller::Poller(int rFD, int cFD)
    pipeBlen        = 0;
    pipePoll.fd     = rFD;
    pipePoll.events = POLLIN | POLLRDNORM;
+   tmoMask         = 255;
 }
 
 /******************************************************************************/
@@ -607,8 +609,7 @@ bool XrdSys::IOEvents::Poller::CbkXeq(XrdSys::IOEvents::Channel *cP, int events,
 {
    XrdSysMutexHelper cbkMHelp(cP->chMutex);
    char oldEvents;
-   int isRead = 0, isWrite = 0;
-   bool cbok, retval, isLocked = true;
+   bool cbok, retval, isRead, isWrite, isLocked = true;
 
 // Perform any required tracing
 //
@@ -617,6 +618,7 @@ bool XrdSys::IOEvents::Poller::CbkXeq(XrdSys::IOEvents::Channel *cP, int events,
                             (cP->chPoller == &pollInit    ? "init" :
                             (cP->chPoller == &pollWait    ? "wait" : "err")));
        DO_TRACE(CbkXeq,cP->chFD,"callback events=" <<events
+            <<" chev=" <<static_cast<int>(cP->chEvents)
             <<" toq=" <<(cP->inTOQ != 0) <<" erc=" <<eNum
             <<" callback " <<(cP->chCB ? "present" : "missing")
             <<" poller=" <<cbtype);
@@ -627,10 +629,14 @@ bool XrdSys::IOEvents::Poller::CbkXeq(XrdSys::IOEvents::Channel *cP, int events,
 //
    if (cP->inTOQ)
       {TmoDel(cP);
-       isRead = (events & (CallBack::ReadyToRead  | CallBack:: ReadTimeOut));
+       cP->dlType |= (events & CallBack::ValidEvents) << 4;
+       isRead = events & (CallBack::ReadyToRead | CallBack:: ReadTimeOut);
        if (isRead)  cP->rdDL = maxTime;
-       isWrite= (events & (CallBack::ReadyToWrite | CallBack::WriteTimeOut));
+       isWrite= events & (CallBack::ReadyToWrite | CallBack::WriteTimeOut);
        if (isWrite) cP->wrDL = maxTime;
+      } else {
+       cP->dlType &= CallBack::ValidEvents;
+       isRead = isWrite = false;
       }
 
 // Verify that there is a callback here and the channel is ready. If not,
@@ -647,7 +653,7 @@ bool XrdSys::IOEvents::Poller::CbkXeq(XrdSys::IOEvents::Channel *cP, int events,
        oldEvents = cP->chEvents;
        cP->chEvents = 0;
        retval = cP->chPoller->Modify(cP, eNum, 0, isLocked);
-       TRACE_MOD(CbkXeq,cP->chFD,0,oldEvents);
+       TRACE_MOD(CbkXeq,cP->chFD,0);
        if (!isLocked) cP->chMutex.Lock();
        cP->chEvents = oldEvents;
        return true;
@@ -664,11 +670,14 @@ bool XrdSys::IOEvents::Poller::CbkXeq(XrdSys::IOEvents::Channel *cP, int events,
            chDead       = false;
            cbkMHelp.UnLock();
            cP->chCB->Fatal(cP,cP->chCBA, eNum, eTxt);
-           return (chDead ? true : false);
+           if (chDead) return true;
+           cbkMHelp.Lock(&(cP->chMutex));
+           cP->inPSet   = 0;
+           return false;
           }
             if (REVENTS(cP->chEvents)) events = CallBack::ReadyToRead;
        else if (WEVENTS(cP->chEvents)) events = CallBack::ReadyToWrite;
-       else    {cP->chPoller = &pollErr1; cP->chFault = eNum;
+       else    {cP->chPoller = &pollErr1; cP->chFault = eNum; cP->inPSet = 0;
                 return false;
                }
       }
@@ -703,14 +712,8 @@ bool XrdSys::IOEvents::Poller::CbkXeq(XrdSys::IOEvents::Channel *cP, int events,
 // the timeout if it hasn't been handled via a call from the callback.
 //
         if (!cbok) Detach(cP,isLocked,false);
-   else if (!(cP->inTOQ) && (cP->chRTO || cP->chWTO))
-           {time_t tNow = time(0);
-            if (isRead)  cP->rdDL = (REVENTS(cP->chEvents) && cP->chRTO
-                                  ?  cP->chRTO + tNow : maxTime);
-            if (isWrite) cP->wrDL = (WEVENTS(cP->chEvents) && cP->chWTO
-                                  ?  cP->chWTO + tNow : maxTime);
-            if (cP->rdDL != maxTime || cP->wrDL != maxTime) TmoAdd(cP);
-           }
+   else if ((isRead || isWrite) && !(cP->inTOQ) && (cP->chRTO || cP->chWTO))
+           TmoAdd(cP, 0);
 
 // All done. While the mutex should not have been unlocked, we relock it if
 // it has to keep the mutex helper from croaking.
@@ -724,7 +727,8 @@ bool XrdSys::IOEvents::Poller::CbkXeq(XrdSys::IOEvents::Channel *cP, int events,
 /******************************************************************************/
   
 XrdSys::IOEvents::Poller *XrdSys::IOEvents::Poller::Create(int         &eNum,
-                                                           const char **eTxt)
+                                                           const char **eTxt,
+                                                           int          crOpts)
 {
    int fildes[2];
    struct pollArg pArg;
@@ -732,13 +736,11 @@ XrdSys::IOEvents::Poller *XrdSys::IOEvents::Poller::Create(int         &eNum,
 
 // Create a pipe used to break the poll wait loop
 //
-   if (pipe(fildes))
+   if (XrdSysFD_Pipe(fildes))
       {eNum = errno;
        if (eTxt) *eTxt = "creating poll pipe";
        return 0;
       }
-   fcntl(fildes[0], F_SETFD, FD_CLOEXEC);
-   fcntl(fildes[1], F_SETFD, FD_CLOEXEC);
 
 // Create an actual implementation of a poller
 //
@@ -755,8 +757,11 @@ XrdSys::IOEvents::Poller *XrdSys::IOEvents::Poller::Create(int         &eNum,
       {if (eTxt) *eTxt = "creating poller thread"; return 0;}
 
 // Now wait for the thread to finish initializing before we allow use
+// Note that the bootstrap takes ownership of the semaphore and will delete it
+// once the thread positing the semaphore actually ends. This is to avoid
+// semaphore bugs present in certain (e.g. Linux) kernels.
 //
-   pArg.pollSync.Wait();
+   pArg.pollSync->Wait();
 
 // Check if all went well
 //
@@ -767,6 +772,11 @@ XrdSys::IOEvents::Poller *XrdSys::IOEvents::Poller::Create(int         &eNum,
        return 0;
       }
 
+// Set creation options in the new poller
+//
+   if (crOpts & optTOM)
+      pArg.pollP->tmoMask = ~(CallBack::ReadTimeOut|CallBack::WriteTimeOut);
+
 // All done
 //
    eNum = 0;
@@ -781,6 +791,8 @@ XrdSys::IOEvents::Poller *XrdSys::IOEvents::Poller::Create(int         &eNum,
 void XrdSys::IOEvents::Poller::Detach(XrdSys::IOEvents::Channel *cP,
                                       bool &isLocked, bool keep)
 {
+// The caller must hold the channel lock!
+//
    bool detFD = (cP->inPSet != 0);
 
 // First remove the channel from the timeout queue
@@ -809,7 +821,10 @@ void XrdSys::IOEvents::Poller::Detach(XrdSys::IOEvents::Channel *cP,
 // Exclude this channel from the associated poll set, don't hold the ad lock
 //
    adMutex.UnLock();
-   if (detFD && cmdFD >= 0) Exclude(cP, isLocked, !ISPOLLER);
+   if (detFD)
+      {cP->inPSet = 0;
+       if (cmdFD >= 0) Exclude(cP, isLocked, !ISPOLLER);
+      }
 }
   
 /******************************************************************************/
@@ -910,7 +925,7 @@ bool XrdSys::IOEvents::Poller::Init(XrdSys::IOEvents::Channel *cP, int &eNum,
             if (cP->reMod)
                {cP->chEvents =  cP->reMod;
                 retval = cP->chPoller->Modify(cP, eNum, eTxt, isLocked);
-                TRACE_MOD(Init,cP->chFD,int(cP->reMod),0);
+                TRACE_MOD(Init,cP->chFD,int(cP->reMod));
                 if (!isLocked) {cP->chMutex.Lock(); isLocked = true;}
                } else {
                 TRACE_NOD(Init,cP->chFD,0);
@@ -1039,11 +1054,12 @@ void XrdSys::IOEvents::Poller::Stop()
 /*                                T m o A d d                                 */
 /******************************************************************************/
   
-bool XrdSys::IOEvents::Poller::TmoAdd(XrdSys::IOEvents::Channel *cP)
+bool XrdSys::IOEvents::Poller::TmoAdd(XrdSys::IOEvents::Channel *cP, int tmoSet)
 {
    XrdSysMutexHelper mHelper(toMutex);
-   time_t rdDL, wrDL;
+   time_t tNow;
    Channel *ncP;
+   bool setRTO, setWTO;
 
 // Remove element from timeout queue if it is there
 //
@@ -1052,15 +1068,30 @@ bool XrdSys::IOEvents::Poller::TmoAdd(XrdSys::IOEvents::Channel *cP)
        cP->inTOQ = 0;
       }
 
+// Determine which timeouts need to be reset
+//
+   tmoSet|= cP->dlType >> 4;
+   setRTO = (tmoSet&tmoMask) & (CallBack::ReadyToRead |CallBack:: ReadTimeOut);
+   setWTO = (tmoSet&tmoMask) & (CallBack::ReadyToWrite|CallBack::WriteTimeOut);
+
+// Reset the required deadlines
+//
+   tNow = time(0);
+   if (setRTO && REVENTS(cP->chEvents) && cP->chRTO)
+      cP->rdDL = cP->chRTO + tNow;
+   if (setWTO && WEVENTS(cP->chEvents) && cP->chWTO)
+      cP->wrDL = cP->chWTO + tNow;
+
 // Calculate the closest enabled deadline
 //
-   rdDL = (cP->chEvents & Channel:: readEvents ? cP->rdDL : maxTime);
-   wrDL = (cP->chEvents & Channel::writeEvents ? cP->wrDL : maxTime);
-   if (rdDL < wrDL) {cP->deadLine = rdDL; cP->dlType  = CallBack::ReadTimeOut;}
-      else {if (rdDL != wrDL) cP->dlType = CallBack::WriteTimeOut;
-               else cP->dlType  = CallBack::ReadTimeOut|CallBack::WriteTimeOut;
-            cP->deadLine = wrDL;
-           }
+   if (cP->rdDL < cP->wrDL)
+      {cP->deadLine  = cP->rdDL; cP->dlType  = CallBack:: ReadTimeOut;
+      } else {
+       cP->deadLine  = cP->wrDL; cP->dlType  = CallBack::WriteTimeOut;
+       if (cP->rdDL == cP->wrDL) cP->dlType |= CallBack:: ReadTimeOut;
+      }
+   IF_TRACE(TmoAdd, cP->chFD, "t=" <<tNow <<" rdDL=" <<setRTO <<' ' <<cP->rdDL
+                                          <<" wrDL=" <<setWTO <<' ' <<cP->wrDL);
 
 // If no timeout really applies, we are done
 //
@@ -1122,7 +1153,7 @@ int XrdSys::IOEvents::Poller::TmoGet()
 
 // Return the value
 //
-   wakePend = false;
+   CPP_ATOMIC_STORE(wakePend, false, std::memory_order_release);
    toMutex.UnLock();
    return wtval;
 }
@@ -1133,14 +1164,24 @@ int XrdSys::IOEvents::Poller::TmoGet()
 
 void XrdSys::IOEvents::Poller::WakeUp()
 {
-   static PipeData cmdbuff = {PipeData::NoOp, 0, 0, 0, 0};
+   static PipeData cmdbuff(PipeData::NoOp);
 
 // Send it off to wakeup the poller thread, but only if here is no wakeup in
 // progress.
 //
+// We use a mutex here because we want to produce a synchronization point - all
+// threads that might be interested timeouts and wakeups are going to incur a
+// cache bounce for the page where wakePend resides; they will see a consistent
+// view of the wakePend flag.  For those threads, this is equivalent to
+// an atomic with memory_order std::memory_order_seq_cst (the strongest ordering).
+// However, the threads that are not interested in timeouts will not get a flush
+// for their copy of the wakePend page.  They will still have the weaker memory
+// ordering of consume/release (which is guaranteed anyway on all current architectures
+// except for DEC Alpha).
    toMutex.Lock();
-   if (wakePend) toMutex.UnLock();
-      else {wakePend = true;
+   bool isWakePend = CPP_ATOMIC_LOAD(wakePend, std::memory_order_consume);
+   if (isWakePend) {toMutex.UnLock();}
+      else {CPP_ATOMIC_STORE(wakePend, true, std::memory_order_release);
             toMutex.UnLock();
             SendCmd(cmdbuff);
            }
@@ -1154,6 +1195,8 @@ void XrdSys::IOEvents::Poller::WakeUp()
 #include "XrdSys/XrdSysIOEventsPollPort.icc"
 #elif defined( __linux__ )
 #include "XrdSys/XrdSysIOEventsPollE.icc"
+#elif defined(__APPLE__)
+#include "XrdSys/XrdSysIOEventsPollKQ.icc"
 #else
 #include "XrdSys/XrdSysIOEventsPollPoll.icc"
 #endif
