@@ -35,7 +35,6 @@
 #include <sys/param.h>
 #include <sys/stat.h>
 #include <sys/types.h>
-#include <sys/un.h>
 
 #ifdef __solaris__
 #include <sys/isa_defs.h>
@@ -57,7 +56,7 @@
 #include "XrdOuc/XrdOucStream.hh"
 #include "XrdOuc/XrdOucTrace.hh"
 #include "XrdOuc/XrdOucUtils.hh"
-#include "XrdSys/XrdSysDNS.hh"
+#include "XrdSec/XrdSecLoadSecurity.hh"
 #include "XrdSys/XrdSysError.hh"
 #include "XrdSys/XrdSysHeaders.hh"
 #include "XrdSys/XrdSysLogger.hh"
@@ -77,6 +76,7 @@
 #include "XrdXrootd/XrdXrootdXPath.hh"
 
 #include "Xrd/XrdBuffer.hh"
+#include "Xrd/XrdInet.hh"
 
 /******************************************************************************/
 /*         P r o t o c o l   C o m m a n d   L i n e   O p t i o n s          */
@@ -109,8 +109,6 @@ extern          XrdOucTrace       *XrdXrootdTrace;
 
                 XrdXrootdPrepare  *XrdXrootdPrepQ;
 
-                XrdOucReqID       *XrdXrootdReqID;
-
                 const char        *XrdXrootdInstance;
 
                 int                XrdXrootdPort;
@@ -133,15 +131,22 @@ int XrdXrootdProtocol::Configure(char *parms, XrdProtocol_Config *pi)
                              XrdSysLogger     *Logger,
                              const char       *configFn,
                              XrdOucEnv        *EnvInfo);
-   extern XrdSecService    *XrdXrootdloadSecurity(XrdSysError *, char *, 
-                                                  char *, void **);
-   extern XrdSfsFileSystem *XrdXrootdloadFileSystem(XrdSysError *, char *, 
-                                                    const char *);
+
+   extern XrdSfsFileSystem *XrdXrootdloadFileSystem(XrdSysError *, 
+                                                    XrdSfsFileSystem *,
+                                                    char *, int,
+                                                    const char *, XrdOucEnv *);
+   extern XrdSfsFileSystem *XrdDigGetFS
+                            (XrdSfsFileSystem *nativeFS,
+                             XrdSysLogger     *Logger,
+                             const char       *configFn,
+                             const char       *theParms);
    extern int optind, opterr;
 
+   XrdOucEnv myEnv;
    XrdXrootdXPath *xp;
-   void *secGetProt = 0;
-   char *adminp, *fsver, *rdf, *bP, *tmp, c, buff[1024];
+   XrdSecGetProt_t secGetProt = 0;
+   char *adminp, *rdf, *bP, *tmp, c, buff[1024];
    int i, n, deper = 0;
 
 // Copy out the special info we want to use at top level
@@ -208,6 +213,7 @@ int XrdXrootdProtocol::Configure(char *parms, XrdProtocol_Config *pi)
 //
    if (!(as_miniosz = as_segsize/2)) as_miniosz = as_segsize;
    maxTransz = maxBuffsz = BPool->MaxSize();
+   memset(Route, 0, sizeof(Route));
 
 // Now process and configuration parameters
 //
@@ -215,13 +221,18 @@ int XrdXrootdProtocol::Configure(char *parms, XrdProtocol_Config *pi)
    if (rdf && Config(rdf)) return 0;
    if (pi->DebugON) XrdXrootdTrace->What = TRACE_ALL;
 
+// Check if we are exporting a generic object name
+//
+   if (XPList.Opts() & XROOTDXP_NOSLASH)
+      {eDest.Say("Config exporting ", XPList.Path(n)); n += 2;}
+      else n = 0;
+
 // Check if we are exporting anything
 //
-   if (!(xp = XPList.Next()))
+   if (!(xp = XPList.Next()) && !n)
       {XPList.Insert("/tmp"); n = 8;
        eDest.Say("Config warning: only '/tmp' will be exported.");
       } else {
-       n = 0;
        while(xp) {eDest.Say("Config exporting ", xp->Path(i));
                   n += i+2; xp = xp->Next();
                  }
@@ -230,6 +241,8 @@ int XrdXrootdProtocol::Configure(char *parms, XrdProtocol_Config *pi)
 // Export the exports
 //
    bP = tmp = (char *)malloc(n);
+   if (XPList.Opts() & XROOTDXP_NOSLASH)
+      {strcpy(bP, XPList.Path(i)); bP += i, *bP++ = ' ';}
    xp = XPList.Next();
    while(xp) {strcpy(bP, xp->Path(i)); bP += i; *bP++ = ' '; xp = xp->Next();}
    *(bP-1) = '\0';
@@ -240,46 +253,80 @@ int XrdXrootdProtocol::Configure(char *parms, XrdProtocol_Config *pi)
    if (!SecLib) eDest.Say("Config warning: 'xrootd.seclib' not specified;"
                           " strong authentication disabled!");
       else {TRACE(DEBUG, "Loading security library " <<SecLib);
-            if (!(CIA = XrdXrootdloadSecurity(&eDest, SecLib, pi->ConfigFN,
-                                              &secGetProt)))
+            if (!(CIA = XrdSecLoadSecService(&eDest, pi->ConfigFN,
+                        (strcmp(SecLib,"default") ? SecLib : 0), &secGetProt)))
                {eDest.Emsg("Config", "Unable to load security system.");
                 return 0;
                }
            }
 
+// Set up the network for self-identification and display it
+//
+   pi->NetTCP->netIF.Port(Port);
+   pi->NetTCP->netIF.Display("Config ");
+
+// Establish our specific environment that will be passed along
+//
+   myEnv.PutPtr("XrdInet*", (void *)(pi->NetTCP));
+   myEnv.PutPtr("XrdNetIF*", (void *)(&(pi->NetTCP->netIF)));
+   myEnv.PutPtr("XrdSecGetProtocol*", (void *)secGetProt);
+   myEnv.PutPtr("XrdScheduler*", Sched);
+
+// Copy over the xrd environment which contains plugin argv's
+//
+   if (pi->theEnv) myEnv.PutPtr("xrdEnv*", pi->theEnv);
+
 // Get the filesystem to be used
 //
-   if (FSLib)
-      {TRACE(DEBUG, "Loading filesystem library " <<FSLib);
-       osFS = XrdXrootdloadFileSystem(&eDest, FSLib, pi->ConfigFN);
+   if (FSLib[0])
+      {TRACE(DEBUG, "Loading base filesystem library " <<FSLib[0]);
+       osFS = XrdXrootdloadFileSystem(&eDest, 0, FSLib[0], FSLvn[0],
+                                      pi->ConfigFN, &myEnv);
       } else {
-       XrdOucEnv myEnv;
-       myEnv.PutPtr("XrdInet*", (void *)(pi->NetTCP));
-       myEnv.PutPtr("XrdSecGetProtocol*", secGetProt);
        osFS = XrdSfsGetDefaultFileSystem(0,eDest.logger(),pi->ConfigFN,&myEnv);
       }
    if (!osFS)
       {eDest.Emsg("Config", "Unable to load file system.");
        return 0;
-      } else SI->setFS(osFS);
+      } else {
+       SI->setFS(osFS);
+       if (FSLib[0]) osFS->EnvInfo(&myEnv);
+      }
 
-// Check if the file system version matches our version
+// Check if we have a wrapper library
 //
-   if (chkfsV)
-      {fsver = (char *)osFS->getVersion();
-       if (strcmp(XrdVERSION, fsver))
-          eDest.Emsg("Config", "Warning! xrootd build version " XrdVERSION
-                               "differs from file system version ", fsver);
+   if (FSLib[1])
+      {TRACE(DEBUG, "Loading wrapper filesystem library " <<FSLib[1]);
+       osFS = XrdXrootdloadFileSystem(&eDest, osFS, FSLib[1], FSLvn[1],
+                                      pi->ConfigFN, &myEnv);
+       if (!osFS)
+          {eDest.Emsg("Config", "Unable to load file system wrapper.");
+           return 0;
+          } else osFS->EnvInfo(&myEnv);
+      }
+
+// Check if the diglib should be loaded. We only support the builtin one. In
+// the future we will have to change this code to be like the above.
+//
+   if (digParm)
+      {TRACE(DEBUG, "Loading dig filesystem builtin");
+       digFS = XrdDigGetFS(osFS, eDest.logger(), pi->ConfigFN, digParm);
+       if (!digFS) eDest.Emsg("Config","Unable to load digFS; "
+                                       "remote debugging disabled!");
       }
 
 // Check if we are going to be processing checksums locally
 //
    if (JobCKT && JobLCL)
       {XrdOucErrInfo myError("Config");
-       if (osFS->chksum(XrdSfsFileSystem::csSize,JobCKT,0,myError))
-          {eDest.Emsg("Config", JobCKT, " checksum is not natively supported.");
-           return 0;
-          }
+       XrdOucTList *tP = JobCKTLST;
+       do {if (osFS->chksum(XrdSfsFileSystem::csSize,tP->text,0,myError))
+              {eDest.Emsg("Config",tP->text,"checksum is not natively supported.");
+               return 0;
+              }
+           tP->ival[1] = myError.getErrInfo();
+           tP = tP->next;
+          } while(tP);
       }
 
 // Initialiaze for AIO
@@ -301,8 +348,7 @@ int XrdXrootdProtocol::Configure(char *parms, XrdProtocol_Config *pi)
 
 // Initialize the request ID generation object
 //
-   XrdXrootdReqID = new XrdOucReqID((int)Port, pi->myName,
-                                    XrdSysDNS::IPAddr(pi->myAddr));
+   PrepID = new XrdOucReqID(pi->urAddr, (int)Port);
 
 // Initialize for prepare processing
 //
@@ -326,9 +372,12 @@ int XrdXrootdProtocol::Configure(char *parms, XrdProtocol_Config *pi)
 //
    if ((xp = RPList.Next()))
       {int k;
-       char buff[512];
+       char buff[1024], puff[1024];
        do {k = xp->Opts();
-           sprintf(buff, " to %s:%d", Route[k].Host, Route[k].Port);
+           if (Route[k].Host[0] == Route[k].Host[1]
+           &&  Route[k].Port[0] == Route[k].Port[1]) *puff = 0;
+              else sprintf(puff, "%%%s:%d", Route[k].Host[1], Route[k].Port[1]);
+           sprintf(buff," to %s:%d%s",Route[k].Host[0],Route[k].Port[0],puff);
            eDest.Say("Config redirect static ", xp->Path(), buff);
            xp = xp->Next();
           } while(xp);
@@ -337,28 +386,39 @@ int XrdXrootdProtocol::Configure(char *parms, XrdProtocol_Config *pi)
    if ((xp = RQList.Next()))
       {int k;
        const char *cgi1, *cgi2;
-       char buff[1024], xCgi[RD_Num] = {0};
+       char buff[1024], puff[1024], xCgi[RD_Num] = {0};
        if (isRedir) {cgi1 = "+"; cgi2 = getenv("XRDCMSCLUSTERID");}
           else      {cgi1 = "";  cgi2 = pi->myName;}
+       myCNlen = snprintf(buff, sizeof(buff), "%s%s", cgi1, cgi2);
+       myCName = strdup(buff);
        do {k = xp->Opts();
-           sprintf(buff, " to %s:%d", Route[k].Host, Route[k].Port);
+           if (Route[k].Host[0] == Route[k].Host[1]
+           &&  Route[k].Port[0] == Route[k].Port[1]) *puff = 0;
+              else sprintf(puff, "%%%s:%d", Route[k].Host[1], Route[k].Port[1]);
+           sprintf(buff," to %s:%d%s",Route[k].Host[0],Route[k].Port[0],puff);
            eDest.Say("Config redirect enoent ", xp->Path(), buff);
            if (!xCgi[k] && cgi2)
-              {snprintf(buff,sizeof(buff), "%s?tried=%s%s", Route[k].Host,
-                                           cgi1, cgi2);
-               free(Route[k].Host); Route[k].Host = strdup(buff);
+              {bool isdup = Route[k].Host[0] == Route[k].Host[1]
+                         && Route[k].Port[0] == Route[k].Port[1];
+               for (i = 0; i < 2; i++)
+                   {n = snprintf(buff,sizeof(buff), "%s?tried=%s%s",
+                                 Route[k].Host[i], cgi1, cgi2);
+                    free(Route[k].Host[i]); Route[k].Host[i] = strdup(buff);
+                    Route[k].RDSz[i] = n;
+                    if (isdup) {Route[k].Host[1] = Route[k].Host[0];
+                                Route[k].RDSz[1] = n; break;
+                               }
+                   }
               }
            xCgi[k] = 1;
            xp = xp->Next();
           } while(xp);
       }
 
-// Check if monitoring should be enabled
+// Initialize monitoring (it won't do anything if it wasn't enabled)
 //
-   if (!isRedir || XrdXrootdMonitor::Redirect())
-      {if (!XrdXrootdMonitor::Init(Sched, &eDest, pi->myName, pi->myProg,
-                                   myInst, Port)) return 0;
-      }
+   if (!XrdXrootdMonitor::Init(Sched, &eDest, pi->myName, pi->myProg,
+                               myInst, Port)) return 0;
 
 // Add all jobs that we can run to the admin object
 //
@@ -414,6 +474,7 @@ int XrdXrootdProtocol::Config(const char *ConfigFN)
          if (ismine)
             {     if TS_Xeq("async",         xasync);
              else if TS_Xeq("chksum",        xcksum);
+             else if TS_Xeq("diglib",        xdig);
              else if TS_Xeq("export",        xexp);
              else if TS_Xeq("fslib",         xfsl);
              else if TS_Xeq("log",           xlog);
@@ -503,8 +564,8 @@ int XrdXrootdProtocol::xasync(XrdOucStream &Config)
     int  V_limit=-1, V_msegs=-1, V_mtot=-1, V_minsz=-1, V_segsz=-1;
     int  V_minsf=-1;
     long long llp;
-    static struct asyncopts {const char *opname; int minv; int *oploc;
-                             const char *opmsg;} asopts[] =
+    struct asyncopts {const char *opname; int minv; int *oploc;
+                      const char *opmsg;} asopts[] =
        {
         {"force",     -1, &V_force, ""},
         {"off",       -1, &V_off,   ""},
@@ -589,10 +650,13 @@ int XrdXrootdProtocol::xasync(XrdOucStream &Config)
 
 /* Function: xcksum
 
-   Purpose:  To parse the directive: chksum [max <n>] <type> [<path>]
+   Purpose:  To parse the directive: chksum [chkcgi] [max <n>] <type> [<path>]
 
              max       maximum number of simultaneous jobs
-             <type>    algorithm of checksum (e.g., md5)
+             chkcgi    Always check for checksum type in cgo info.
+             <type>    algorithm of checksum (e.g., md5). If more than one
+                       checksum is supported then they should be listed with
+                       each separated by a space.
              <path>    the path of the program performing the checksum
                        If no path is given, the checksum is local.
 
@@ -603,13 +667,22 @@ int XrdXrootdProtocol::xcksum(XrdOucStream &Config)
 {
    static XrdOucProg *theProg = 0;
    int (*Proc)(XrdOucStream *, char **, int) = 0;
+   XrdOucTList *tP, *algFirst = 0, *algLast = 0;
    char *palg, prog[2048];
-   int jmax = 4;
+   int jmax = 4, anum[2] = {0,0};
 
 // Get the algorithm name and the program implementing it
 //
+   JobCKCGI = 0;
    while ((palg = Config.GetWord()) && *palg != '/')
-         {if (strcmp(palg, "max")) break;
+         {if (!strcmp(palg,"chkcgi")) {JobCKCGI = 1; continue;}
+          if (strcmp(palg, "max"))
+             {XrdOucTList *xalg = new XrdOucTList(palg, anum); anum[0]++;
+              if (algLast) algLast->next = xalg;
+                 else      algFirst      = xalg;
+              algLast = xalg;
+              continue;
+             }
           if (!(palg = Config.GetWord()))
              {eDest.Emsg("Config", "chksum max not specified"); return 1;}
           if (XrdOuca2x::a2i(eDest, "chksum max", palg, &jmax, 0)) return 1;
@@ -617,16 +690,31 @@ int XrdXrootdProtocol::xcksum(XrdOucStream &Config)
 
 // Verify we have an algoritm
 //
-   if (!palg || *palg == '/')
+   if (!algFirst)
       {eDest.Emsg("Config", "chksum algorithm not specified"); return 1;}
    if (JobCKT) free(JobCKT);
-   JobCKT = strdup(palg);
+   JobCKT = strdup(algFirst->text);
+
+// Handle alternate checksums
+//
+   while((tP = JobCKTLST)) {JobCKTLST = tP->next; delete tP;}
+   JobCKTLST = algFirst;
+   if (algFirst->next) JobCKCGI = 2;
+
+// Handle program if we have one
+//
+   if (palg)
+      {int n = strlen(palg);
+       if (n+2 >= (int)sizeof(prog))
+          {eDest.Emsg("Config", "cksum program too long"); return 1;}
+       strcpy(prog, palg); palg = prog+n; *palg++ = ' '; n = sizeof(prog)-n-1;
+       if (!Config.GetRest(palg, n))
+          {eDest.Emsg("Config", "cksum parameters too long"); return 1;}
+      } else *prog = 0;
 
 // Check if we have a program. If not, then this will be a local checksum and
 // the algorithm will be verified after we load the filesystem.
 //
-   if (!Config.GetRest(prog, sizeof(prog)))
-      {eDest.Emsg("Config", "cksum parameters too long"); return 1;}
    if (*prog) JobLCL = 0;
       else {  JobLCL = 1; Proc = &CheckSum; strcpy(prog, "chksum");}
 
@@ -640,6 +728,46 @@ int XrdXrootdProtocol::xcksum(XrdOucStream &Config)
    return 0;
 }
   
+/******************************************************************************/
+/*                                  x d i g                                   */
+/******************************************************************************/
+
+/* Function: xdig
+
+   Purpose:  To parse the directive: diglib * <parms>
+
+             *         use builtin digfs library (only one supported now).
+             parms     parameters for digfs.
+
+  Output: 0 upon success or !0 upon failure.
+*/
+
+int XrdXrootdProtocol::xdig(XrdOucStream &Config)
+{
+    char parms[4096], *val;
+
+// Get the path
+//
+   if (!(val = Config.GetWord()))
+      {eDest.Emsg("Config", "digfslib not specified"); return 1;}
+
+// Make sure it refers to an internal one
+//
+   if (strcmp(val, "*"))
+      {eDest.Emsg("Config", "builtin diglib not specified"); return 1;}
+
+// Grab the parameters
+//
+    if (!Config.GetRest(parms, sizeof(parms)))
+       {eDest.Emsg("Config", "diglib parameters too long"); return 1;}
+    if (digParm) free(digParm);
+    digParm = strdup(parms);
+
+// All done
+//
+   return 0;
+}
+  
 /******************************************************************************/
 /*                                  x e x p                                   */
 /******************************************************************************/
@@ -681,7 +809,19 @@ int XrdXrootdProtocol::xexp(XrdOucStream &Config)
 
 int XrdXrootdProtocol::xexpdo(char *path, int popt)
 {
-   const char *opaque;
+   char *opaque;
+
+// Check if we are exporting a generic name
+//
+   if (*path == '*')
+      {popt |= XROOTDXP_NOSLASH | XROOTDXP_NOCGI;
+       if (*(path+1))
+          {if (*(path+1) == '?') popt &= ~XROOTDXP_NOCGI;
+              else {eDest.Emsg("Config","invalid export path -",path);return 1;}
+          }
+       XPList.Set(popt, path);
+       return 0;
+      }
 
 // Make sure path start with a slash
 //
@@ -700,36 +840,92 @@ int XrdXrootdProtocol::xexpdo(char *path, int popt)
 
 /* Function: xfsl
 
-   Purpose:  To parse the directive: fslib [?] <path>
+   Purpose:  To parse the directive: fslib [throttle | [-2] <fspath2>]
+                                           {default  | [-2] <fspath1>}
 
-             ?         check if fslib build version matches our version
-             <path>    the path of the filesystem library to be used.
+             -2        Uses version2 of the plugin initializer.
+                       This is ignored now because it's always done.
+             throttle  load libXrdThrottle.so as the head interface.
+             <fspath2> load the named library as the head interface.
+             default   load libXrdOfs.so ro libXrdPss.so as the tail
+                       interface. This is the default.
+             <fspath1> load the named library as the tail interface.
 
   Output: 0 upon success or !0 upon failure.
 */
 
 int XrdXrootdProtocol::xfsl(XrdOucStream &Config)
 {
-    char *val, *Slash;
+    char *val;
+
+// Clear storage pointers
+//
+   if (FSLib[0]) {free(FSLib[0]); FSLib[0] = 0;}
+   if (FSLib[1]) {free(FSLib[1]); FSLib[1] = 0;}
+   FSLvn[0] = FSLvn[1] = 0;
 
 // Get the path
 //
-   chkfsV = 0;
-   if ((val = Config.GetWord()) && *val == '?' && !val[1])
-      {chkfsV = '?'; val = Config.GetWord();}
-
-   if (!val || !val[0])
+   if (!(val = Config.GetWord()))
       {eDest.Emsg("Config", "fslib not specified"); return 1;}
 
+// Check if this is "thottle"
+//
+   if (!strcmp("throttle", val))
+      {FSLib[1] = strdup("libXrdThrottle.so");
+       if (!(val = Config.GetWord()))
+          {eDest.Emsg("Config","fslib throttle target library not specified");
+           return 1;
+          }
+       return xfsL(Config, val, 0);
+      }
+
+// Check for default or default library, the common case
+//
+   if (xfsL(Config, val, 1))    return 1;
+   if (!FSLib[1])               return 0;
+
+// If we dont have another token, then demote the previous library
+//
+   if (!(val = Config.GetWord()))
+      {FSLib[0] = FSLib[1]; FSLib[1] = 0;
+       FSLvn[0] = FSLvn[1]; FSLvn[1] = 0;
+       return 0;
+      }
+
+// Check for default or default library, the common case
+//
+   return xfsL(Config, val, 0);
+}
+
+/******************************************************************************/
+
+int XrdXrootdProtocol::xfsL(XrdOucStream &Config, char *val, int lix)
+{
+    char *Slash;
+    int lvn = 0;
+
+// Check if this is a version token
+//
+   if (!strcmp(val, "-2"))
+      {lvn = 2;
+       if (!(val = Config.GetWord()))
+          {eDest.Emsg("Config", "fslib not specified"); return 1;}
+      }
+
+// We will play fast and furious with the syntax as "default" should not be
+// prefixed with a version number but will let that pass.
+//
+   if (!strcmp("default", val)) return 0;
+
 // If this is the "standard" name tell the user that we are ignoring this lib.
 // Otherwise, record the path and return.
 //
    if (!(Slash = rindex(val, '/'))) Slash = val;
-   if (!strcmp(Slash, "/libXrdOfs.so"))
+      else Slash++;
+   if (!strcmp(Slash, "libXrdOfs.so"))
       eDest.Say("Config warning: ignoring fslib; libXrdOfs.so is built-in.");
-      else {if (FSLib) free(FSLib);
-            FSLib = strdup(val);
-           }
+      else {FSLib[lix] = strdup(val); FSLvn[lix] = lvn;}
    return 0;
 }
 
@@ -823,7 +1019,7 @@ int XrdXrootdProtocol::xmon(XrdOucStream &Config)
                                   "monitor rbuff value not specified",
                                   "monitor mbuff", "monitor rbuff"
                                  };
-    char  *val, *cp, *monDest[2] = {0, 0};
+    char  *val = 0, *cp, *monDest[2] = {0, 0};
     long long tempval;
     int i, monFlash = 0, monFlush=0, monMBval=0, monRBval=0, monWWval=0;
     int    monIdent = 3600, xmode=0, monMode[2] = {0, 0}, mrType, *flushDest;
@@ -1064,7 +1260,8 @@ int XrdXrootdProtocol::xprep(XrdOucStream &Config)
   
 /* Function: xred
 
-   Purpose:  To parse the directive: redirect <host>:<port> {<funcs>|[?]<path>}
+   Purpose:  To parse the directive: redirect <host>:<port>[%<prvhost>:<port>]
+                                              {<funcs>|[?]<path>}
 
              <funcs>   are one or more of the following functions that will
                        be immediately redirected to <host>:<port>. Each function
@@ -1098,21 +1295,40 @@ int XrdXrootdProtocol::xred(XrdOucStream &Config)
         {"stat",     RD_stat},
         {"trunc",    RD_trunc}
        };
-    char rHost[512], *val, *pp;
-    int i, k, neg, rPort, numopts = sizeof(rdopts)/sizeof(struct rediropts);
-    int isQ = 0;
+    static const int rHLen = 264;
+    char rHost[2][rHLen], *hP[2], *val, *pp;
+    int i, k, neg, numopts = sizeof(rdopts)/sizeof(struct rediropts);
+    int rPort[2], isQ = 0;
 
 // Get the host and port
 //
    val = Config.GetWord();
-   if (!val || !val[0] || val[0] == ':')
-      {eDest.Emsg("Config", "redirect host not specified"); return 1;}
-   if (!(pp = index(val, ':')))
-      {eDest.Emsg("Config", "redirect port not specified"); return 1;}
-   if (!(rPort = atoi(pp+1)))
-      {eDest.Emsg("Config", "redirect port is invalid");    return 1;}
-   *pp = '\0';
-   strlcpy(rHost, val, sizeof(rHost));
+
+// Check if we have two hosts here
+//
+   hP[0] = val;
+   if (!(pp = index(val, '%'))) hP[1] = 0;
+      else {hP[1] = pp+1; *pp = 0;}
+
+// Verify corectness here
+//
+   if (!(*val) || (hP[1] && !hP[1]))
+      {eDest.Emsg("Config", "malformed redirect host specification"); return 1;}
+
+// Process the hosts
+//
+   for (i = 0; i < 2; i++)
+       {if (!(val = hP[i])) break;
+        if (!val || !val[0] || val[0] == ':')
+           {eDest.Emsg("Config", "redirect host not specified"); return 1;}
+        if (!(pp = rindex(val, ':')))
+           {eDest.Emsg("Config", "redirect port not specified"); return 1;}
+        if (!(rPort[i] = atoi(pp+1)))
+           {eDest.Emsg("Config", "redirect port is invalid");    return 1;}
+        *pp = '\0';
+        strlcpy(rHost[i], val, rHLen);
+        hP[i] = rHost[i];
+       }
 
 // Set all redirect target functions
 //
@@ -1132,11 +1348,10 @@ int XrdXrootdProtocol::xred(XrdOucStream &Config)
                }
            }
         for (k = static_cast<int>(RD_open1); k < RD_Num; k++)
-            if (!Route[k].Host
-            || (!strcmp(Route[k].Host, rHost) && Route[k].Port == rPort)) break;
+            if (xred_xok(k, hP, rPort)) break;
         if (k >= RD_Num)
            {eDest.Emsg("Config", "too many diffrent path redirects"); return 1;}
-        xred_set(RD_func(k), rHost, rPort);
+        xred_set(RD_func(k), hP, rPort);
         do {if (isQ) RQList.Insert(val, k, 0);
                else  RPList.Insert(val, k, 0);
             if ((val = Config.GetWord()) && *val != '/')
@@ -1150,13 +1365,13 @@ int XrdXrootdProtocol::xred(XrdOucStream &Config)
     while (val)
           {if (!strcmp(val, "all"))
               {for (i = 0; i < numopts; i++)
-                   xred_set(rdopts[i].opval, rHost, rPort);
+                   xred_set(rdopts[i].opval, hP, rPort);
               }
               else {if ((neg = (val[0] == '-' && val[1]))) val++;
                     for (i = 0; i < numopts; i++)
                        {if (!strcmp(val, rdopts[i].opname))
                            {if (neg) xred_set(rdopts[i].opval, 0, 0);
-                               else  xred_set(rdopts[i].opval, rHost, rPort);
+                               else  xred_set(rdopts[i].opval, hP, rPort);
                             break;
                            }
                        }
@@ -1168,14 +1383,46 @@ int XrdXrootdProtocol::xred(XrdOucStream &Config)
    return 0;
 }
 
-void XrdXrootdProtocol::xred_set(RD_func func, const char *rHost, int rPort)
+
+void XrdXrootdProtocol::xred_set(RD_func func, char *rHost[2], int rPort[2])
 {
 
 // Reset static redirection
 //
-   if (Route[func].Host) free(Route[func].Host);
-   Route[func].Host = (rHost ? strdup(rHost) : 0);
-   Route[func].Port = rPort;
+   if (Route[func].Host[0]) free(Route[func].Host[0]);
+   if (Route[func].Host[0] != Route[func].Host[1]) free(Route[func].Host[1]);
+
+   if (rHost)
+      {Route[func].Host[0] = strdup(rHost[0]);
+       Route[func].Port[0] = rPort[0];
+      } else {
+       Route[func].Host[0] = Route[func].Host[1] = 0;
+       Route[func].Port[0] = Route[func].Port[1] = 0;
+       return;
+      }
+
+   if (!rHost[1])
+      {Route[func].Host[1] = Route[func].Host[0];
+       Route[func].Port[1] = Route[func].Port[0];
+      } else {
+       Route[func].Host[1] = strdup(rHost[1]);
+       Route[func].Port[1] = rPort[1];
+      }
+}
+
+bool XrdXrootdProtocol::xred_xok(int func, char *rHost[2], int rPort[2])
+{
+   if (!Route[func].Host[0]) return true;
+
+   if (strcmp(Route[func].Host[0], rHost[0])
+   ||  Route[func].Port[0] != rPort[0]) return false;
+
+   if (!rHost[1]) return Route[func].Host[0] == Route[func].Host[1];
+
+   if (strcmp(Route[func].Host[1], rHost[1])
+   ||  Route[func].Port[1] != rPort[1]) return false;
+
+   return true;
 }
 
 /******************************************************************************/
@@ -1184,9 +1431,10 @@ void XrdXrootdProtocol::xred_set(RD_func func, const char *rHost, int rPort)
 
 /* Function: xsecl
 
-   Purpose:  To parse the directive: seclib <path>
+   Purpose:  To parse the directive: seclib {default | <path>}
 
              <path>    the path of the security library to be used.
+                       "default" uses the default security library.
 
   Output: 0 upon success or !0 upon failure.
 */
